INCLUDE(CMakeForceCompiler)

IF(NOT CMAKE_TOOLCHAIN_FILE)
	MESSAGE(FATAL_ERROR "CMAKE_TOOLCHAIN_FILE not defined")
ENDIF()

SET(CMAKE_EXECUTABLE_SUFFIX "$CACHE{CMAKE_EXECUTABLE_SUFFIX}")

IF(${CMAKE_BUILD_TYPE} MATCHES "Release")
	SET(CTR_BUILD_TYPE "release")
	ADD_DEFINITIONS(-DNDEBUG)
ELSEIF(${CMAKE_BUILD_TYPE} MATCHES "Debug")
	SET(CTR_BUILD_TYPE "noopt")
ELSE()
	SET(CTR_BUILD_TYPE "verbose")
ENDIF()

GET_FILENAME_COMPONENT(
	tmp
	"${NLIB_RELEASE_ROOT}/libraries/CTR-TS.Process.MPCore/${CTR_BUILD_TYPE}"
	ABSOLUTE)
nlib_cygwin_getwinpath(${tmp} NLIB_DIR)

MESSAGE(STATUS "CTR_BUILD_TYPE=${CTR_BUILD_TYPE}")

SET(CAT_DIR "vc140amd64")  # only for pc tools
SET(ZLIB_FOUND 1)

SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_DEFINITIONS "NLIB_HAS_ZLIB")

MACRO(library_default_build_rule target)
	ADD_LIBRARY(${target} STATIC ${LIB_SRC} ${LIB_HEADER})
	SET_PROPERTY(TARGET ${target} APPEND PROPERTY INTERFACE_INCLUDE_DIRECTORIES "$<BUILD_INTERFACE:${NLIB_PKGROOT_DIR}/include>")
	TARGET_INCLUDE_DIRECTORIES(${target} BEFORE PRIVATE "${NLIB_PKGROOT_DIR}/include" "${NLIB_PKGROOT_DIR}/sources")
	INSTALL(TARGETS ${target} DESTINATION libraries/CTR-TS.Process.MPCore/${CTR_BUILD_TYPE})
	INSTALL(DIRECTORY ${NLIB_PKGROOT_DIR}/include DESTINATION .)
ENDMACRO()

MACRO(generate_ctr_makeromopt target)
	GET_TARGET_PROPERTY(ctr_spec_desc ${target} CTR_SPEC_DESC)
	GET_TARGET_PROPERTY(ctr_spec_rsf ${target} CTR_SPEC_RSF)
	GET_TARGET_PROPERTY(ctr_banner ${target} CTR_BANNER)
	GET_TARGET_PROPERTY(ctr_icon ${target} CTR_ICON)
	nlib_cygwin_getwinpath(${ctr_spec_desc} ctr_spec_desc)
	nlib_cygwin_getwinpath(${ctr_spec_rsf} ctr_spec_rsf)
	nlib_cygwin_getwinpath(${ctr_banner} ctr_banner)
	nlib_cygwin_getwinpath(${ctr_icon} ctr_icon)
	FILE(WRITE ${CMAKE_BINARY_DIR}/CMakeFiles/${target}.dir/makeromopt.txt
		"-desc ${ctr_spec_desc} -rsf ${ctr_spec_rsf} -DROMFS_ROOT= -DTITLE=CtrApp -banner ${ctr_banner} -icon ${ctr_icon} -f card")
ENDMACRO()

MACRO(generate_ctr_default_makeromopt target)
	SET_TARGET_PROPERTIES(${target} PROPERTIES CTR_SPEC_DESC "${WIN_CTRSDK_ROOT}/resources/specfiles/Application.desc")
	SET_TARGET_PROPERTIES(${target} PROPERTIES CTR_SPEC_RSF "${WIN_CTRSDK_ROOT}/resources/specfiles/Application.rsf")
	SET_TARGET_PROPERTIES(${target} PROPERTIES CTR_BANNER "${WIN_CTRSDK_ROOT}/resources/banner/Default.bnr")
	SET_TARGET_PROPERTIES(${target} PROPERTIES CTR_ICON "${WIN_CTRSDK_ROOT}/resources/banner/Default.icn")
	generate_ctr_makeromopt(${target})
ENDMACRO()

MACRO(target_link_libraries_default target)
	TARGET_LINK_LIBRARIES(${target}
		gles2 nn_ac nn_applet nn_boss nn_camera nn_cec
		nn_cfg nn_crypto nn_cryptoRsa nn_cx nn_dbg nn_dbm
		nn_dev nn_dlp nn_dsp nn_ec nn_err nn_erreula
		nn_fnd nn_font nn_friends nn_fs nn_fshio nn_gd
		nn_gr nn_gxlow nn_hid nn_hidlow nn_hio nn_http
		nn_init nn_ir nn_jpegmp nn_math nn_mic nn_midi
		nn_ndm nn_news nn_ngc nn_nstd nn_nwm nn_os
		nn_phtsel nn_pl nn_ptm nn_rdt nn_snd nn_socket
		nn_srv nn_ssl nn_svc nn_svcDbg nn_swkbd nn_tpl
		nn_ubl nn_uds nn_ulcd nn_util nn_voicesel
		nn_y2r rtport zlib)
ENDMACRO()

