
#
# uname -a
#
EXECUTE_PROCESS(
	COMMAND uname -a
	OUTPUT_VARIABLE NLIB_UNAME
)
MESSAGE(STATUS "NLIB_UNAME: ${NLIB_UNAME}")
IF("${NLIB_UNAME}" MATCHES "Alpine")
	SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_DEFINITIONS "NLIB_ALPINE")
	SET(ALPINE 1)
ENDIF()

IF(EXISTS "/mnt/c")
	SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_DEFINITIONS "NLIB_BASH_UBUNTU_WINDOWS")
	SET(BASH_UBUNTU_WINDOWS 1)
ENDIF()

#
# GCC or clang
#
STRING(REGEX REPLACE ".*\/(.+)\/.+$" "\\1" CAT_DIR ${CMAKE_BINARY_DIR})
IF (NOT CAT_DIR)
	MESSAGE(FATAL_ERROR "cannot set CAT_DIR")
ENDIF()

IF(CMAKE_COMPILER_IS_GNUCXX)
	EXECUTE_PROCESS(COMMAND ${CMAKE_CXX_COMPILER} -dumpversion OUTPUT_VARIABLE GCC_VERSION)
	MESSAGE(STATUS "Found g++ ${GCC_VERSION}")
	IF(GCC_VERSION VERSION_LESS 4.5)
		MESSAGE(FATAL_ERROR "!!!!! Please use g++ 4.5 or later. !!!!!")
	ENDIF()
	IF(${CMAKE_SYSTEM_NAME} MATCHES "FreeBSD")
		IF(GCC_VERSION VERSION_GREATER 5.0)
			SET(CMAKE_EXE_LINKER_FLAGS "${CMAKE_EXE_LINKER_FLAGS} -Wl,-rpath=/usr/local/lib/gcc5")
			SET(CMAKE_MODULE_LINKER_FLAGS "${CMAKE_MODULE_LINKER_FLAGS} -Wl,-rpath=/usr/local/lib/gcc5")
			SET(CMAKE_SHARED_LINKER_FLAGS "${CMAKE_SHARED_LINKER_FLAGS} -Wl,-rpath=/usr/local/lib/gcc5")
		ENDIF()
	ENDIF()
ENDIF()

#
# Compiler options
#   -ffast-math cannot be specified for SQLite
#
SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_DEFINITIONS "$<$<CONFIG:Debug>:DEBUG;_DEBUG>")
SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_DEFINITIONS "$<$<CONFIG:Release>:NDEBUG>")
SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "-Wall;-Wextra;-Wshadow;-fstack-protector")
SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "$<$<CONFIG:Release>:-O3>")

MESSAGE(STATUS "CMAKE_SYSTEM_PROCESSOR ${CMAKE_SYSTEM_PROCESSOR}")

IF(NOT NLIB_ARM)
	IF(${CMAKE_SYSTEM_NAME} MATCHES "Darwin")
		SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "-msse4.1")
	ELSE()
		SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "-msse4.2")
	ENDIF()
	IF(CMAKE_SIZEOF_VOID_P EQUAL 8)
	ELSE()
		SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "-mfpmath=sse")
	ENDIF()
ENDIF()

SET(CMAKE_EXE_LINKER_FLAGS "${CMAKE_EXE_LINKER_FLAGS} -fstack-protector")
SET(CMAKE_MODULE_LINKER_FLAGS "${CMAKE_MODULE_LINKER_FLAGS} -fstack-protector")
SET(CMAKE_SHARED_LINKER_FLAGS "${CMAKE_SHARED_LINKER_FLAGS} -fstack-protector")

# gcc specific compiler options
# SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "$<$<CXX_COMPILER_ID:GNU>:>")
# SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "$<$<C_COMPILER_ID:GNU>:>")

# clang compiler options
# SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "$<$<OR:$<CXX_COMPILER_ID:Clang>,$<CXX_COMPILER_ID:AppleClang>>:>")
# SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "$<$<OR:$<C_COMPILER_ID:Clang>,$<C_COMPILER_ID:AppleClang>>:>")

SET(CMAKE_CXX_STANDARD 14)

IF(CMAKE_COMPILER_IS_GNUCXX)
	SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "$<$<NOT:$<VERSION_LESS:$<CXX_COMPILER_VERSION>,4.7>>:-Wstack-usage=65536>")
	SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "$<$<VERSION_LESS:$<CXX_COMPILER_VERSION>,4.7>:-std=gnu++0x>")
ENDIF()

IF(DEFINED ENV{FSANITIZE})
	# export FSANITIZE=address
	# export FSANITIZE=thread
	# export FSANITIZE=memory
	# export FSANITIZE=undefined
	# export FSANITIZE=dataflow
	# export FSANITIZE=cfi
	# export FSANITIZE=safe-stack
	SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "-fsanitize=$ENV{FSANITIZE}")
	SET(CMAKE_EXE_LINKER_FLAGS "${CMAKE_EXE_LINKER_FLAGS} -fsanitize=$ENV{FSANITIZE}")
	IF($ENV{FSANITIZE} MATCHES "address")
		SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "-fno-omit-frame-pointer;-fno-optimize-sibling-calls")
	ENDIF()
	IF($ENV{FSANITIZE} MATCHES "memory")
		# ubuntu:
		# export MSAN_SYMBOLIZER_PATH=$(which llvm-symbolizer-3.4)
		SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "-fno-omit-frame-pointer;-fno-optimize-sibling-calls;-fsanitize-memory-track-origins")
	ENDIF()
	IF($ENV{FSANITIZE} MATCHES "cfi")
		SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "-flto")
	ENDIF()
ENDIF()

SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_DEFINITIONS "_GNU_SOURCE")
IF(CYGWIN)
	OPTION(BUILD_SHARED_LIBS "Build shared libraries (DLLs)." OFF)
	SET(CMAKE_DEBUG_POSTFIX "_d")
ELSEIF(UNIX)
	SET(CMAKE_DEBUG_POSTFIX "_d")
	SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "-fPIC")
	IF(${CMAKE_SYSTEM_NAME} MATCHES "Linux")
		OPTION(BUILD_SHARED_LIBS "Build shared libraries (DLLs)." ON)
	ELSEIF(${CMAKE_SYSTEM_NAME} MATCHES "FreeBSD")
		OPTION(BUILD_SHARED_LIBS "Build shared libraries (DLLs)." ON)
		IF(CMAKE_CXX_COMPILER_ID STREQUAL Clang)
			SET(CMAKE_CXX_FLAGS "${CMAKE_CXX_FLAGS} -stdlib=libc++")
		ENDIF()
	ELSEIF(${CMAKE_SYSTEM_NAME} MATCHES "Darwin")
		SET(CMAKE_CXX_FLAGS "${CMAKE_CXX_FLAGS} -stdlib=libc++")
		IF(NOT NLIB_APPLE_DIR_WORKAROUND)
			OPTION(BUILD_SHARED_LIBS "Build shared libraries (DLLs)." ON)
		ELSE()
			OPTION(BUILD_SHARED_LIBS "Build shared libraries (DLLs)." OFF)
		ENDIF()
	ELSE()
		MESSAGE(FATAL_ERROR "Unknown environment for nlib")
	ENDIF()
ENDIF()

MACRO(library_default_build_rule target)
	IF(UNIX AND NOT CYGWIN)
		IF(NOT NLIB_APPLE_DIR_WORKAROUND)
			ADD_LIBRARY(${target} SHARED ${LIB_SRC} ${LIB_HEADER})
			IF(${CMAKE_SYSTEM_NAME} MATCHES "Darwin")
				MATH(EXPR BN0 "${NLIB_BUILD_NUMBER} / 1000")
				MATH(EXPR BN1 "${NLIB_BUILD_NUMBER} % 1000")
				SET_PROPERTY(TARGET ${target} PROPERTY VERSION "${NLIB_PRODUCT_NUMBER}.${NLIB_PRODUCT_VERSION}.${BN0}.${BN1}")
			ELSEIF(NOT BASH_UBUNTU_WINDOWS)
				SET_PROPERTY(TARGET ${target} PROPERTY VERSION "${NLIB_PRODUCT_NUMBER}.${NLIB_PRODUCT_VERSION}.${NLIB_BUILD_NUMBER}")
			ENDIF()
			IF(NOT BASH_UBUNTU_WINDOWS)
				SET_PROPERTY(TARGET ${target} PROPERTY SOVERSION "${NLIB_PRODUCT_NUMBER}.${NLIB_PRODUCT_VERSION}")
			ENDIF()
			SET_PROPERTY(TARGET ${target} PROPERTY CXX_VISIBILITY_PRESET hidden)
			SET_PROPERTY(TARGET ${target} PROPERTY C_VISIBILITY_PRESET hidden)
			SET_PROPERTY(TARGET ${target} PROPERTY VISIBILITY_INLINES_HIDDEN ON)
		ELSE()
			ADD_LIBRARY(${target} STATIC ${LIB_SRC} ${LIB_HEADER})
		ENDIF()
	ELSE()
		ADD_LIBRARY(${target} STATIC ${LIB_SRC} ${LIB_HEADER})
	ENDIF()
	SET_PROPERTY(TARGET ${target} APPEND PROPERTY INTERFACE_INCLUDE_DIRECTORIES "$<BUILD_INTERFACE:${NLIB_PKGROOT_DIR}/include>")
	TARGET_INCLUDE_DIRECTORIES(${target} BEFORE PRIVATE "${NLIB_PKGROOT_DIR}/include" "${NLIB_PKGROOT_DIR}/sources")
	target_link_libraries_default(${target})
	IF(BASH_UBUNTU_WINDOWS)
		INSTALL(TARGETS ${target} LIBRARY DESTINATION lib/cmake/${CAT_DIR} NAMELINK_SKIP)
	ELSEIF(NLIB_APPLE_DIR_WORKAROUND)
		INSTALL(DIRECTORY ${CMAKE_CURRENT_BINARY_DIR}/${CMAKE_BUILD_TYPE}${CMAKE_XCODE_EFFECTIVE_PLATFORMS}/
				DESTINATION lib/cmake/${CAT_DIR})
	ELSE()
		INSTALL(TARGETS ${target} DESTINATION lib/cmake/${CAT_DIR})
	ENDIF()
	INSTALL(DIRECTORY ${NLIB_PKGROOT_DIR}/include DESTINATION .)
ENDMACRO()

MACRO(target_link_libraries_default target)
ENDMACRO()

GET_FILENAME_COMPONENT(
	NLIB_DIR
	"${NLIB_RELEASE_ROOT}/lib/cmake/${CAT_DIR}"
	ABSOLUTE)

