﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_FILESTYLEURI_H_
#define INCLUDE_NN_NLIB_FILESTYLEURI_H_

#include "nn/nlib/Config.h"
#include "nn/nlib/Uri.h"

NLIB_NAMESPACE_BEGIN

class NLIB_VIS_PUBLIC FileStyleUri NLIB_FINAL {
 public:
    FileStyleUri() NLIB_NOEXCEPT : prv_(NULL) {}
    ~FileStyleUri() NLIB_NOEXCEPT;
    bool FromUri(const Uri& uri) NLIB_NOEXCEPT;
    bool ToUri(Uri* uri) const NLIB_NOEXCEPT NLIB_NONNULL;
    bool Parse(const char* str) NLIB_NOEXCEPT NLIB_NONNULL;
    bool ComposeString(
        char* buf,
        size_t size) const NLIB_NOEXCEPT NLIB_NONNULL;
    template <size_t N>
    bool ComposeString(
        char (&buf)[N]) const NLIB_NOEXCEPT {
        return this->ComposeString(buf, N);
    }
    bool SetPath(const char* path) NLIB_NOEXCEPT;
    size_t GetNumSegment() const NLIB_NOEXCEPT;
    const char* GetSegment(size_t i) const NLIB_NOEXCEPT;
    bool AddBaseUri(const Uri& relative, const FileStyleUri& base) NLIB_NOEXCEPT;
    void Reset() NLIB_NOEXCEPT;
    bool IsPathDirectory() const NLIB_NOEXCEPT;

 private:
    NLIB_VIS_HIDDEN bool CheckInit() const NLIB_NOEXCEPT;
    struct FileStyleUriPrivate;
    mutable FileStyleUriPrivate* prv_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(FileStyleUri);
};

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_FILESTYLEURI_H_
