﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_HTTPSTYLEURI_H_
#define INCLUDE_NN_NLIB_HTTPSTYLEURI_H_

#include <utility>  // for std::pair
#include "nn/nlib/Config.h"
#include "nn/nlib/Uri.h"

NLIB_NAMESPACE_BEGIN

class NLIB_VIS_PUBLIC HttpStyleUri NLIB_FINAL {
 public:
    typedef std::pair<const char*, const char*> KeyValue;
    HttpStyleUri() NLIB_NOEXCEPT : prv_(NULL) {}
    ~HttpStyleUri() NLIB_NOEXCEPT;
    bool FromUri(const Uri& uri) NLIB_NOEXCEPT;
    bool ToUri(Uri* uri) const NLIB_NOEXCEPT;
    bool Parse(const char* str) NLIB_NOEXCEPT;
    bool ComposeString(
        char* buf,
        size_t size) const NLIB_NOEXCEPT;
    template <size_t N>
    bool ComposeString(
        char (&buf)[N]) const NLIB_NOEXCEPT {
        return this->ComposeString(buf, N);
    }
    bool SetScheme(const char* scheme) NLIB_NOEXCEPT;
    bool SetUserInfo(const char* userinfo) NLIB_NOEXCEPT;
    bool SetHost(const char* host) NLIB_NOEXCEPT;
    bool SetPort(int port) NLIB_NOEXCEPT;
    bool SetPath(const char* path) NLIB_NOEXCEPT;
    bool SetQuery(const char* query) NLIB_NOEXCEPT;
    bool SetFragment(const char* fragment) NLIB_NOEXCEPT;
    bool AddQuery(const char* key, const char* value) NLIB_NOEXCEPT;
    const char* GetScheme() const NLIB_NOEXCEPT;
    const char* GetUserInfo() const NLIB_NOEXCEPT;
    const char* GetHost() const NLIB_NOEXCEPT;
    int GetPort() const NLIB_NOEXCEPT;
    size_t GetNumSegment() const NLIB_NOEXCEPT;
    const char* GetSegment(size_t i) const NLIB_NOEXCEPT;
    bool IsPathDirectory() const NLIB_NOEXCEPT;
    const char* GetQueryValue(const char* key) const NLIB_NOEXCEPT;
    size_t GetNumQuery() const NLIB_NOEXCEPT;
    const KeyValue* GetQueryKeyValue(size_t i) const NLIB_NOEXCEPT;
    const char* GetFragment() const NLIB_NOEXCEPT;
    bool AddBaseUri(const Uri& relative, const HttpStyleUri& base) NLIB_NOEXCEPT;
    void Reset() NLIB_NOEXCEPT;

 private:
    NLIB_VIS_HIDDEN bool CheckInit() const NLIB_NOEXCEPT;
    struct HttpStyleUriPrivate;
    mutable HttpStyleUriPrivate* prv_;
};

NLIB_NAMESPACE_END
#endif  // INCLUDE_NN_NLIB_HTTPSTYLEURI_H_
