﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_PLATFORM_WIN32_H_
#define INCLUDE_NN_NLIB_PLATFORM_WIN32_H_
#ifndef INCLUDE_NN_NLIB_PLATFORM_H_
# error do not include directly
#endif

#if defined(_MSC_VER)

#if !defined(_M_IX86) && !defined(_M_X64)
#error Sorry, not supported
#endif

#if _MSC_VER < 1700
# error Use Visual Studio 2012 or later
#endif

#ifndef WIN32_LEAN_AND_MEAN
# pragma message(__FILE__ ": WIN32_LEAN_AND_MEAN not defined, compile may fail")
# define WIN32_LEAN_AND_MEAN
#endif

#ifndef NOMINMAX
# pragma message(__FILE__ ": NOMINMAX not defined, compile may fail")
# define NOMINMAX
#endif

#ifndef _USE_MATH_DEFINES
# pragma message(__FILE__ ": _USE_MATH_DEFINES not defined, compile may fail")
# define _USE_MATH_DEFINES
#endif

#ifdef _WIN64
# define NLIB_64BIT
#endif

#include <intrin.h>

#pragma intrinsic(_BitScanReverse)
#pragma intrinsic(_BitScanForward)
#ifdef NLIB_64BIT
# pragma intrinsic(_BitScanReverse64)
# pragma intrinsic(_BitScanForward64)
#endif

#include <stdlib.h>  // for malloc, free, calloc, realloc
#include <limits.h>  // for RSIZE_MAX
#include <WinSock2.h>
#include <Windows.h>  // for HANDLE, CONDITION_VARIABLE, ....
#include <fcntl.h>  // for O_RDONLY, O_RDWR, O_CREAT, ....

#include <in6addr.h>
#include <ws2ipdef.h>
#include <sal.h>

#if _MSC_VER == 1800 && WINVER < 0x0603
# pragma message(__FILE__ ": Please update your Windows SDK")
#endif

#if _MSC_VER == 1700 && WINVER < 0x0602
# pragma message(__FILE__ ": Please update your Windows SDK")
#endif

#if (!defined(_M_IX86_FP) || _M_IX86_FP < 2) && !defined(_M_X64) && !defined(_M_AMD64)
# error /arch:SSE2 or higher must be set
#endif

#ifndef NLIB_SSE41
# define NLIB_SSE41
#endif

#ifndef NLIB_SSE42
# define NLIB_SSE42
#endif

#define NLIB_HAS_STDHEADER_STDINT
#include <stdint.h>
#if (_MSC_VER >= 1800)
# define NLIB_HAS_STDHEADER_INTTYPES
# include <inttypes.h>
#endif
#if (_MSC_VER < 1900)
# ifndef __func__
#  define __func__ __FUNCTION__
# endif
#endif

#ifndef NLIB_HAS_STDHEADER_INTTYPES
# include "nn/nlib/IntTypes.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifndef va_copy
# define va_copy(dest, src) (dest = src)
#endif
#define NLIB_ALWAYS_INLINE __forceinline
#define NLIB_NEVER_INLINE __declspec(noinline)
#define NLIB_LIKELY(x) (x)
#define NLIB_UNLIKELY(x) (x)
#define NLIB_EXPECT(var, exp_value) (var)
#define NLIB_CHECK_RESULT
#define NLIB_NORETURN
#define NLIB_NONNULL
#define NLIB_NONNULL_1
#define NLIB_NONNULL_2
#define NLIB_NONNULL_3
#define NLIB_NONNULL_4
#define NLIB_NONNULL_5
#define NLIB_ATTRIBUTE_MALLOC
#define NLIB_ATTRIBUTE_PURE
#define NLIB_ATTRIBUTE_CONST
#define NLIB_ATTRIBUTE_ALLOC_SIZE1(n)
#define NLIB_ATTRIBUTE_ALLOC_SIZE2(n0, n1)
#define NLIB_ATTRIBUTE_ALLOC_ALIGN(algn)
#define NLIB_ATTRIBUTE_ASSUME_ALIGNED(n)
#ifndef NLIB_DEPRECATED
#define NLIB_DEPRECATED __declspec(deprecated)
#endif
#ifndef NLIB_DEPRECATED_MSG
#define NLIB_DEPRECATED_MSG(msg) __declspec(deprecated(msg))
#endif
#define NLIB_WEAKSYMBOL
#define NLIB_VIS_HIDDEN
#ifdef NLIB_WINDLL
# define NLIB_EXPIMP_TEMPLATE_(x) template class NLIB_VIS_PUBLIC x
# define NLIB_WINEXPORT __declspec(dllexport)
# define NLIB_WINIMPORT __declspec(dllimport)
# define NLIB_VIS_PUBLIC NLIB_WINIMPORT
# define NLIB_EXPIMP_TEMPLATE(x) extern NLIB_EXPIMP_TEMPLATE_(x)
#else
# define NLIB_VIS_PUBLIC
# define NLIB_EXPIMP_TEMPLATE(x)
#endif

#if defined(n_EXPORTS)
# undef NLIB_VIS_PUBLIC
# define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#elif defined(nx_misc_EXPORTS)
# undef NLIB_EXPIMP_TEMPLATE
# define NLIB_EXPIMP_TEMPLATE(x) NLIB_EXPIMP_TEMPLATE_(x)
#endif

#define NLIB_VIS_PUBLIC_ALT

#define NLIB_LITTLE_ENDIAN

// NOTE:
//   _WriteBarrier() for the compiler, _mm_sfence() for the processor, ...
// NOTE:
//   You have to comment out error lines in intrin.h if you use VC++2005.
//   See:
//     https://connect.microsoft.com/VisualStudio/feedback/ViewFeedback.aspx?FeedbackID=262047
//     http://www.hydrogenaudio.org/forums/index.php?showtopic=50972
#define NLIB_MEMORY_ORDER_RELEASE                   \
    __pragma(warning(push))                         \
    __pragma(warning(disable:4127))                 \
    do { _WriteBarrier(); _mm_sfence(); } while (0) \
    __pragma(warning(pop))
#define NLIB_MEMORY_ORDER_ACQUIRE                  \
    __pragma(warning(push))                        \
    __pragma(warning(disable:4127))                \
    do { _ReadBarrier(); _mm_lfence(); } while (0) \
    __pragma(warning(pop))
#define NLIB_MEMORY_ORDER_ACQ_REL                       \
    __pragma(warning(push))                             \
    __pragma(warning(disable:4127))                     \
    do { _ReadWriteBarrier(); _mm_mfence(); } while (0) \
    __pragma(warning(pop))
#define NLIB_MEMORY_ORDER_SEQ_CST NLIB_MEMORY_ORDER_ACQ_REL

// because "%zu" not supported ...
#ifdef NLIB_64BIT
# define __PRIS_PREFIX "ll"
#else
# define __PRIS_PREFIX
#endif
#define NLIB_WARN__(x) #x
#define NLIB_WARN_(x) NLIB_WARN__(x)
#define NLIB_WARN(exp) (__FILE__ "(" NLIB_WARN_(__LINE__) ") : WARNING: " exp)

typedef unsigned int nlib_tls;

#if _MSC_VER >= 1900
#define NLIB_TIMESPEC_HAS_NATIVE
#endif

#pragma pack(push, 8)
typedef struct nlib_mutex_ {
    int32_t mark;
    uint32_t cat;
    union {
        HANDLE h;
        CRITICAL_SECTION cs;
    } data;
} nlib_mutex;
#pragma pack(pop)
#define NLIB_MUTEX_INITIALIZER { 0x6768696AU, 0 }
#define NLIB_RECURSIVE_MUTEX_INITIALIZER { 0x6768696AU, 1 }
#define NLIB_RECURSIVE_TIMED_MUTEX_INITIALIZER { 0x6768696AU, 2 }

typedef HANDLE nlib_semaphore;

typedef struct nlib_cond_ {
    CONDITION_VARIABLE cond;
    nlib_mutex mutex;
} nlib_cond;
#define NLIB_COND_INITIALIZER { 0, NLIB_MUTEX_INITIALIZER }

typedef HANDLE nlib_thread;

#define NLIB_ATTRIBUTE_PRINTF(x, y)

#define NLIB_ASSUME(cond) __assume(cond)

#define NLIB_RWLOCK_HAS_NATIVE
#ifdef NLIB_RWLOCK_HAS_NATIVE
typedef SRWLOCK nlib_rwlock;
#define NLIB_RWLOCK_INITIALIZER SRWLOCK_INIT
#endif

NLIB_VIS_PUBLIC void nlib_free_size_(void* ptr, size_t size);

#define NLIB_LIBC_nlib_memcmp
#define NLIB_LIBC_nlib_strlen
#define NLIB_LIBC_nlib_strnlen
#define NLIB_LIBC_nlib_wcslen
#define NLIB_LIBC_nlib_wcsnlen
#define NLIB_LIBC_nlib_strcpy
#define NLIB_LIBC_nlib_strncpy
#define NLIB_LIBC_nlib_wcscpy
#define NLIB_LIBC_nlib_wcsncpy
#define NLIB_LIBC_nlib_strchr
#define NLIB_LIBC_nlib_strrchr

#if !defined(__INTELLISENSE__)
#define NLIB_ATOMIC_RELAXED (0)
#define NLIB_ATOMIC_ACQUIRE (1)
#define NLIB_ATOMIC_RELEASE (2)
#define NLIB_ATOMIC_ACQ_REL (3)
#define NLIB_ATOMIC_SEQ_CST (7)

//
// NOTE:
// Windows 8 has xxxNoFence version
//

static __inline int32_t nlib_atomic_load32(const int32_t* ptr, int memorder) {
    int32_t result;
    _ReadWriteBarrier();
    result = *ptr;
    (void)memorder;
    // if (memorder & NLIB_ATOMIC_ACQUIRE)
    //    NLIB_MEMORY_ORDER_ACQUIRE;
    _ReadWriteBarrier();
    return result;
}

static __inline void nlib_atomic_store32(int32_t* ptr, int32_t val, int memorder) {
    _ReadWriteBarrier();
    if (memorder != NLIB_ATOMIC_SEQ_CST) {
        // if (memorder & NLIB_ATOMIC_RELEASE)
        //    NLIB_MEMORY_ORDER_RELEASE;
        *ptr = val;
    } else {
        _InterlockedExchange((volatile long*)ptr, val);  // NOLINT
    }
    _ReadWriteBarrier();
}

static __inline int32_t nlib_atomic_exchange32(int32_t* ptr, int32_t val,
                                               int memorder) {
    int32_t result;
    (void)memorder;
    _ReadWriteBarrier();
    result = (int32_t)_InterlockedExchange((volatile long*)ptr, val);  // NOLINT
    _ReadWriteBarrier();
    return result;
}

static __inline int nlib_atomic_compare_exchange32(int32_t* ptr, int32_t* expected,
                                                   int32_t desired, int weak,
                                                   int success_memorder, int failure_memorder) {
    int32_t result;
    _ReadWriteBarrier();
    result = _InterlockedCompareExchange((volatile long*)ptr, desired, *expected);  // NOLINT
    _ReadWriteBarrier();
    (void)weak;
    (void)success_memorder;
    (void)failure_memorder;
    if (result == *expected) {
        return 1;
    } else {
        *expected = result;
        return 0;
    }
}

static __inline int32_t nlib_atomic_add_fetch32(int32_t* ptr, int32_t val,
                                                int memorder) {
    int32_t result;
    (void)memorder;
    _ReadWriteBarrier();
    result = _InterlockedExchangeAdd((volatile long*)ptr, val) + val;  // NOLINT
    _ReadWriteBarrier();
    return result;
}

static __inline int32_t nlib_atomic_sub_fetch32(int32_t* ptr, int32_t val,
                                                int memorder) {
    int32_t result;
    (void)memorder;
    _ReadWriteBarrier();
    result = _InterlockedExchangeAdd((volatile long*)ptr, -val) - val;  // NOLINT
    _ReadWriteBarrier();
    return result;
}

static __inline int32_t nlib_atomic_and_fetch32(int32_t* ptr, int32_t val,
                                                int memorder) {
    int32_t result;
    (void)memorder;
    _ReadWriteBarrier();
    result = _InterlockedAnd((volatile long*)ptr, val) & val;  // NOLINT
    _ReadWriteBarrier();
    return result;
}

static __inline int32_t nlib_atomic_xor_fetch32(int32_t* ptr, int32_t val,
                                                int memorder) {
    int32_t result;
    (void)memorder;
    _ReadWriteBarrier();
    result = _InterlockedXor((volatile long*)ptr, val) ^ val;  // NOLINT
    _ReadWriteBarrier();
    return result;
}

static __inline int32_t nlib_atomic_or_fetch32(int32_t* ptr, int32_t val,
                                               int memorder) {
    int32_t result;
    (void)memorder;
    _ReadWriteBarrier();
    result = _InterlockedOr((volatile long*)ptr, val) | val;  // NOLINT
    _ReadWriteBarrier();
    return result;
}

static __inline int32_t nlib_atomic_fetch_add32(int32_t* ptr, int32_t val,
                                                int memorder) {
    int32_t result;
    (void)memorder;
    _ReadWriteBarrier();
    result = _InterlockedExchangeAdd((volatile long*)ptr, val);  // NOLINT
    _ReadWriteBarrier();
    return result;
}

static __inline int32_t nlib_atomic_fetch_sub32(int32_t* ptr, int32_t val,
                                                int memorder) {
    int32_t result;
    (void)memorder;
    _ReadWriteBarrier();
    result = _InterlockedExchangeAdd((volatile long*)ptr, -val);  // NOLINT
    _ReadWriteBarrier();
    return result;
}

static __inline int32_t nlib_atomic_fetch_and32(int32_t* ptr, int32_t val,
                                                int memorder) {
    int32_t result;
    (void)memorder;
    _ReadWriteBarrier();
    result = _InterlockedAnd((volatile long*)ptr, val);  // NOLINT
    _ReadWriteBarrier();
    return result;
}

static __inline int32_t nlib_atomic_fetch_xor32(int32_t* ptr, int32_t val,
                                                int memorder) {
    int32_t result;
    (void)memorder;
    _ReadWriteBarrier();
    result = _InterlockedXor((volatile long*)ptr, val);  // NOLINT
    _ReadWriteBarrier();
    return result;
}

static __inline int32_t nlib_atomic_fetch_or32(int32_t* ptr, int32_t val,
                                               int memorder) {
    int32_t result;
    (void)memorder;
    _ReadWriteBarrier();
    result = _InterlockedOr((volatile long*)ptr, val);  // NOLINT
    _ReadWriteBarrier();
    return result;
}

static __inline int64_t nlib_atomic_load64(const int64_t* ptr, int memorder) {
    int64_t result;
    _ReadWriteBarrier();
    result = *ptr;
    _ReadWriteBarrier();
    (void)memorder;
    // if (memorder & NLIB_ATOMIC_ACQUIRE)
    //    NLIB_MEMORY_ORDER_ACQUIRE;
    return result;
}

static __inline void nlib_atomic_store64(int64_t* ptr, int64_t val, int memorder) {
    _ReadWriteBarrier();
    if (memorder != NLIB_ATOMIC_SEQ_CST) {
        // if (memorder & NLIB_ATOMIC_RELEASE)
        //    NLIB_MEMORY_ORDER_RELEASE;
        *ptr = val;
    } else {
#ifndef NLIB_64BIT
        InterlockedExchange64((volatile long long*)ptr, val);  // NOLINT
#else
        _InterlockedExchange64((volatile long long*)ptr, val);  // NOLINT
#endif
    }
    _ReadWriteBarrier();
}

static __inline int64_t nlib_atomic_exchange64(int64_t* ptr, int64_t val,
                                               int memorder) {
    int64_t result;
    (void)memorder;
    _ReadWriteBarrier();
#ifndef NLIB_64BIT
    result = (int64_t)InterlockedExchange64((volatile long long*)ptr, val);  // NOLINT
#else
    result = (int64_t)_InterlockedExchange64((volatile long long*)ptr, val);  // NOLINT
#endif
    _ReadWriteBarrier();
    return result;
}

static __inline int nlib_atomic_compare_exchange64(int64_t* ptr, int64_t* expected,
                                                   int64_t desired, int weak,
                                                   int success_memorder, int failure_memorder) {
    int64_t result;
    _ReadWriteBarrier();
#ifndef NLIB_64BIT
    result = InterlockedCompareExchange64((volatile long long*)ptr, desired, *expected);  // NOLINT
#else
    result = _InterlockedCompareExchange64((volatile long long*)ptr, desired, *expected);  // NOLINT
#endif
    _ReadWriteBarrier();
    (void)weak;
    (void)success_memorder;
    (void)failure_memorder;
    if (result == *expected) {
        return 1;
    } else {
        *expected = result;
        return 0;
    }
}

static __inline int64_t nlib_atomic_add_fetch64(int64_t* ptr, int64_t val,
                                                int memorder) {
    int64_t result;
    (void)memorder;
    _ReadWriteBarrier();
#ifndef NLIB_64BIT
    result = InterlockedExchangeAdd64((volatile long long*)ptr, val) + val;  // NOLINT
#else
    result = _InterlockedExchangeAdd64((volatile long long*)ptr, val) + val;  // NOLINT
#endif
    _ReadWriteBarrier();
    return result;
}

static __inline int64_t nlib_atomic_sub_fetch64(int64_t* ptr, int64_t val,
                                                int memorder) {
    int64_t result;
    (void)memorder;
    _ReadWriteBarrier();
#ifndef NLIB_64BIT
    result = InterlockedExchangeAdd64((volatile long long*)ptr, -val) - val;  // NOLINT
#else
    result = _InterlockedExchangeAdd64((volatile long long*)ptr, -val) - val;  // NOLINT
#endif
    _ReadWriteBarrier();
    return result;
}

static __inline int64_t nlib_atomic_and_fetch64(int64_t* ptr, int64_t val,
                                                int memorder) {
    int64_t result;
    (void)memorder;
    _ReadWriteBarrier();
#ifndef NLIB_64BIT
    result = InterlockedAnd64((volatile long long*)ptr, val) & val;  // NOLINT
#else
    result = _InterlockedAnd64((volatile long long*)ptr, val) & val;  // NOLINT
#endif
    _ReadWriteBarrier();
    return result;
}

static __inline int64_t nlib_atomic_xor_fetch64(int64_t* ptr, int64_t val,
                                                int memorder) {
    int64_t result;
    (void)memorder;
    _ReadWriteBarrier();
#ifndef NLIB_64BIT
    result = InterlockedXor64((volatile long long*)ptr, val) ^ val;  // NOLINT
#else
    result = _InterlockedXor64((volatile long long*)ptr, val) ^ val;  // NOLINT
#endif
    _ReadWriteBarrier();
    return result;
}

static __inline int64_t nlib_atomic_or_fetch64(int64_t* ptr, int64_t val,
                                               int memorder) {
    int64_t result;
    (void)memorder;
    _ReadWriteBarrier();
#ifndef NLIB_64BIT
    result = InterlockedOr64((volatile long long*)ptr, val) | val;  // NOLINT
#else
    result = _InterlockedOr64((volatile long long*)ptr, val) | val;  // NOLINT
#endif
    _ReadWriteBarrier();
    return result;
}

static __inline int64_t nlib_atomic_fetch_add64(int64_t* ptr, int64_t val,
                                                int memorder) {
    int64_t result;
    (void)memorder;
    _ReadWriteBarrier();
#ifndef NLIB_64BIT
    result = InterlockedExchangeAdd64((volatile long long*)ptr, val);  // NOLINT
#else
    result = _InterlockedExchangeAdd64((volatile long long*)ptr, val);  // NOLINT
#endif
    _ReadWriteBarrier();
    return result;
}

static __inline int64_t nlib_atomic_fetch_sub64(int64_t* ptr, int64_t val,
                                                int memorder) {
    int64_t result;
    (void)memorder;
    _ReadWriteBarrier();
#ifndef NLIB_64BIT
    result = InterlockedExchangeAdd64((volatile long long*)ptr, -val);  // NOLINT
#else
    result = _InterlockedExchangeAdd64((volatile long long*)ptr, -val);  // NOLINT
#endif
    _ReadWriteBarrier();
    return result;
}

static __inline int64_t nlib_atomic_fetch_and64(int64_t* ptr, int64_t val,
                                                int memorder) {
    int64_t result;
    (void)memorder;
    _ReadWriteBarrier();
#ifndef NLIB_64BIT
    result = InterlockedAnd64((volatile long long*)ptr, val);  // NOLINT
#else
    result = _InterlockedAnd64((volatile long long*)ptr, val);  // NOLINT
#endif
    _ReadWriteBarrier();
    return result;
}

static __inline int64_t nlib_atomic_fetch_xor64(int64_t* ptr, int64_t val,
                                                int memorder) {
    int64_t result;
    (void)memorder;
    _ReadWriteBarrier();
#ifndef NLIB_64BIT
    result = InterlockedXor64((volatile long long*)ptr, val);  // NOLINT
#else
    result = _InterlockedXor64((volatile long long*)ptr, val);  // NOLINT
#endif
    _ReadWriteBarrier();
    return result;
}

static __inline int64_t nlib_atomic_fetch_or64(int64_t* ptr, int64_t val,
                                               int memorder) {
    int64_t result;
    (void)memorder;
    _ReadWriteBarrier();
#ifndef NLIB_64BIT
    result = InterlockedOr64((volatile long long*)ptr, val);  // NOLINT
#else
    result = _InterlockedOr64((volatile long long*)ptr, val);  // NOLINT
#endif
    _ReadWriteBarrier();
    return result;
}

static __inline void* nlib_atomic_loadptr(void* const* ptr, int memorder) {
    void* result;
    _ReadWriteBarrier();
    result = *ptr;
    _ReadWriteBarrier();
    (void)memorder;
    // if (memorder & NLIB_ATOMIC_ACQUIRE)
    //    NLIB_MEMORY_ORDER_ACQUIRE;
    return result;
}

static __inline void nlib_atomic_storeptr(void** ptr, void* val, int memorder) {
    _ReadWriteBarrier();
    if (memorder != NLIB_ATOMIC_SEQ_CST) {
        // if (memorder & NLIB_ATOMIC_RELEASE)
        //    NLIB_MEMORY_ORDER_RELEASE;
        *ptr = val;
    } else {
#if _MSC_VER < 1800 || (_MSC_VER == 1800 && !defined(NLIB_64BIT))
        InterlockedExchangePointer((void* volatile*)ptr, val);  // NOLINT
#else
        _InterlockedExchangePointer((void* volatile*)ptr, val);  // NOLINT
#endif
    }
    _ReadWriteBarrier();
}

static __inline int nlib_atomic_compare_exchangeptr(void** ptr, void** expected,
                                                    void* desired, int weak,
                                                    int success_memorder, int failure_memorder) {
    void* result;
    _ReadWriteBarrier();
#if _MSC_VER < 1800
    result = InterlockedCompareExchangePointer((void* volatile *)ptr, desired, *expected);
#else
    result = _InterlockedCompareExchangePointer((void* volatile *)ptr, desired, *expected);
#endif
    _ReadWriteBarrier();
    (void)weak;
    (void)success_memorder;
    (void)failure_memorder;
    if (result == *expected) {
        return 1;
    } else {
        *expected = result;
        return 0;
    }
}

static __inline void nlib_atomic_thread_fence(int memorder) {
    switch (memorder) {
    case NLIB_ATOMIC_RELAXED:
        break;
    case NLIB_ATOMIC_ACQUIRE:
        NLIB_MEMORY_ORDER_ACQUIRE;
        break;
    case NLIB_ATOMIC_RELEASE:
        NLIB_MEMORY_ORDER_RELEASE;
        break;
    default:
        NLIB_MEMORY_ORDER_SEQ_CST;
        break;
    }
}
#endif

#ifdef __cplusplus
}
#endif

#ifndef _WIN64
# pragma comment(linker, "/alternatename:_nlib_malloc=_malloc")
# pragma comment(linker, "/alternatename:_nlib_free=_free")
# pragma comment(linker, "/alternatename:_nlib_calloc=_calloc")
# pragma comment(linker, "/alternatename:_nlib_realloc=_realloc")
# pragma comment(linker, "/alternatename:_nlib_free_size=_nlib_free_size_")
#else
# pragma comment(linker, "/alternatename:nlib_malloc=malloc")
# pragma comment(linker, "/alternatename:nlib_free=free")
# pragma comment(linker, "/alternatename:nlib_calloc=calloc")
# pragma comment(linker, "/alternatename:nlib_realloc=realloc")
# pragma comment(linker, "/alternatename:nlib_free_size=nlib_free_size_")
#endif

#if defined(NLIB_SIMD)
# ifndef _WIN64
# ifndef NLIB_LIBC_nlib_strlen
#  pragma comment(linker, "/alternatename:_nlib_strlen=_nlib_strlen_simd")
# endif
# ifndef NLIB_LIBC_nlib_strlen
#  pragma comment(linker, "/alternatename:_nlib_strnlen=_nlib_strnlen_simd")
# endif
#  pragma comment(linker, "/alternatename:_nlib_utf16len_=_nlib_utf16len_simd")
#  pragma comment(linker, "/alternatename:_nlib_utf16nlen_=_nlib_utf16nlen_simd")
#  pragma comment(linker, "/alternatename:_nlib_utf32len_=_nlib_utf32len_generic")
#  pragma comment(linker, "/alternatename:_nlib_utf32nlen_=_nlib_utf32nlen_generic")
#  pragma comment(linker, "/alternatename:_nlib_utf16cplen_ex_=_nlib_utf16cplen_ex_simd")
#  pragma comment(linker, "/alternatename:_nlib_strchr=_nlib_strchr_simd")
#  pragma comment(linker, "/alternatename:_nlib_strrchr=_strrchr")
#  pragma comment(linker, "/alternatename:_nlib_strchr_mb=_nlib_strchr_mb_simd")
#  pragma comment(linker, "/alternatename:_nlib_memcmp=_nlib_memcmp_simd")
#  pragma comment(linker, "/alternatename:_nlib_memchr=_nlib_memchr_simd")
#  pragma comment(linker, "/alternatename:_nlib_memrchr=_nlib_memrchr_simd")
#  pragma comment(linker, "/alternatename:_nlib_memchr_not=_nlib_memchr_not_simd")
#  pragma comment(linker, "/alternatename:_nlib_memchr_range_not=_nlib_memchr_range_not_simd")
#  pragma comment(linker, "/alternatename:_nlib_memchr_lt=_nlib_memchr_lt_simd")
#  pragma comment(linker, "/alternatename:_nlib_memchr_gt=_nlib_memchr_gt_simd")
#  pragma comment(linker, "/alternatename:_nlib_memchr_mb=_nlib_memchr_mb_simd")
#  pragma comment(linker, "/alternatename:_nlib_skipws=_nlib_skipws_simd")
#  pragma comment(linker, "/alternatename:_nlib_swapendian_16=_nlib_swapendian_16_simd")
#  pragma comment(linker, "/alternatename:_nlib_swapendian_32=_nlib_swapendian_32_simd")
#  pragma comment(linker, "/alternatename:_nlib_swapendian_64=_nlib_swapendian_64_simd")
#  pragma comment(linker, "/alternatename:_nlib_strcplen_ex=_nlib_strcplen_ex")
#  pragma comment(linker, "/alternatename:_nlib_popcnt=_nlib_popcnt_simd")
#  pragma comment(linker, "/alternatename:_nlib_popcnt64=_nlib_popcnt64_simd")
# else
# ifndef NLIB_LIBC_nlib_strlen
#  pragma comment(linker, "/alternatename:nlib_strlen=nlib_strlen_simd")
# endif
# ifndef NLIB_LIBC_nlib_strlen
#  pragma comment(linker, "/alternatename:nlib_strnlen=nlib_strnlen_simd")
# endif
#  pragma comment(linker, "/alternatename:nlib_utf16len_=nlib_utf16len_simd")
#  pragma comment(linker, "/alternatename:nlib_utf16nlen_=nlib_utf16nlen_simd")
#  pragma comment(linker, "/alternatename:nlib_utf32len_=nlib_utf32len_generic")
#  pragma comment(linker, "/alternatename:nlib_utf32nlen_=nlib_utf32nlen_generic")
#  pragma comment(linker, "/alternatename:nlib_utf16cplen_ex_=nlib_utf16cplen_ex_simd")
#  pragma comment(linker, "/alternatename:nlib_strchr=nlib_strchr_simd")
#  pragma comment(linker, "/alternatename:nlib_strrchr=strrchr")
#  pragma comment(linker, "/alternatename:nlib_strchr_mb=nlib_strchr_mb_simd")
#  pragma comment(linker, "/alternatename:nlib_memcmp=nlib_memcmp_simd")
#  pragma comment(linker, "/alternatename:nlib_memchr=nlib_memchr_simd")
#  pragma comment(linker, "/alternatename:nlib_memrchr=nlib_memrchr_simd")
#  pragma comment(linker, "/alternatename:nlib_memchr_not=nlib_memchr_not_simd")
#  pragma comment(linker, "/alternatename:nlib_memchr_range_not=nlib_memchr_range_not_simd")
#  pragma comment(linker, "/alternatename:nlib_memchr_lt=nlib_memchr_lt_simd")
#  pragma comment(linker, "/alternatename:nlib_memchr_gt=nlib_memchr_gt_simd")
#  pragma comment(linker, "/alternatename:nlib_memchr_mb=nlib_memchr_mb_simd")
#  pragma comment(linker, "/alternatename:nlib_skipws=nlib_skipws_simd")
#  pragma comment(linker, "/alternatename:nlib_swapendian_16=nlib_swapendian_16_simd")
#  pragma comment(linker, "/alternatename:nlib_swapendian_32=nlib_swapendian_32_simd")
#  pragma comment(linker, "/alternatename:nlib_swapendian_64=nlib_swapendian_64_simd")
#  pragma comment(linker, "/alternatename:nlib_strcplen_ex=nlib_strcplen_ex")
#  pragma comment(linker, "/alternatename:nlib_popcnt=nlib_popcnt_simd")
#  pragma comment(linker, "/alternatename:nlib_popcnt64=nlib_popcnt64_simd")
# endif
#else
# ifndef _WIN64
# ifndef NLIB_LIBC_nlib_strlen
#  pragma comment(linker, "/alternatename:_nlib_strlen=_nlib_strlen_generic")
# endif
# ifndef NLIB_LIBC_nlib_strlen
#  pragma comment(linker, "/alternatename:_nlib_strnlen=_nlib_strnlen_generic")
# endif
#  pragma comment(linker, "/alternatename:_nlib_utf16len_=_nlib_utf16len_generic")
#  pragma comment(linker, "/alternatename:_nlib_utf16nlen_=_nlib_utf16nlen_generic")
#  pragma comment(linker, "/alternatename:_nlib_utf32len_=_nlib_utf32len_generic")
#  pragma comment(linker, "/alternatename:_nlib_utf32nlen_=_nlib_utf32nlen_generic")
#  pragma comment(linker, "/alternatename:_nlib_utf16cplen_ex_=_nlib_utf16cplen_ex_generic")
#  pragma comment(linker, "/alternatename:_nlib_strchr=_nlib_strchr_generic")
#  pragma comment(linker, "/alternatename:_nlib_strrchr=_nlib_strrchr_generic")
#  pragma comment(linker, "/alternatename:_nlib_strchr_mb=_nlib_strchr_mb_generic")
#  pragma comment(linker, "/alternatename:_nlib_memcmp=_nlib_memcmp_generic")
#  pragma comment(linker, "/alternatename:_nlib_memchr=_nlib_memchr_generic")
#  pragma comment(linker, "/alternatename:_nlib_memrchr=_nlib_memrchr_generic")
#  pragma comment(linker, "/alternatename:_nlib_memchr_not=_nlib_memchr_not_generic")
#  pragma comment(linker, "/alternatename:_nlib_memchr_range_not=_nlib_memchr_range_not_generic")
#  pragma comment(linker, "/alternatename:_nlib_memchr_lt=_nlib_memchr_lt_generic")
#  pragma comment(linker, "/alternatename:_nlib_memchr_gt=_nlib_memchr_gt_generic")
#  pragma comment(linker, "/alternatename:_nlib_memchr_mb=_nlib_memchr_mb_generic")
#  pragma comment(linker, "/alternatename:_nlib_skipws=_nlib_skipws_generic")
#  pragma comment(linker, "/alternatename:_nlib_swapendian_16=_nlib_swapendian_16_generic")
#  pragma comment(linker, "/alternatename:_nlib_swapendian_32=_nlib_swapendian_32_generic")
#  pragma comment(linker, "/alternatename:_nlib_swapendian_64=_nlib_swapendian_64_generic")
#  pragma comment(linker, "/alternatename:_nlib_strcplen_ex=_nlib_strcplen_ex")
#  pragma comment(linker, "/alternatename:_nlib_popcnt=_nlib_popcnt_generic")
#  pragma comment(linker, "/alternatename:_nlib_popcnt64=_nlib_popcnt64_generic")
# else
# ifndef NLIB_LIBC_nlib_strlen
#  pragma comment(linker, "/alternatename:nlib_strlen=nlib_strlen_generic")
# endif
# ifndef NLIB_LIBC_nlib_strlen
#  pragma comment(linker, "/alternatename:nlib_strnlen=nlib_strnlen_generic")
# endif
#  pragma comment(linker, "/alternatename:nlib_utf16len_=nlib_utf16len_generic")
#  pragma comment(linker, "/alternatename:nlib_utf16nlen_=nlib_utf16nlen_generic")
#  pragma comment(linker, "/alternatename:nlib_utf32len_=nlib_utf32len_generic")
#  pragma comment(linker, "/alternatename:nlib_utf32nlen_=nlib_utf32nlen_generic")
#  pragma comment(linker, "/alternatename:nlib_utf16cplen_ex_=nlib_utf16cplen_ex_generic")
#  pragma comment(linker, "/alternatename:nlib_strchr=nlib_strchr_generic")
#  pragma comment(linker, "/alternatename:nlib_strrchr=nlib_strrchr_generic")
#  pragma comment(linker, "/alternatename:nlib_strchr_mb=nlib_strchr_mb_generic")
#  pragma comment(linker, "/alternatename:nlib_memcmp=nlib_memcmp_generic")
#  pragma comment(linker, "/alternatename:nlib_memchr=nlib_memchr_generic")
#  pragma comment(linker, "/alternatename:nlib_memrchr=nlib_memrchr_generic")
#  pragma comment(linker, "/alternatename:nlib_memchr_not=nlib_memchr_not_generic")
#  pragma comment(linker, "/alternatename:nlib_memchr_range_not=nlib_memchr_range_not_generic")
#  pragma comment(linker, "/alternatename:nlib_memchr_lt=nlib_memchr_lt_generic")
#  pragma comment(linker, "/alternatename:nlib_memchr_gt=nlib_memchr_gt_generic")
#  pragma comment(linker, "/alternatename:nlib_memchr_mb=nlib_memchr_mb_generic")
#  pragma comment(linker, "/alternatename:nlib_skipws=nlib_skipws_generic")
#  pragma comment(linker, "/alternatename:nlib_swapendian_16=nlib_swapendian_16_generic")
#  pragma comment(linker, "/alternatename:nlib_swapendian_32=nlib_swapendian_32_generic")
#  pragma comment(linker, "/alternatename:nlib_swapendian_64=nlib_swapendian_64_generic")
#  pragma comment(linker, "/alternatename:nlib_strcplen_ex=nlib_strcplen_ex")
#  pragma comment(linker, "/alternatename:nlib_popcnt=nlib_popcnt_generic")
#  pragma comment(linker, "/alternatename:nlib_popcnt64=nlib_popcnt64_generic")
# endif
#endif

#endif
#endif  // INCLUDE_NN_NLIB_PLATFORM_WIN32_H_
