﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_STDINT_H_
#define INCLUDE_NN_NLIB_STDINT_H_

typedef signed char int8_t;
typedef short int16_t;  // NOLINT
typedef int int32_t;
typedef long long int64_t;  // NOLINT
typedef unsigned char uint8_t;
typedef unsigned short uint16_t;  // NOLINT
typedef unsigned int uint32_t;
typedef unsigned long long uint64_t;  // NOLINT

typedef signed char int_least8_t;
typedef short int_least16_t;  // NOLINT
typedef int int_least32_t;
typedef long long int_least64_t;  // NOLINT
typedef unsigned char uint_least8_t;
typedef unsigned short uint_least16_t;  // NOLINT
typedef unsigned int uint_least32_t;
typedef unsigned long long uint_least64_t;  // NOLINT

typedef signed char int_fast8_t;
typedef short int_fast16_t;  // NOLINT
typedef int int_fast32_t;
typedef long long int_fast64_t;  // NOLINT
typedef unsigned char uint_fast8_t;
typedef unsigned short uint_fast16_t;  // NOLINT
typedef unsigned int uint_fast32_t;
typedef unsigned long long uint_fast64_t;  // NOLINT

// NOTE:
// boost/cstdint.hpp may have defined them.
#ifndef INT8_MIN
#define INT8_MIN (-127 - 1)
#endif
#ifndef INT16_MIN
#define INT16_MIN (-32767 - 1)
#endif
#ifndef INT32_MIN
#define INT32_MIN (-2147483647 - 1)
#endif
#ifndef INT64_MIN
#define INT64_MIN (-9223372036854775807LL - 1LL)
#endif
#ifndef INT8_MAX
#define INT8_MAX +127
#endif
#ifndef INT16_MAX
#define INT16_MAX +32767
#endif
#ifndef INT32_MAX
#define INT32_MAX +2147483647
#endif
#ifndef INT64_MAX
#define INT64_MAX +9223372036854775807LL
#endif
#ifndef UINT8_MAX
#define UINT8_MAX 255
#endif
#ifndef UINT16_MAX
#define UINT16_MAX 65535
#endif
#ifndef UINT32_MAX
#define UINT32_MAX 4294967295U
#endif
#ifndef UINT64_MAX
#define UINT64_MAX 18446744073709551615ULL
#endif

#ifndef INT_LEAST8_MIN
#define INT_LEAST8_MIN INT8_MIN
#endif
#ifndef INT_LEAST16_MIN
#define INT_LEAST16_MIN INT16_MIN
#endif
#ifndef INT_LEAST32_MIN
#define INT_LEAST32_MIN INT32_MIN
#endif
#ifndef INT_LEAST64_MIN
#define INT_LEAST64_MIN INT64_MIN
#endif
#ifndef INT_LEAST8_MAX
#define INT_LEAST8_MAX INT8_MAX
#endif
#ifndef INT_LEAST16_MAX
#define INT_LEAST16_MAX INT16_MAX
#endif
#ifndef INT_LEAST32_MAX
#define INT_LEAST32_MAX INT32_MAX
#endif
#ifndef INT_LEAST64_MAX
#define INT_LEAST64_MAX INT64_MAX
#endif
#ifndef UINT_LEAST8_MAX
#define UINT_LEAST8_MAX UINT8_MAX
#endif
#ifndef UINT_LEAST16_MAX
#define UINT_LEAST16_MAX UINT16_MAX
#endif
#ifndef UINT_LEAST32_MAX
#define UINT_LEAST32_MAX UINT32_MAX
#endif
#ifndef UINT_LEAST64_MAX
#define UINT_LEAST64_MAX UINT64_MAX
#endif

#ifndef INT_FAST8_MIN
#define INT_FAST8_MIN INT8_MIN
#endif
#ifndef INT_FAST16_MIN
#define INT_FAST16_MIN INT32_MIN
#endif
#ifndef INT_FAST32_MIN
#define INT_FAST32_MIN INT32_MIN
#endif
#ifndef INT_FAST64_MIN
#define INT_FAST64_MIN INT64_MIN
#endif
#ifndef INT_FAST8_MAX
#define INT_FAST8_MAX INT8_MAX
#endif
#ifndef INT_FAST16_MAX
#define INT_FAST16_MAX INT32_MAX
#endif
#ifndef INT_FAST32_MAX
#define INT_FAST32_MAX INT32_MAX
#endif
#ifndef INT_FAST64_MAX
#define INT_FAST64_MAX INT64_MAX
#endif
#ifndef UINT_FAST8_MAX
#define UINT_FAST8_MAX UINT8_MAX
#endif
#ifndef UINT_FAST16_MAX
#define UINT_FAST16_MAX UINT32_MAX
#endif
#ifndef UINT_FAST32_MAX
#define UINT_FAST32_MAX UINT32_MAX
#endif
#ifndef UINT_FAST64_MAX
#define UINT_FAST64_MAX UINT64_MAX
#endif

#ifndef INTMAX_MIN
#define INTMAX_MIN INT64_MIN
#endif
#ifndef INTMAX_MAX
#define INTMAX_MAX INT64_MAX
#endif
#ifndef UINTMAX_MAX
#define UINTMAX_MAX UINT64_MAX
#endif

#ifndef SIG_ATOMIC_MIN
#define SIG_ATOMIC_MIN INT32_MIN
#endif
#ifndef SIG_ATOMIC_MAX
#define SIG_ATOMIC_MAX INT32_MAX
#endif

#ifndef INT8_C
#define INT8_C(x) (x)
#endif
#ifndef INT16_C
#define INT16_C(x) (x)
#endif
#ifndef INT32_C
#define INT32_C(x) (x)
#endif
#ifndef INT64_C
#define INT64_C(x) (x##LL)
#endif
#ifndef UINT8_C
#define UINT8_C(x) (x)
#endif
#ifndef UINT16_C
#define UINT16_C(x) (x)
#endif
#ifndef UINT32_C
#define UINT32_C(x) (x##U)
#endif
#ifndef UINT64_C
#define UINT64_C(x) (x##ULL)
#endif
#ifndef INTMAX_C
#define INTMAX_C(x) INT64_C(x)
#endif
#ifndef UINTMAX_C
#define UINTMAX_C(x) UINT64_C(x)
#endif

#ifndef WINT_MIN
#define WINT_MIN 0x0000
#endif
#ifndef WINT_MAX
#define WINT_MAX 0xffff
#endif

#ifdef _MSC_VER
#ifdef _WIN64
typedef long long intptr_t;            // NOLINT
typedef unsigned long long uintptr_t;  // NOLINT
#define INTPTR_MIN INT64_MIN
#define INTPTR_MAX INT64_MAX
#define UINTPTR_MAX UINT64_MAX
#else
typedef __w64 int intptr_t;
typedef __w64 unsigned int uintptr_t;
#define INTPTR_MIN INT32_MIN
#define INTPTR_MAX INT32_MAX
#define UINTPTR_MAX UINT32_MAX
#endif
#elif defined(NLIB_64BIT)
typedef long long intptr_t;            // NOLINT
typedef unsigned long long uintptr_t;  // NOLINT
#define INTPTR_MIN INT64_MIN
#define INTPTR_MAX INT64_MAX
#define UINTPTR_MAX UINT64_MAX
#else
typedef int intptr_t;
typedef unsigned int uintptr_t;
#define INTPTR_MIN INT32_MIN
#define INTPTR_MAX INT32_MAX
#define UINTPTR_MAX UINT32_MAX
#endif

typedef long long intmax_t;            // NOLINT
typedef unsigned long long uintmax_t;  // NOLINT

#ifndef SIZE_MAX
#define SIZE_MAX UINTPTR_MAX
#endif

#ifndef PTRDIFF_MIN
#define PTRDIFF_MIN INTPTR_MIN
#endif
#ifndef PTRDIFF_MAX
#define PTRDIFF_MAX INTPTR_MAX
#endif

#endif  // INCLUDE_NN_NLIB_STDINT_H_
