﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_TEXTWRITER_H_
#define INCLUDE_NN_NLIB_TEXTWRITER_H_

#include "nn/nlib/Config.h"
#include "nn/nlib/Swap.h"

NLIB_NAMESPACE_BEGIN

class OutputStream;

// code snippets:
//   TextWriter writer;
//   if (writer.Init() != 0 || writer.Open(&stream) != 0) { error; }
//   # note that wide characters(UTF16/UTF32) are converted into UTF8
//   if (!writer.Write(....)) { error; }
//   if (!writer.Flush()) { error; }
//   if (!writer.Close()) { error; }
class NLIB_VIS_PUBLIC TextWriter NLIB_FINAL {
 public:
    TextWriter() NLIB_NOEXCEPT;
    ~TextWriter() NLIB_NOEXCEPT {}
    NLIB_MOVE_MEMBER_HELPER_2(TextWriter, stream_, errno_);  // NOLINT
    errno_t Init() NLIB_NOEXCEPT;
    errno_t Open(OutputStream* stream) NLIB_NOEXCEPT;
    bool Write(wchar_t c) NLIB_NOEXCEPT;
    bool Write(const wchar_t* str) NLIB_NOEXCEPT NLIB_NONNULL;
    bool Write(char c) NLIB_NOEXCEPT;
    bool Write(const char* str) NLIB_NOEXCEPT NLIB_NONNULL;
    bool Write(nlib_utf16_t c) NLIB_NOEXCEPT;
    bool Write(const nlib_utf16_t* str) NLIB_NOEXCEPT NLIB_NONNULL;
    bool Write(nlib_utf32_t c) NLIB_NOEXCEPT;
    bool Write(const nlib_utf32_t* str) NLIB_NOEXCEPT NLIB_NONNULL;
    bool Flush() NLIB_NOEXCEPT;
    bool Close() NLIB_NOEXCEPT;
    errno_t GetErrorValue() const NLIB_NOEXCEPT { return errno_; }
    OutputStream* GetStream() NLIB_NOEXCEPT { return stream_; }
    bool WriteFormat(_Printf_format_string_ const wchar_t* fmt, ...) NLIB_NOEXCEPT;
    bool WriteFormat(_Printf_format_string_ const char* fmt, ...) NLIB_NOEXCEPT;
    void SetError(errno_t e) const NLIB_NOEXCEPT {
        if (errno_ == 0) errno_ = e;
    }

    void swap(TextWriter& rhs) NLIB_NOEXCEPT {
        using std::swap;
        swap(stream_, rhs.stream_);
        swap(errno_, rhs.errno_);
        swap(cr_, rhs.cr_);
    }
    NLIB_SAFE_BOOL(TextWriter, GetErrorValue() == 0)

 private:
    NLIB_VIS_HIDDEN bool Write_(const char* str, size_t len) NLIB_NOEXCEPT;

 private:
    OutputStream* stream_;
    mutable ErrnoT errno_;
    unsigned char cr_;
    unsigned char utf8idx_;
    unsigned char utf16idx_;
    union {
        char utf8[5];
        nlib_utf16_t utf16[3];
    } buf_;

    NLIB_DISALLOW_COPY_AND_ASSIGN(TextWriter);
};

NLIB_NAMESPACE_END

NLIB_DEFINE_STD_SWAP(::nlib_ns::TextWriter)

#endif  // INCLUDE_NN_NLIB_TEXTWRITER_H_
