﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_HEAP_STACKHEAP_H_
#define INCLUDE_NN_NLIB_HEAP_STACKHEAP_H_

#include "nn/nlib/Config.h"

#if defined(_MSC_VER) && defined(nx_heap_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

NLIB_NAMESPACE_BEGIN
namespace heap {

class NLIB_VIS_PUBLIC StackHeap NLIB_FINAL {
 public:
    StackHeap() NLIB_NOEXCEPT : cur_(NULL), heap_option_(0), start_(NULL), end_(NULL) {}
    errno_t Init(void* start, size_t size, uint32_t heap_option) NLIB_NOEXCEPT;
    void FreeAll() NLIB_NOEXCEPT { cur_ = start_; }
    errno_t ResizeTop(void* p, size_t from_size, size_t to_size) NLIB_NOEXCEPT;
    errno_t Adjust(void** start, size_t* size) NLIB_NOEXCEPT;

    void* Alloc(size_t n, size_t algn) NLIB_NOEXCEPT {
        if (n == 0) return NULL;
        if (algn == 0 || 0 != (algn & (algn - 1))) return NULL;
        uint8_t* ret = this->RoundUp(cur_, algn);
        uint8_t* new_top = ret + n;
        if (new_top > end_) return NULL;
        cur_ = new_top;
        return ret;
    }
    void* Alloc(size_t n) NLIB_NOEXCEPT { return this->Alloc(n, 8); }
    errno_t Free(void* p) NLIB_NOEXCEPT {
        uint8_t* ptr = reinterpret_cast<uint8_t*>(p);
        if (ptr >= start_ && ptr < cur_) {
            cur_ = ptr;
            return 0;
        } else {
            return EFAULT;
        }
    }
    void Dump() NLIB_NOEXCEPT;

 private:
    static uint8_t* RoundUp(uint8_t* ptr, size_t algn) NLIB_NOEXCEPT {
        // algn is power of 2.
        intptr_t tmp = reinterpret_cast<intptr_t>(ptr);
        return reinterpret_cast<uint8_t*>(tmp + (-tmp & (algn - 1)));
    }
    void ChangeHeapRange(void* start, void* end) NLIB_NOEXCEPT {
        start_ = reinterpret_cast<uint8_t*>(start);
        end_ = reinterpret_cast<uint8_t*>(end);
    }

 private:
    uint8_t* cur_;

    uint32_t heap_option_;
    uint8_t* start_;
    uint8_t* end_;
};

}  // namespace heap
NLIB_NAMESPACE_END

#if defined(_MSC_VER) && defined(nx_heap_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_HEAP_STACKHEAP_H_
