﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_MSGPACK_JSONRPC_JSONRPCREQUEST_H_
#define INCLUDE_NN_NLIB_MSGPACK_JSONRPC_JSONRPCREQUEST_H_

#include <iterator>

#include "nn/nlib/Config.h"
#include "nn/nlib/UniquePtr.h"
#include "nn/nlib/Nlist.h"
#include "nn/nlib/ReallocOutputStream.h"
#include "nn/nlib/msgpack/MpObject.h"
#include "nn/nlib/msgpack/jsonrpc/JsonRpcResponse.h"

#if defined(_MSC_VER) && defined(nx_msgpack_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

NLIB_NAMESPACE_BEGIN
namespace msgpack {
namespace jsonrpc {

namespace fortest {
class JsonRpcErrorRequestTest;
}  // namespace fortest

typedef uint32_t reqid_t;

class NLIB_VIS_PUBLIC JsonRpcRequest NLIB_FINAL {
 public:
    // NOTE:
    // If id_ is 0, the request becomes notification(ID is not sent.).
    JsonRpcRequest() NLIB_NOEXCEPT : id_(0), method_(NULL) {}
    ~JsonRpcRequest() NLIB_NOEXCEPT;
    reqid_t GetId() const NLIB_NOEXCEPT { return id_; }
    const char* GetMethod() const NLIB_NOEXCEPT { return method_; }
    const MpObject& GetParams() const NLIB_NOEXCEPT { return params_; }
    MpObject& GetParams() NLIB_NOEXCEPT { return params_; }

    void SetId(reqid_t id) NLIB_NOEXCEPT { id_ = id; }
    bool SetMethod(const char* method) NLIB_NOEXCEPT;
    void MoveParamsFrom(MpObject& params) NLIB_NOEXCEPT {  // NOLINT
        MpObject tmp;
        tmp.swap(params);
        tmp.swap(params_);
    }

 private:
    NLIB_VIS_HIDDEN JsonRpcResponse& GetResponse() NLIB_NOEXCEPT { return response_; }
    NLIB_VIS_HIDDEN bool SetMethodRaw(const void* p, size_t n) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN errno_t Init(MpObject* rhs) NLIB_NOEXCEPT;

 private:
    reqid_t id_;
    char* method_;
    MpObject params_;
    JsonRpcResponse response_;

    friend NLIB_VIS_PUBLIC size_t JsonRpcServerExec(const void* p, size_t n,
                                                    ReallocOutputStream::UniquePtrType* ptr);
    friend class JsonRpcRequestReader;
    friend class fortest::JsonRpcErrorRequestTest;
    NLIB_DISALLOW_COPY_AND_ASSIGN(JsonRpcRequest);
};

class NLIB_VIS_PUBLIC JsonRpcRequestWriter NLIB_FINAL {
 public:
    JsonRpcRequestWriter() NLIB_NOEXCEPT : prv_(NULL), is_msgpack_(false) {}
    ~JsonRpcRequestWriter() NLIB_NOEXCEPT;
    explicit JsonRpcRequestWriter(bool use_msgpack) NLIB_NOEXCEPT
        : prv_(NULL), is_msgpack_(use_msgpack) {}
    errno_t BeginWriteRequest(uint32_t n) NLIB_NOEXCEPT;
    errno_t WriteRequest(const char* method, reqid_t id, const MpObject& params) NLIB_NOEXCEPT;
    errno_t WriteNotification(const char* method, const MpObject& params) NLIB_NOEXCEPT {
        return WriteRequest(method, 0, params);
    }
    errno_t WriteRequest(const JsonRpcRequest& req) NLIB_NOEXCEPT {
        return WriteRequest(req.GetMethod(), req.GetId(), req.GetParams());
    }
    errno_t EndWriteRequest(ReallocOutputStream::UniquePtrType* ptr, size_t* n) NLIB_NOEXCEPT;

 private:
    struct JsonRpcRequestWriterPrivate;
    JsonRpcRequestWriterPrivate* prv_;
    bool is_msgpack_;

    NLIB_DISALLOW_COPY_AND_ASSIGN(JsonRpcRequestWriter);
};

class NLIB_VIS_PUBLIC JsonRpcRequestReader NLIB_FINAL {
 public:
    typedef Nlist<JsonRpcRequest> ListType;
    static errno_t ReadRequest(const void* p, size_t n, ListType* request_list) NLIB_NOEXCEPT;

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(JsonRpcRequestReader);
};

}  // namespace jsonrpc
}  // namespace msgpack
NLIB_NAMESPACE_END

#if defined(_MSC_VER) && defined(nx_msgpack_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_MSGPACK_JSONRPC_JSONRPCREQUEST_H_
