﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_MSGPACK_JSONRPC_JSONRPCRESPONSE_H_
#define INCLUDE_NN_NLIB_MSGPACK_JSONRPC_JSONRPCRESPONSE_H_

#include "nn/nlib/Config.h"
#include "nn/nlib/UniquePtr.h"
#include "nn/nlib/Nlist.h"
#include "nn/nlib/ReallocOutputStream.h"
#include "nn/nlib/msgpack/MpObject.h"

#if defined(_MSC_VER) && defined(nx_msgpack_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

NLIB_NAMESPACE_BEGIN
namespace msgpack {
namespace jsonrpc {
typedef uint32_t reqid_t;

class NLIB_VIS_PUBLIC JsonRpcResponse NLIB_FINAL {
 public:
    enum ErrorCode {
        OK = 0,
        PARSE_ERROR = -32700,
        INVALID_REQUEST = -32600,
        METHOD_NOT_FOUND = -32601,
        INVALID_PARAMS = -32602,
        INTERNAL_ERROR = -32603,
        CLIENT_ABORT = -31000
    };
    JsonRpcResponse() NLIB_NOEXCEPT
        : id_(0), is_error_(false), error_code_(0), error_msg_(NULL) {}
    ~JsonRpcResponse() NLIB_NOEXCEPT;

    void MoveResultFrom(MpObject& obj) NLIB_NOEXCEPT;  // NOLINT
    void SetError(int errcode, const char* msg, MpObject& data) NLIB_NOEXCEPT;  // NOLINT
    void SetError(int errcode, const char* msg) NLIB_NOEXCEPT {
        MpObject data;
        this->SetError(errcode, msg, data);
    }

    ErrorCode GetError() const NLIB_NOEXCEPT {
        return static_cast<ErrorCode>(error_code_);
    }
    reqid_t GetId() const NLIB_NOEXCEPT { return id_; }
    const char* GetErrorMessage() const NLIB_NOEXCEPT {
        return error_msg_ ? error_msg_ : "";
    }
    MpObject& GetMpObject() NLIB_NOEXCEPT { return obj_; }
    const MpObject& GetMpObject() const NLIB_NOEXCEPT { return obj_; }
    NLIB_SAFE_BOOL(JsonRpcResponse, !is_error_);

 private:
    void SetId(reqid_t id) NLIB_NOEXCEPT { id_ = id; }  // from JsonRpcRequest::Init()
    NLIB_VIS_HIDDEN errno_t Init(MpObject* rhs) NLIB_NOEXCEPT;  // from JsonRpcResponseReader

 private:
    reqid_t id_;
    bool is_error_;
    int error_code_;
    char* error_msg_;
    MpObject obj_;

    friend class JsonRpcResponseReader;
    friend class JsonRpcRequest;
    NLIB_DISALLOW_COPY_AND_ASSIGN(JsonRpcResponse);
};

class NLIB_VIS_PUBLIC JsonRpcResponseReader {
 public:
    typedef UniquePtr<JsonRpcResponse> ValueType;
    typedef Nlist<ValueType> ListType;
    static errno_t ReadResponse(const void* p, size_t n, ListType* result_list) NLIB_NOEXCEPT;
};

class NLIB_VIS_PUBLIC JsonRpcResponseWriter NLIB_FINAL {
 public:
    JsonRpcResponseWriter() NLIB_NOEXCEPT : prv_(NULL), is_msgpack_(false) {}
    ~JsonRpcResponseWriter() NLIB_NOEXCEPT;
    explicit JsonRpcResponseWriter(bool msgpack) NLIB_NOEXCEPT
        : prv_(NULL), is_msgpack_(msgpack) {}
    errno_t BeginWriteResponse(uint32_t n) NLIB_NOEXCEPT;
    errno_t WriteResponse(JsonRpcResponse* result) NLIB_NOEXCEPT;
    errno_t EndWriteResponse(ReallocOutputStream::UniquePtrType* ptr, size_t* n) NLIB_NOEXCEPT;

 private:
    struct JsonRpcResponseWriterPrivate;
    JsonRpcResponseWriterPrivate* prv_;
    bool is_msgpack_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(JsonRpcResponseWriter);
};

}  // namespace jsonrpc
}  // namespace msgpack
NLIB_NAMESPACE_END

#if defined(_MSC_VER) && defined(nx_msgpack_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_MSGPACK_JSONRPC_JSONRPCRESPONSE_H_
