﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_SIMD_SIMDINT_H_
#define INCLUDE_NN_NLIB_SIMD_SIMDINT_H_

#include "nn/nlib/Config.h"

#if defined(NLIB_SSE41)
typedef __m128i nlib_i128_t;
#elif defined(NLIB_NEON)
typedef int8x16_t nlib_i128_t;
#endif

NLIB_NAMESPACE_BEGIN
namespace simd {

// use each_int8 for the argument value
struct each_int8_tag {};
// use each_int16 for the argument value
struct each_int16_tag {};
// use each_int32 for the argument value
struct each_int32_tag {};
// use each_int64 for the argument value
struct each_int64_tag {};
// use each_uint8 for the argument value
struct each_uint8_tag {};
// use each_uint16 for the argument value
struct each_uint16_tag {};
// use each_uint32 for the argument value
struct each_uint32_tag {};
// use each_uint64 for the argument value
struct each_uint64_tag {};

NLIB_CEXPR const each_int8_tag each_int8 = {};
NLIB_CEXPR const each_int16_tag each_int16 = {};
NLIB_CEXPR const each_int32_tag each_int32 = {};
NLIB_CEXPR const each_int64_tag each_int64 = {};
NLIB_CEXPR const each_uint8_tag each_uint8 = {};
NLIB_CEXPR const each_uint16_tag each_uint16 = {};
NLIB_CEXPR const each_uint32_tag each_uint32 = {};
NLIB_CEXPR const each_uint64_tag each_uint64 = {};

// use each_select32 for the argument value
struct each_select32_tag {};
// use each_select16 for the argument value
struct each_select16_tag {};
// use each_select8 for the argument value
struct each_select8_tag {};

NLIB_CEXPR const each_select32_tag each_select32 = {};
NLIB_CEXPR const each_select16_tag each_select16 = {};
NLIB_CEXPR const each_select8_tag each_select8 = {};

#if !defined(_MSC_VER) || _MSC_VER < 1800
#ifndef __vectorcall
#define __vectorcall
#endif
#endif

#if defined(NLIB_SIMD)

// __m128i(SSE), int8x16_t(NEON)
typedef nlib_i128_t i128;

#if _MSC_VER < 1800
typedef const i128& i128arg;
#else
typedef const i128 i128arg;
#endif

struct NLIB_VIS_HIDDEN I128 {
    static i128 __vectorcall SetValue(int8_t v, each_int8_tag) NLIB_NOEXCEPT;
    static i128 __vectorcall SetValue(int16_t v, each_int16_tag) NLIB_NOEXCEPT;
    static i128 __vectorcall SetValue(int32_t v, each_int32_tag) NLIB_NOEXCEPT;
    static i128 __vectorcall SetValue(int64_t v, each_int64_tag) NLIB_NOEXCEPT;
    static i128 __vectorcall SetValue(uint8_t v, each_uint8_tag) NLIB_NOEXCEPT;
    static i128 __vectorcall SetValue(uint16_t v, each_uint16_tag) NLIB_NOEXCEPT;
    static i128 __vectorcall SetValue(uint32_t v, each_uint32_tag) NLIB_NOEXCEPT;
    static i128 __vectorcall SetValue(uint64_t v, each_uint64_tag) NLIB_NOEXCEPT;

    template <size_t N>
    static i128 __vectorcall SetValue(i128 value, each_select32_tag) NLIB_NOEXCEPT;
    template <size_t N>
    static i128 __vectorcall SetValue(i128 value, each_select16_tag) NLIB_NOEXCEPT;
    template <size_t N>
    static i128 __vectorcall SetValue(i128 value, each_select8_tag) NLIB_NOEXCEPT;

    static i128 __vectorcall SetZero() NLIB_NOEXCEPT;
    static i128 __vectorcall SetFull(i128arg dummy) NLIB_NOEXCEPT;

    static i128 __vectorcall LoadA16(const void* p) NLIB_NOEXCEPT;
    static i128 __vectorcall LoadA8(const void* p) NLIB_NOEXCEPT;
    static i128 __vectorcall LoadA4(const void* p) NLIB_NOEXCEPT;
    static i128 __vectorcall LoadA2(const void* p) NLIB_NOEXCEPT;
    static i128 __vectorcall LoadA1(const void* p) NLIB_NOEXCEPT;
    static i128 __vectorcall LoadLoA8(const void* p) NLIB_NOEXCEPT;
    static i128 __vectorcall LoadLoA4(const void* p) NLIB_NOEXCEPT;
    static i128 __vectorcall LoadLoA2(const void* p) NLIB_NOEXCEPT;
    static i128 __vectorcall LoadLoA1(const void* p) NLIB_NOEXCEPT;
    static i128 __vectorcall LoadHiA8(const void* p) NLIB_NOEXCEPT;
    static i128 __vectorcall LoadHiA4(const void* p) NLIB_NOEXCEPT;
    static i128 __vectorcall LoadHiA2(const void* p) NLIB_NOEXCEPT;
    static i128 __vectorcall LoadHiA1(const void* p) NLIB_NOEXCEPT;

#define NLIB_LOAD_REDIRECT(func)                               \
    static i128 __vectorcall func(uintptr_t p) NLIB_NOEXCEPT { \
        return func(reinterpret_cast<void*>(p));               \
    }                                                          \
    static i128 __vectorcall func(intptr_t p) NLIB_NOEXCEPT {  \
        return func(reinterpret_cast<void*>(p));               \
    }
    NLIB_LOAD_REDIRECT(LoadA16)
    NLIB_LOAD_REDIRECT(LoadA8)
    NLIB_LOAD_REDIRECT(LoadA4)
    NLIB_LOAD_REDIRECT(LoadA2)
    NLIB_LOAD_REDIRECT(LoadA1)
    NLIB_LOAD_REDIRECT(LoadLoA8)
    NLIB_LOAD_REDIRECT(LoadLoA4)
    NLIB_LOAD_REDIRECT(LoadLoA2)
    NLIB_LOAD_REDIRECT(LoadLoA1)
    NLIB_LOAD_REDIRECT(LoadHiA8)
    NLIB_LOAD_REDIRECT(LoadHiA4)
    NLIB_LOAD_REDIRECT(LoadHiA2)
    NLIB_LOAD_REDIRECT(LoadHiA1)
#undef NLIB_LOAD_REDIRECT

    static void __vectorcall StoreA16(void* p, i128arg value) NLIB_NOEXCEPT;
    static void __vectorcall StoreA8(void* p, i128arg value) NLIB_NOEXCEPT;
    static void __vectorcall StoreA4(void* p, i128arg value) NLIB_NOEXCEPT;
    static void __vectorcall StoreA2(void* p, i128arg value) NLIB_NOEXCEPT;
    static void __vectorcall StoreA1(void* p, i128arg value) NLIB_NOEXCEPT;
    static void __vectorcall StoreLoA8(void* p, i128arg value) NLIB_NOEXCEPT;
    static void __vectorcall StoreLoA4(void* p, i128arg value) NLIB_NOEXCEPT;
    static void __vectorcall StoreLoA2(void* p, i128arg value) NLIB_NOEXCEPT;
    static void __vectorcall StoreLoA1(void* p, i128arg value) NLIB_NOEXCEPT;
    static void __vectorcall StoreHiA8(void* p, i128arg value) NLIB_NOEXCEPT;
    static void __vectorcall StoreHiA4(void* p, i128arg value) NLIB_NOEXCEPT;
    static void __vectorcall StoreHiA2(void* p, i128arg value) NLIB_NOEXCEPT;
    static void __vectorcall StoreHiA1(void* p, i128arg value) NLIB_NOEXCEPT;

#define NLIB_STORE_REDIRECT(func)                                             \
    static void __vectorcall func(uintptr_t p, i128arg value) NLIB_NOEXCEPT { \
        func(reinterpret_cast<void*>(p), value);                              \
    }                                                                         \
    static void __vectorcall func(intptr_t p, i128arg value) NLIB_NOEXCEPT {  \
        func(reinterpret_cast<void*>(p), value);                              \
    }
    NLIB_STORE_REDIRECT(StoreA16)
    NLIB_STORE_REDIRECT(StoreA8)
    NLIB_STORE_REDIRECT(StoreA4)
    NLIB_STORE_REDIRECT(StoreA2)
    NLIB_STORE_REDIRECT(StoreA1)
    NLIB_STORE_REDIRECT(StoreLoA8)
    NLIB_STORE_REDIRECT(StoreLoA4)
    NLIB_STORE_REDIRECT(StoreLoA2)
    NLIB_STORE_REDIRECT(StoreLoA1)
    NLIB_STORE_REDIRECT(StoreHiA8)
    NLIB_STORE_REDIRECT(StoreHiA4)
    NLIB_STORE_REDIRECT(StoreHiA2)
    NLIB_STORE_REDIRECT(StoreHiA1)
#undef NLIB_STORE_REDIRECT

    //
    // Get/Set
    //
    template <size_t N>
    static uint8_t __vectorcall GetUint8FromLane(i128arg value) NLIB_NOEXCEPT;
    template <size_t N>
    static uint16_t __vectorcall GetUint16FromLane(i128arg value) NLIB_NOEXCEPT;
    template <size_t N>
    static uint32_t __vectorcall GetUint32FromLane(i128arg value) NLIB_NOEXCEPT;
    template <size_t N>
    static uint64_t __vectorcall GetUint64FromLane(i128arg value) NLIB_NOEXCEPT;
    template <size_t N>
    static i128 __vectorcall SetUint8ToLane(i128arg value, uint8_t v) NLIB_NOEXCEPT;
    template <size_t N>
    static i128 __vectorcall SetUint16ToLane(i128arg value, uint16_t v) NLIB_NOEXCEPT;
    template <size_t N>
    static i128 __vectorcall SetUint32ToLane(i128arg value, uint32_t v) NLIB_NOEXCEPT;
    template <size_t N>
    static i128 __vectorcall SetUint64ToLane(i128arg value, uint64_t v) NLIB_NOEXCEPT;

    //
    // Arithmetic Operations
    //
    static i128 __vectorcall Add8(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Add16(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Add32(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Add64(i128arg a, i128arg b) NLIB_NOEXCEPT;

    static i128 __vectorcall AddInt8Saturated(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall AddInt16Saturated(i128arg a, i128arg b) NLIB_NOEXCEPT;

    static i128 __vectorcall AddUint8Saturated(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall AddUint16Saturated(i128arg a, i128arg b) NLIB_NOEXCEPT;

    static i128 __vectorcall Sub8(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Sub16(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Sub32(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Sub64(i128arg a, i128arg b) NLIB_NOEXCEPT;

    static i128 __vectorcall SubInt8Saturated(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall SubInt16Saturated(i128arg a, i128arg b) NLIB_NOEXCEPT;

    static i128 __vectorcall SubUint8Saturated(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall SubUint16Saturated(i128arg a, i128arg b) NLIB_NOEXCEPT;

    static i128 __vectorcall PairwiseAdd8(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall PairwiseAdd16(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall PairwiseAdd32(i128arg a, i128arg b) NLIB_NOEXCEPT;
    // PairwiseMaxInt8, PairwiseMaxInt16, PairwiseMaxInt32
    // PairwiseMaxUint8, PairwiseMaxUint16, PairwiseMaxUint32
    // PairwiseMinInt8, PairwiseMinInt16, PairwiseMinInt32
    // PairwiseMinUint8, PairwiseMinUint16, PairwiseMinUint32

    static i128 __vectorcall Mult16(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall MultAdd16(i128arg a, i128arg b, i128arg c) NLIB_NOEXCEPT;
    static i128 __vectorcall MultSub16(i128arg a, i128arg b, i128arg c) NLIB_NOEXCEPT;
    static i128 __vectorcall Mult32(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall MultAdd32(i128arg a, i128arg b, i128arg c) NLIB_NOEXCEPT;
    static i128 __vectorcall MultSub32(i128arg a, i128arg b, i128arg c) NLIB_NOEXCEPT;

    static i128 __vectorcall NegateInt8(i128arg value) NLIB_NOEXCEPT;
    static i128 __vectorcall NegateInt16(i128arg value) NLIB_NOEXCEPT;
    static i128 __vectorcall NegateInt32(i128arg value) NLIB_NOEXCEPT;

    static i128 __vectorcall MaxInt8(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall MaxInt16(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall MaxInt32(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall MaxUint8(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall MaxUint16(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall MaxUint32(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall MinInt8(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall MinInt16(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall MinInt32(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall MinUint8(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall MinUint16(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall MinUint32(i128arg a, i128arg b) NLIB_NOEXCEPT;

    static i128 __vectorcall AbsInt8(i128arg value) NLIB_NOEXCEPT;
    static i128 __vectorcall AbsInt16(i128arg value) NLIB_NOEXCEPT;
    static i128 __vectorcall AbsInt32(i128arg value) NLIB_NOEXCEPT;
    static i128 __vectorcall AbsDiffInt8(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall AbsDiffInt16(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall AbsDiffInt32(i128arg a, i128arg b) NLIB_NOEXCEPT;

    //
    // Logical Operations
    //
    static i128 __vectorcall And(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Or(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Xor(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Not(i128arg a) NLIB_NOEXCEPT;
    static i128 __vectorcall AndNot(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall OrNot(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Test8(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Test16(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Test32(i128arg a, i128arg b) NLIB_NOEXCEPT;

    //
    // Comparison Operations
    //
    static i128 __vectorcall CmpEq8(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpEq16(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpEq32(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpEq64(i128arg a, i128arg b) NLIB_NOEXCEPT;

    static i128 __vectorcall CmpLtInt8(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpLtInt16(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpLtInt32(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpLtInt64(i128arg a, i128arg b) NLIB_NOEXCEPT;

    static i128 __vectorcall CmpGtInt8(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpGtInt16(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpGtInt32(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpGtInt64(i128arg a, i128arg b) NLIB_NOEXCEPT;

    static i128 __vectorcall CmpLtUint8(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpLtUint16(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpLtUint32(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpLtUint64(i128arg a, i128arg b) NLIB_NOEXCEPT;

    static i128 __vectorcall CmpGtUint8(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpGtUint16(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpGtUint32(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpGtUint64(i128arg a, i128arg b) NLIB_NOEXCEPT;

    static i128 __vectorcall CmpLeInt8(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpLeInt16(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpLeInt32(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpLeInt64(i128arg a, i128arg b) NLIB_NOEXCEPT;

    static i128 __vectorcall CmpGeInt8(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpGeInt16(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpGeInt32(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpGeInt64(i128arg a, i128arg b) NLIB_NOEXCEPT;

    static i128 __vectorcall CmpLeUint8(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpLeUint16(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpLeUint32(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpLeUint64(i128arg a, i128arg b) NLIB_NOEXCEPT;

    static i128 __vectorcall CmpGeUint8(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpGeUint16(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpGeUint32(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpGeUint64(i128arg a, i128arg b) NLIB_NOEXCEPT;

    static i128 __vectorcall CmpEqZero8(i128arg value) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpEqZero16(i128arg value) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpEqZero32(i128arg value) NLIB_NOEXCEPT;
    static i128 __vectorcall CmpEqZero64(i128arg value) NLIB_NOEXCEPT;

    //
    // Bit Shift
    //
    static i128 __vectorcall ShiftLeftLogical8(i128arg value, int count) NLIB_NOEXCEPT;
    static i128 __vectorcall ShiftRightLogical8(i128arg value, int count) NLIB_NOEXCEPT;
    static i128 __vectorcall ShiftRightArithmetic8(i128arg value, int count) NLIB_NOEXCEPT;

    static i128 __vectorcall ShiftLeftLogical16(i128arg value, int count) NLIB_NOEXCEPT;
    static i128 __vectorcall ShiftRightLogical16(i128arg value, int count) NLIB_NOEXCEPT;
    static i128 __vectorcall ShiftRightArithmetic16(i128arg value, int count) NLIB_NOEXCEPT;

    static i128 __vectorcall ShiftLeftLogical32(i128arg value, int count) NLIB_NOEXCEPT;
    static i128 __vectorcall ShiftRightLogical32(i128arg value, int count) NLIB_NOEXCEPT;
    static i128 __vectorcall ShiftRightArithmetic32(i128arg value, int count) NLIB_NOEXCEPT;

    static i128 __vectorcall ShiftLeftLogical64(i128arg value, int count) NLIB_NOEXCEPT;
    static i128 __vectorcall ShiftRightLogical64(i128arg value, int count) NLIB_NOEXCEPT;

    //
    // Bit Shift(Constant)
    //
    template <size_t N>
    static i128 __vectorcall ShiftLeftLogical8(i128arg value) NLIB_NOEXCEPT;
    template <size_t N>
    static i128 __vectorcall ShiftRightLogical8(i128arg value) NLIB_NOEXCEPT;
    template <size_t N>
    static i128 __vectorcall ShiftRightArithmetic8(i128arg value) NLIB_NOEXCEPT;

    template <size_t N>
    static i128 __vectorcall ShiftLeftLogical16(i128arg value) NLIB_NOEXCEPT;
    template <size_t N>
    static i128 __vectorcall ShiftRightLogical16(i128arg value) NLIB_NOEXCEPT;
    template <size_t N>
    static i128 __vectorcall ShiftRightArithmetic16(i128arg value) NLIB_NOEXCEPT;

    template <size_t N>
    static i128 __vectorcall ShiftLeftLogical32(i128arg value) NLIB_NOEXCEPT;
    template <size_t N>
    static i128 __vectorcall ShiftRightLogical32(i128arg value) NLIB_NOEXCEPT;
    template <size_t N>
    static i128 __vectorcall ShiftRightArithmetic32(i128arg value) NLIB_NOEXCEPT;

    template <size_t N>
    static i128 __vectorcall ShiftLeftLogical64(i128arg value) NLIB_NOEXCEPT;
    template <size_t N>
    static i128 __vectorcall ShiftRightLogical64(i128arg value) NLIB_NOEXCEPT;

    //
    // 128bit wide Byte Shift/Rotate
    //
    template <size_t N>
    static i128 __vectorcall ByteShiftLeft(i128arg value) NLIB_NOEXCEPT;
    template <size_t N>
    static i128 __vectorcall ByteShiftRight(i128arg value) NLIB_NOEXCEPT;
    template <size_t N>
    static i128 __vectorcall ByteRotateRight(i128arg value) NLIB_NOEXCEPT;
    template <size_t N>
    static i128 __vectorcall AlignR(i128arg a, i128arg b) NLIB_NOEXCEPT;

    //
    // Conversion
    //
    static i128 __vectorcall NarrowFrom16To8(i128arg lo, i128arg hi) NLIB_NOEXCEPT;
    static i128 __vectorcall NarrowFrom32To16(i128arg lo, i128arg hi) NLIB_NOEXCEPT;
    static i128 __vectorcall NarrowFrom64To32(i128arg lo, i128arg hi) NLIB_NOEXCEPT;

    static i128 __vectorcall
        ConvertFromUint16ToUint8Saturated(i128arg lo, i128arg hi) NLIB_NOEXCEPT;  // NOLINT
    static i128 __vectorcall ConvertFromInt16ToInt8Saturated(i128arg lo, i128arg hi) NLIB_NOEXCEPT;
    static i128 __vectorcall
        ConvertFromUint32ToUint16Saturated(i128arg lo, i128arg hi) NLIB_NOEXCEPT;  // NOLINT
    static i128 __vectorcall
        ConvertFromInt32ToInt16Saturated(i128arg lo, i128arg hi) NLIB_NOEXCEPT;  // NOLINT

    static i128 __vectorcall ConvertFromInt8ToInt16Lo(i128arg value) NLIB_NOEXCEPT;
    static i128 __vectorcall ConvertFromInt8ToInt16Hi(i128arg value) NLIB_NOEXCEPT;
    static i128 __vectorcall ConvertFromInt16ToInt32Lo(i128arg value) NLIB_NOEXCEPT;
    static i128 __vectorcall ConvertFromInt16ToInt32Hi(i128arg value) NLIB_NOEXCEPT;
    static i128 __vectorcall ConvertFromInt32ToInt64Lo(i128arg value) NLIB_NOEXCEPT;
    static i128 __vectorcall ConvertFromInt32ToInt64Hi(i128arg value) NLIB_NOEXCEPT;
    static i128 __vectorcall ConvertFromUint8ToUint16Lo(i128arg value) NLIB_NOEXCEPT;
    static i128 __vectorcall ConvertFromUint8ToUint16Hi(i128arg value) NLIB_NOEXCEPT;
    static i128 __vectorcall ConvertFromUint16ToUint32Lo(i128arg value) NLIB_NOEXCEPT;
    static i128 __vectorcall ConvertFromUint16ToUint32Hi(i128arg value) NLIB_NOEXCEPT;
    static i128 __vectorcall ConvertFromUint32ToUint64Lo(i128arg value) NLIB_NOEXCEPT;
    static i128 __vectorcall ConvertFromUint32ToUint64Hi(i128arg value) NLIB_NOEXCEPT;

    static i128 __vectorcall Zip8Lo(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Zip8Hi(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Unzip8Lo(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Unzip8Hi(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Zip16Lo(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Zip16Hi(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Unzip16Lo(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Unzip16Hi(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Zip32Lo(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Zip32Hi(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Unzip32Lo(i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Unzip32Hi(i128arg a, i128arg b) NLIB_NOEXCEPT;

    template<int V0, int V1, int V2, int V3, int V4, int V5, int V6, int V7,
             int V8, int V9, int V10, int V11, int V12, int V13, int V14, int V15>
    static i128 __vectorcall Permute8(i128arg a, i128arg b) NLIB_NOEXCEPT;
    template<int V0, int V1, int V2, int V3, int V4, int V5, int V6, int V7>
    static i128 __vectorcall Permute16(i128arg a, i128arg b) NLIB_NOEXCEPT;
    template<int V0, int V1, int V2, int V3>
    static i128 __vectorcall Permute32(i128arg a, i128arg b) NLIB_NOEXCEPT;

    //
    // Swap endian
    //
    static i128 __vectorcall Reverse16(i128arg value) NLIB_NOEXCEPT;
    static i128 __vectorcall Reverse32(i128arg value) NLIB_NOEXCEPT;
    static i128 __vectorcall Reverse64(i128arg value) NLIB_NOEXCEPT;

    //
    // Misc
    //
    static int __vectorcall MoveMask8(i128arg value) NLIB_NOEXCEPT;
    static int __vectorcall MoveMask16(i128arg value) NLIB_NOEXCEPT;
    static int __vectorcall MoveMask32(i128arg value) NLIB_NOEXCEPT;
    static i128 __vectorcall SetMask8(int mask) NLIB_NOEXCEPT;
    static i128 __vectorcall SetMask16(int mask) NLIB_NOEXCEPT;
    static i128 __vectorcall SetMask32(int mask) NLIB_NOEXCEPT;
    static bool __vectorcall IsZero(i128arg value) NLIB_NOEXCEPT;
    static bool __vectorcall IsFull(i128arg value) NLIB_NOEXCEPT;
    static i128 __vectorcall Select(i128arg mask, i128arg a, i128arg b) NLIB_NOEXCEPT;
    static i128 __vectorcall Shuffle8(i128arg value, i128arg shuffle) NLIB_NOEXCEPT;
    static int __vectorcall PopCntMask8(i128arg value) NLIB_NOEXCEPT;
    static int __vectorcall ClzMask8(i128arg value) NLIB_NOEXCEPT;
    static int __vectorcall CtzMask8(i128arg value) NLIB_NOEXCEPT;

 private:
    I128();  // forbidden
};

#ifndef NLIB_DOXYGEN

#define NLIB_M(tp) NLIB_ALWAYS_INLINE tp __vectorcall
#define NLIB_M2(tp) inline tp __vectorcall

#ifdef NLIB_NEON
# undef vreinterpret_s8_s8
# undef NLIB_OP1
# undef NLIB_OP2
# undef NLIB_OP3
# undef NLIB_CMP
# undef NLIB_SFT
# undef NLIB_CMB

#define vreinterpretq_s8_s8(a) (a)
#define NLIB_OP1(intrin, tp, a) \
    vreinterpretq_s8_##tp(intrin##_##tp(vreinterpretq_##tp##_s8(a)))
#define NLIB_OP2(intrin, tp, a, b) \
    vreinterpretq_s8_##tp(intrin##_##tp(vreinterpretq_##tp##_s8(a), \
                                        vreinterpretq_##tp##_s8(b)))
#define NLIB_OP3(intrin, tp, a, b, c) \
    vreinterpretq_s8_##tp(intrin##_##tp(vreinterpretq_##tp##_s8(a), \
                                        vreinterpretq_##tp##_s8(b), \
                                        vreinterpretq_##tp##_s8(c)))
#define NLIB_CMP(intrin, tp, a, b, utp) \
    vreinterpretq_s8_##utp(intrin##_##tp(vreinterpretq_##tp##_s8(a), \
                                         vreinterpretq_##tp##_s8(b)))
#define NLIB_SFT(intrin, tp, a, cnt, stp) \
    vreinterpretq_s8_##tp(intrin##_##tp(vreinterpretq_##tp##_s8(a), vdupq_n_##stp(cnt)))
#define NLIB_CMB(tp, l, h) vreinterpretq_s8_##tp(vcombine_##tp(l, h))
#endif

// r.s8[i] = v for all i_mm_cvtsi32_si128
NLIB_M(i128) I128::SetValue(int8_t v, each_int8_tag) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    // faster than _mm_set1_epi8(v)
    return _mm_shuffle_epi8(_mm_cvtsi32_si128(static_cast<uint8_t>(v)), _mm_setzero_si128());
#elif defined(NLIB_NEON)
    return vdupq_n_s8(v);
#endif
}

// r.s16[i] = v for all i
NLIB_M(i128) I128::SetValue(int16_t v, each_int16_tag) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_set1_epi16(v);
#elif defined(NLIB_NEON)
    return vreinterpretq_s8_s16(vdupq_n_s16(v));
#endif
}

// r.s32[i] = v for all i
NLIB_M(i128) I128::SetValue(int32_t v, each_int32_tag) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_set1_epi32(v);
#elif defined(NLIB_NEON)
    return vreinterpretq_s8_s32(vdupq_n_s32(v));
#endif
}

// r.s64[i] = v for all i
NLIB_M(i128) I128::SetValue(int64_t v, each_int64_tag) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
#ifdef _MSC_VER
    // conflicts with ffloor()?
    // Do not use MMX anyway
    // return _mm_set1_epi64(*reinterpret_cast<__m64*>(&x));
    NLIB_ALIGNAS(16) int64_t tmp[2] = {v, v};
    return I128::LoadA16(tmp);
#else
    return _mm_set1_epi64x(v);
#endif
#elif defined(NLIB_NEON)
    return vreinterpretq_s8_s64(vdupq_n_s64(v));
#endif
}

// r.u8[i] = v for all i
NLIB_M(i128) I128::SetValue(uint8_t v, each_uint8_tag) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    // faster than _mm_set1_epi8(v)
    return _mm_shuffle_epi8(_mm_cvtsi32_si128(v), _mm_setzero_si128());
#elif defined(NLIB_NEON)
    return vreinterpretq_s8_u8(vdupq_n_u8(v));
#endif
}

// r.u16[i] = v for all i
NLIB_M(i128) I128::SetValue(uint16_t v, each_uint16_tag) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_set1_epi16(static_cast<int16_t>(v));
#elif defined(NLIB_NEON)
    return vreinterpretq_s8_u16(vdupq_n_u16(v));
#endif
}

// r.u32[i] = v for all i
NLIB_M(i128) I128::SetValue(uint32_t v, each_uint32_tag) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_set1_epi32(static_cast<int32_t>(v));
#elif defined(NLIB_NEON)
    return vreinterpretq_s8_u32(vdupq_n_u32(v));
#endif
}

// r.u64[i] = v for all i
NLIB_M(i128) I128::SetValue(uint64_t v, each_uint64_tag) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
#ifdef _MSC_VER
    NLIB_ALIGNAS(16) uint64_t tmp[2] = {v, v};
    return I128::LoadA16(tmp);
#else
    return _mm_set1_epi64x(static_cast<int64_t>(v));
#endif
#elif defined(NLIB_NEON)
    return vreinterpretq_s8_u64(vdupq_n_u64(v));
#endif
}

#if defined(NLIB_SSE41)
template <size_t N>
// r.u32[i] = value.u32[N] for all i
NLIB_M(i128) I128::SetValue(i128 value, each_select32_tag) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N < 4);
    return _mm_shuffle_epi32(value, _MM_SHUFFLE(N, N, N, N));
}
#elif defined(NLIB_NEON)
# ifdef __aarch64__
template <size_t N>
// r.u32[i] = value.u32[N] for all i
NLIB_M(i128) I128::SetValue(i128 value, each_select32_tag) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N < 4);
    uint32x4_t v = vreinterpretq_u32_s8(value);
    return vreinterpretq_s8_u32(vdupq_laneq_u32(v, N));
}
# else
template <>
NLIB_M(i128) I128::SetValue<0>(i128 value, each_select32_tag) NLIB_NOEXCEPT {
    uint32x2_t v = vget_low_u32(vreinterpretq_u32_s8(value));
    return vreinterpretq_s8_u32(vdupq_lane_u32(v, 0));
}
template <>
NLIB_M(i128) I128::SetValue<1>(i128 value, each_select32_tag) NLIB_NOEXCEPT {
    uint32x2_t v = vget_low_u32(vreinterpretq_u32_s8(value));
    return vreinterpretq_s8_u32(vdupq_lane_u32(v, 1));
}
template <>
NLIB_M(i128) I128::SetValue<2>(i128 value, each_select32_tag) NLIB_NOEXCEPT {
    uint32x2_t v = vget_high_u32(vreinterpretq_u32_s8(value));
    return vreinterpretq_s8_u32(vdupq_lane_u32(v, 0));
}
template <>
NLIB_M(i128) I128::SetValue<3>(i128 value, each_select32_tag) NLIB_NOEXCEPT {
    uint32x2_t v = vget_high_u32(vreinterpretq_u32_s8(value));
    return vreinterpretq_s8_u32(vdupq_lane_u32(v, 1));
}
# endif
#endif

#if defined(NLIB_SSE41)
template <size_t N>
// r.u16[i] = value.u16[N] for all i
NLIB_M2(i128) I128::SetValue(i128 value, each_select16_tag) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N < 8);
    NLIB_ALIGNAS(16) const int8_t mask[16] = {
        2 * N, 2 * N + 1, 2 * N, 2 * N + 1, 2 * N, 2 * N + 1, 2 * N, 2 * N + 1,
        2 * N, 2 * N + 1, 2 * N, 2 * N + 1, 2 * N, 2 * N + 1, 2 * N, 2 * N + 1
    };
    return _mm_shuffle_epi8(value, *reinterpret_cast<const __m128i*>(mask));
}
#elif defined(NLIB_NEON)
template <>
NLIB_M(i128) I128::SetValue<0>(i128 value, each_select16_tag) NLIB_NOEXCEPT {
# ifdef __aarch64__
    uint16x8_t v = vreinterpretq_u16_s8(value);
    return vreinterpretq_s8_u16(vdupq_laneq_u16(v, 0));
# else
    uint16x4_t v = vget_low_u16(vreinterpretq_u16_s8(value));
    return vreinterpretq_s8_u16(vdupq_lane_u16(v, 0));
# endif
}

template <>
NLIB_M(i128) I128::SetValue<1>(i128 value, each_select16_tag) NLIB_NOEXCEPT {
# ifdef __aarch64__
    uint16x8_t v = vreinterpretq_u16_s8(value);
    return vreinterpretq_s8_u16(vdupq_laneq_u16(v, 1));
# else
    uint16x4_t v = vget_low_u16(vreinterpretq_u16_s8(value));
    return vreinterpretq_s8_u16(vdupq_lane_u16(v, 1));
# endif
}

template <>
NLIB_M(i128) I128::SetValue<2>(i128 value, each_select16_tag) NLIB_NOEXCEPT {
# ifdef __aarch64__
    uint16x8_t v = vreinterpretq_u16_s8(value);
    return vreinterpretq_s8_u16(vdupq_laneq_u16(v, 2));
# else
    uint16x4_t v = vget_low_u16(vreinterpretq_u16_s8(value));
    return vreinterpretq_s8_u16(vdupq_lane_u16(v, 2));
# endif
}

template <>
NLIB_M(i128) I128::SetValue<3>(i128 value, each_select16_tag) NLIB_NOEXCEPT {
# ifdef __aarch64__
    uint16x8_t v = vreinterpretq_u16_s8(value);
    return vreinterpretq_s8_u16(vdupq_laneq_u16(v, 3));
# else
    uint16x4_t v = vget_low_u16(vreinterpretq_u16_s8(value));
    return vreinterpretq_s8_u16(vdupq_lane_u16(v, 3));
# endif
}

template <>
NLIB_M(i128) I128::SetValue<4>(i128 value, each_select16_tag) NLIB_NOEXCEPT {
# ifdef __aarch64__
    uint16x8_t v = vreinterpretq_u16_s8(value);
    return vreinterpretq_s8_u16(vdupq_laneq_u16(v, 4));
# else
    uint16x4_t v = vget_high_u16(vreinterpretq_u16_s8(value));
    return vreinterpretq_s8_u16(vdupq_lane_u16(v, 0));
# endif
}

template <>
NLIB_M(i128) I128::SetValue<5>(i128 value, each_select16_tag) NLIB_NOEXCEPT {
# ifdef __aarch64__
    uint16x8_t v = vreinterpretq_u16_s8(value);
    return vreinterpretq_s8_u16(vdupq_laneq_u16(v, 5));
# else
    uint16x4_t v = vget_high_u16(vreinterpretq_u16_s8(value));
    return vreinterpretq_s8_u16(vdupq_lane_u16(v, 1));
# endif
}

template <>
NLIB_M(i128) I128::SetValue<6>(i128 value, each_select16_tag) NLIB_NOEXCEPT {
# ifdef __aarch64__
    uint16x8_t v = vreinterpretq_u16_s8(value);
    return vreinterpretq_s8_u16(vdupq_laneq_u16(v, 6));
# else
    uint16x4_t v = vget_high_u16(vreinterpretq_u16_s8(value));
    return vreinterpretq_s8_u16(vdupq_lane_u16(v, 2));
# endif
}

template <>
NLIB_M(i128) I128::SetValue<7>(i128 value, each_select16_tag) NLIB_NOEXCEPT {
# ifdef __aarch64__
    uint16x8_t v = vreinterpretq_u16_s8(value);
    return vreinterpretq_s8_u16(vdupq_laneq_u16(v, 7));
# else
    uint16x4_t v = vget_high_u16(vreinterpretq_u16_s8(value));
    return vreinterpretq_s8_u16(vdupq_lane_u16(v, 3));
# endif
}
#endif

#if defined(NLIB_SSE41)
template <size_t N>
// r.u8[i] = value.u8[N] for all i
NLIB_M2(i128) I128::SetValue(i128 value, each_select8_tag) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N < 16);
    NLIB_ALIGNAS(16) const int8_t mask[16] = {
        N, N, N, N, N, N, N, N, N, N, N, N, N, N, N, N
    };
    return _mm_shuffle_epi8(value, *reinterpret_cast<const __m128i*>(&mask[0]));
}
#elif defined(NLIB_NEON)
namespace detail {
template <size_t N, bool IsLower>
struct SetValue8Helper {
    NLIB_M(i128) operator()(i128 value) NLIB_NOEXCEPT {
        return vdupq_lane_s8(vget_low_s8(value), N);
    }
};

template <size_t N>
struct SetValue8Helper<N, false> {
    NLIB_M(i128) operator()(i128 value) NLIB_NOEXCEPT {
        return vdupq_lane_s8(vget_high_s8(value), N - 8);
    }
};

}  // namespace detail

template <size_t N>
NLIB_M(i128) I128::SetValue(i128 value, each_select8_tag) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N < 16);
# ifdef __aarch64__
    return vdupq_laneq_s8(value, N);
# else
    return detail::SetValue8Helper<N, (N < 8)>()(value);
# endif
}
#endif

// set 0 to all bits
NLIB_M(i128) I128::SetZero() NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_setzero_si128();
#elif defined(NLIB_NEON)
    return vdupq_n_s8(0);
#endif
}

// set 1 to all bits
NLIB_M(i128) I128::SetFull(i128arg dummy) NLIB_NOEXCEPT { return I128::CmpEq8(dummy, dummy); }

// r[i] = p[i], p is 16 bytes aligned
NLIB_M(i128) I128::LoadA16(const void* p) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_load_si128(reinterpret_cast<const __m128i*>(p));
#elif defined(NLIB_NEON)
    uint64x2_t tmp = vld1q_u64(reinterpret_cast<const uint64_t*>(p));
    return vreinterpretq_s8_u64(tmp);
#endif
}

// r[i] = p[i], p is 8 bytes aligned
NLIB_M(i128) I128::LoadA8(const void* p) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_loadu_si128(reinterpret_cast<const __m128i*>(p));
#elif defined(NLIB_NEON)
    uint64x2_t tmp = vld1q_u64(reinterpret_cast<const uint64_t*>(p));
    return vreinterpretq_s8_u64(tmp);
#endif
}

// r[i] = p[i], p is 4 bytes aligned
NLIB_M(i128) I128::LoadA4(const void* p) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_loadu_si128(reinterpret_cast<const __m128i*>(p));
#elif defined(NLIB_NEON)
    uint32x4_t tmp = vld1q_u32(reinterpret_cast<const uint32_t*>(p));
    return vreinterpretq_s8_u32(tmp);
#endif
}

// r[i] = p[i], p is 2 bytes aligned
NLIB_M(i128) I128::LoadA2(const void* p) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_loadu_si128(reinterpret_cast<const __m128i*>(p));
#elif defined(NLIB_NEON)
    uint16x8_t tmp = vld1q_u16(reinterpret_cast<const uint16_t*>(p));
    return vreinterpretq_s8_u16(tmp);
#endif
}

// r[i] = p[i]
NLIB_M(i128) I128::LoadA1(const void* p) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_loadu_si128(reinterpret_cast<const __m128i*>(p));
#elif defined(NLIB_NEON)
    return vld1q_s8(reinterpret_cast<const int8_t*>(p));
#endif
}

NLIB_M(i128) I128::LoadLoA8(const void* p) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_loadl_epi64(reinterpret_cast<const __m128i*>(p));
#elif defined(NLIB_NEON)
    int8x8_t lo = vreinterpret_s8_u64(vld1_u64(reinterpret_cast<const uint64_t*>(p)));
    return vcombine_s8(lo, vdup_n_s8(0));
#endif
}

NLIB_M(i128) I128::LoadLoA4(const void* p) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_loadl_epi64(reinterpret_cast<const __m128i*>(p));
#elif defined(NLIB_NEON)
    int8x8_t lo = vreinterpret_s8_u32(vld1_u32(reinterpret_cast<const uint32_t*>(p)));
    return vcombine_s8(lo, vdup_n_s8(0));
#endif
}

NLIB_M(i128) I128::LoadLoA2(const void* p) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_loadl_epi64(reinterpret_cast<const __m128i*>(p));
#elif defined(NLIB_NEON)
    int8x8_t lo = vreinterpret_s8_u16(vld1_u16(reinterpret_cast<const uint16_t*>(p)));
    return vcombine_s8(lo, vdup_n_s8(0));
#endif
}

NLIB_M(i128) I128::LoadLoA1(const void* p) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_loadl_epi64(reinterpret_cast<const __m128i*>(p));
#elif defined(NLIB_NEON)
    int8x8_t lo = vld1_s8(reinterpret_cast<const int8_t*>(p));
    return vcombine_s8(lo, vdup_n_s8(0));
#endif
}

NLIB_M(i128) I128::LoadHiA8(const void* p) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    __m128i tmp = _mm_loadl_epi64(reinterpret_cast<const __m128i*>(p));
    return _mm_shuffle_epi32(tmp, _MM_SHUFFLE(1, 0, 3, 2));
#elif defined(NLIB_NEON)
    int8x8_t hi = vreinterpret_s8_u64(vld1_u64(reinterpret_cast<const uint64_t*>(p)));
    return vcombine_s8(vdup_n_s8(0), hi);
#endif
}

NLIB_M(i128) I128::LoadHiA4(const void* p) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    __m128i tmp = _mm_loadl_epi64(reinterpret_cast<const __m128i*>(p));
    return _mm_shuffle_epi32(tmp, _MM_SHUFFLE(1, 0, 3, 2));
#elif defined(NLIB_NEON)
    int8x8_t hi = vreinterpret_s8_u32(vld1_u32(reinterpret_cast<const uint32_t*>(p)));
    return vcombine_s8(vdup_n_s8(0), hi);
#endif
}

NLIB_M(i128) I128::LoadHiA2(const void* p) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    __m128i tmp = _mm_loadl_epi64(reinterpret_cast<const __m128i*>(p));
    return _mm_shuffle_epi32(tmp, _MM_SHUFFLE(1, 0, 3, 2));
#elif defined(NLIB_NEON)
    int8x8_t hi = vreinterpret_s8_u16(vld1_u16(reinterpret_cast<const uint16_t*>(p)));
    return vcombine_s8(vdup_n_s8(0), hi);
#endif
}

NLIB_M(i128) I128::LoadHiA1(const void* p) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    __m128i tmp = _mm_loadl_epi64(reinterpret_cast<const __m128i*>(p));
    return _mm_shuffle_epi32(tmp, _MM_SHUFFLE(1, 0, 3, 2));
#elif defined(NLIB_NEON)
    int8x8_t hi = vld1_s8(reinterpret_cast<const int8_t*>(p));
    return vcombine_s8(vdup_n_s8(0), hi);
#endif
}

// p[i] = value[i], p is 16 bytes aligned
NLIB_M(void) I128::StoreA16(void* p, i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    _mm_store_si128(reinterpret_cast<i128*>(p), value);
#elif defined(NLIB_NEON)
    vst1q_u64(reinterpret_cast<uint64_t*>(p), vreinterpretq_u64_s8(value));
#endif
}

// p[i] = value[i], p is 8 bytes aligned
NLIB_M(void) I128::StoreA8(void* p, i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    _mm_storeu_si128(reinterpret_cast<i128*>(p), value);
#elif defined(NLIB_NEON)
    vst1q_u64(reinterpret_cast<uint64_t*>(p), vreinterpretq_u64_s8(value));
#endif
}

// p[i] = value[i], p is 4 bytes aligned
NLIB_M(void) I128::StoreA4(void* p, i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    _mm_storeu_si128(reinterpret_cast<i128*>(p), value);
#elif defined(NLIB_NEON)
    vst1q_u32(reinterpret_cast<uint32_t*>(p), vreinterpretq_u32_s8(value));
#endif
}

// p[i] = value[i], p is 2 bytes aligned
NLIB_M(void) I128::StoreA2(void* p, i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    _mm_storeu_si128(reinterpret_cast<i128*>(p), value);
#elif defined(NLIB_NEON)
    vst1q_u16(reinterpret_cast<uint16_t*>(p), vreinterpretq_u16_s8(value));
#endif
}

// p[i] = value[i]
NLIB_M(void) I128::StoreA1(void* p, i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    _mm_storeu_si128(reinterpret_cast<i128*>(p), value);
#elif defined(NLIB_NEON)
    vst1q_s8(reinterpret_cast<int8_t*>(p), value);
#endif
}

NLIB_M(void) I128::StoreLoA8(void* p, i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    _mm_storel_epi64(reinterpret_cast<i128*>(p), value);
#elif defined(NLIB_NEON)
    uint64x1_t x = vreinterpret_u64_s8(vget_low_s8(value));
    vst1_u64(reinterpret_cast<uint64_t*>(p), x);
#endif
}

NLIB_M(void) I128::StoreLoA4(void* p, i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    _mm_storel_epi64(reinterpret_cast<i128*>(p), value);
#elif defined(NLIB_NEON)
    uint32x2_t x = vreinterpret_u32_s8(vget_low_s8(value));
    vst1_u32(reinterpret_cast<uint32_t*>(p), x);
#endif
}

NLIB_M(void) I128::StoreLoA2(void* p, i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    _mm_storel_epi64(reinterpret_cast<i128*>(p), value);
#elif defined(NLIB_NEON)
    uint16x4_t x = vreinterpret_u16_s8(vget_low_s8(value));
    vst1_u16(reinterpret_cast<uint16_t*>(p), x);
#endif
}

NLIB_M(void) I128::StoreLoA1(void* p, i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    _mm_storel_epi64(reinterpret_cast<i128*>(p), value);
#elif defined(NLIB_NEON)
    int8x8_t x = vget_low_s8(value);
    vst1_s8(reinterpret_cast<int8_t*>(p), x);
#endif
}

NLIB_M(void) I128::StoreHiA8(void* p, i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    _mm_storel_epi64(reinterpret_cast<i128*>(p),
                     _mm_shuffle_epi32(value, _MM_SHUFFLE(1, 0, 3, 2)));
#elif defined(NLIB_NEON)
    uint64x1_t x = vreinterpret_u64_s8(vget_high_s8(value));
    vst1_u64(reinterpret_cast<uint64_t*>(p), x);
#endif
}

NLIB_M(void) I128::StoreHiA4(void* p, i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    _mm_storel_epi64(reinterpret_cast<i128*>(p),
                     _mm_shuffle_epi32(value, _MM_SHUFFLE(1, 0, 3, 2)));
#elif defined(NLIB_NEON)
    uint32x2_t x = vreinterpret_u32_s8(vget_high_s8(value));
    vst1_u32(reinterpret_cast<uint32_t*>(p), x);
#endif
}

NLIB_M(void) I128::StoreHiA2(void* p, i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    _mm_storel_epi64(reinterpret_cast<i128*>(p),
                     _mm_shuffle_epi32(value, _MM_SHUFFLE(1, 0, 3, 2)));
#elif defined(NLIB_NEON)
    uint16x4_t x = vreinterpret_u16_s8(vget_high_s8(value));
    vst1_u16(reinterpret_cast<uint16_t*>(p), x);
#endif
}

NLIB_M(void) I128::StoreHiA1(void* p, i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    _mm_storel_epi64(reinterpret_cast<i128*>(p),
                     _mm_shuffle_epi32(value, _MM_SHUFFLE(1, 0, 3, 2)));
#elif defined(NLIB_NEON)
    int8x8_t x = vget_high_s8(value);
    vst1_s8(reinterpret_cast<int8_t*>(p), x);
#endif
}

template <size_t N>
// r = value.u8[N]
NLIB_M(uint8_t) I128::GetUint8FromLane(i128arg value) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N < 16);
#if defined(NLIB_SSE41)
    return static_cast<uint8_t>(_mm_extract_epi8(value, N));
#elif defined(NLIB_NEON)
    return vgetq_lane_u8(vreinterpretq_u8_s8(value), N);
#endif
}

template <size_t N>
// r = value.u16[N]
NLIB_M(uint16_t) I128::GetUint16FromLane(i128arg value) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N < 8);
#if defined(NLIB_SSE41)
    return static_cast<uint16_t>(_mm_extract_epi16(value, N));
#elif defined(NLIB_NEON)
    return vgetq_lane_u16(vreinterpretq_u16_s8(value), N);
#endif
}

template <size_t N>
// r = value.u32[N]
NLIB_M(uint32_t) I128::GetUint32FromLane(i128arg value) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N < 4);
#if defined(NLIB_SSE41)
    return static_cast<uint32_t>(_mm_extract_epi32(value, N));
#elif defined(NLIB_NEON)
    return vgetq_lane_u32(vreinterpretq_u32_s8(value), N);
#endif
}

template <size_t N>
// r = value.u64[N]
NLIB_M(uint64_t) I128::GetUint64FromLane(i128arg value) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N < 2);
#if defined(NLIB_SSE41)
#ifdef NLIB_64BIT
    return static_cast<uint64_t>(_mm_extract_epi64(value, N));
#else
    NLIB_UNUSED(value);
    return 0;  // dummy
#endif
#elif defined(NLIB_NEON)
    return vgetq_lane_u64(vreinterpretq_u64_s8(value), N);
#endif
}

#if defined(NLIB_SSE41) && !defined(NLIB_64BIT)
template <>
NLIB_M(uint64_t) I128::GetUint64FromLane<0>(i128arg value) NLIB_NOEXCEPT {
    uint64_t rval;
    _mm_storel_epi64(reinterpret_cast<i128*>(&rval), value);
    return rval;
}
template <>
NLIB_M(uint64_t) I128::GetUint64FromLane<1>(i128arg value) NLIB_NOEXCEPT {
    uint64_t rval;
    i128 tmp = _mm_shuffle_epi32(value, _MM_SHUFFLE(1, 0, 3, 2));
    _mm_storel_epi64(reinterpret_cast<i128*>(&rval), tmp);
    return rval;
}
#endif

template <size_t N>
// r = value, r.u8[N] = v
NLIB_M(i128) I128::SetUint8ToLane(i128arg value, uint8_t v) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N < 16);
#if defined(NLIB_SSE41)
    return _mm_insert_epi8(value, static_cast<int8_t>(v), N);
#elif defined(NLIB_NEON)
    return __builtin_constant_p(v) ?
        I128::Permute8<
            N == 0 ? 16 : 0,
            N == 1 ? 17 : 1,
            N == 2 ? 18 : 2,
            N == 3 ? 19 : 3,
            N == 4 ? 20 : 4,
            N == 5 ? 21 : 5,
            N == 6 ? 22 : 6,
            N == 7 ? 23 : 7,
            N == 8 ? 24 : 8,
            N == 9 ? 25 : 9,
            N == 10 ? 26 : 10,
            N == 11 ? 27 : 11,
            N == 12 ? 28 : 12,
            N == 13 ? 29 : 13,
            N == 14 ? 30 : 14,
            N == 15 ? 31 : 15>(value, vreinterpretq_s8_u8(vdupq_n_u8(v))) :
        vreinterpretq_s8_u8(vsetq_lane_u8(v, vreinterpretq_u8_s8(value), N));
#endif
}

template <size_t N>
// r = value, r.u16[N] = v
NLIB_M(i128) I128::SetUint16ToLane(i128arg value, uint16_t v) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N < 8);
#if defined(NLIB_SSE41)
    return _mm_insert_epi16(value, static_cast<int16_t>(v), N);
#elif defined(NLIB_NEON)
    return __builtin_constant_p(v) ?
        I128::Permute16<
            N == 0 ? 8 : 0,
            N == 1 ? 9 : 1,
            N == 2 ? 10 : 2,
            N == 3 ? 11 : 3,
            N == 4 ? 12 : 4,
            N == 5 ? 13 : 5,
            N == 6 ? 14 : 6,
            N == 7 ? 15 : 7>(value, vreinterpretq_s8_u16(vdupq_n_u16(v))) :
        vreinterpretq_s8_u16(vsetq_lane_u16(v, vreinterpretq_u16_s8(value), N));
#endif
}

template <size_t N>
// r = value, r.u32[N] = v
NLIB_M(i128) I128::SetUint32ToLane(i128arg value, uint32_t v) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N < 4);
#if defined(NLIB_SSE41)
    return _mm_insert_epi32(value, static_cast<uint32_t>(v), N);
#elif defined(NLIB_NEON)
    return __builtin_constant_p(v) ?
        I128::Permute32<N == 0 ? 4 : 0,
                        N == 1 ? 5 : 1,
                        N == 2 ? 6 : 2,
                        N == 3 ? 7 : 3>(value, vreinterpretq_s8_u32(vdupq_n_u32(v))) :
        vreinterpretq_s8_u32(vsetq_lane_u32(v, vreinterpretq_u32_s8(value), N));
#endif
}

template <size_t N>
// r = value, r.u64[N] = v
NLIB_M(i128) I128::SetUint64ToLane(i128arg value, uint64_t v) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N < 2);
#if defined(NLIB_SSE41)
#ifdef NLIB_64BIT
    return _mm_insert_epi64(value, static_cast<int64_t>(v), N);
#else
    union {
        int32_t i32[2];
        int64_t i64;
    } tmp;
    tmp.i64 = static_cast<int64_t>(v);
    __m128i rval;
    rval = _mm_insert_epi32(value, tmp.i32[0], N * 2 + 0);
    return _mm_insert_epi32(rval, tmp.i32[1], N * 2 + 1);
#endif
#elif defined(NLIB_NEON)
    return vreinterpretq_s8_u64(vsetq_lane_u64(v, vreinterpretq_u64_s8(value), N));
#endif
}

// r.i8[i] = a.i8[i] + b.i8[i]
NLIB_M(i128) I128::Add8(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_add_epi8(a, b);
#elif defined(NLIB_NEON)
    return vaddq_s8(a, b);
#endif
}

// r.i16[i] = a.i16[i] + b.i16[i]
NLIB_M(i128) I128::Add16(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_add_epi16(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vaddq, s16, a, b);
#endif
}

// r.i32[i] = a.i32[i] + b.i32[i]
NLIB_M(i128) I128::Add32(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_add_epi32(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vaddq, s32, a, b);
#endif
}

// r.i64[i] = a.i64[i] + b.i64[i]
NLIB_M(i128) I128::Add64(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_add_epi64(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vaddq, s64, a, b);
#endif
}

// 127 if a[i] + b[i] > 127, -128 if a[i] + b[i] < -128
NLIB_M(i128) I128::AddInt8Saturated(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_adds_epi8(a, b);
#elif defined(NLIB_NEON)
    return vqaddq_s8(a, b);
#endif
}

// 32767 if a[i] + b[i] > 32767, -32768 if a[i] + b[i] < -32768
NLIB_M(i128) I128::AddInt16Saturated(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_adds_epi16(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vqaddq, s16, a, b);
#endif
}

// 255 if a[i] + b[i] > 255
NLIB_M(i128) I128::AddUint8Saturated(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_adds_epu8(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vqaddq, u8, a, b);
#endif
}

// 65535 if a[i] + b[i] > 65535
NLIB_M(i128) I128::AddUint16Saturated(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_adds_epu16(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vqaddq, u16, a, b);
#endif
}

// r.i8[i] = a.i8[i] - b.i8[i]
NLIB_M(i128) I128::Sub8(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_sub_epi8(a, b);
#elif defined(NLIB_NEON)
    return vsubq_s8(a, b);
#endif
}

// r.i16[i] = a.i16[i] - b.i16[i]
NLIB_M(i128) I128::Sub16(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_sub_epi16(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vsubq, s16, a, b);
#endif
}

// r.i32[i] = a.i32[i] - b.i32[i]
NLIB_M(i128) I128::Sub32(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_sub_epi32(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vsubq, s32, a, b);
#endif
}

// r.i64[0] = a.i64[0] - b.i64[0]
NLIB_M(i128) I128::Sub64(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_sub_epi64(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vsubq, s64, a, b);
#endif
}

// 127 if a[i] - b[i] > 127, -128 if a[i] - b[i] < -128
NLIB_M(i128) I128::SubInt8Saturated(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_subs_epi8(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vqsubq, s8, a, b);
#endif
}

// 32767 if a[i] - b[i] > 32767, -32768 if a[i] - b[i] < -32768
NLIB_M(i128) I128::SubInt16Saturated(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_subs_epi16(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vqsubq, s16, a, b);
#endif
}

// 0 if a.u8[i] - b.u8[i] < 0
NLIB_M(i128) I128::SubUint8Saturated(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_subs_epu8(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vqsubq, u8, a, b);
#endif
}

// 0 if a.u16[i] - b.u16[i] < 0
NLIB_M(i128) I128::SubUint16Saturated(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_subs_epu16(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vqsubq, u16, a, b);
#endif
}

// r.i8[0] = value[0] + value[1], ..., r.i8[7] = value[14] + value[15]
NLIB_M(i128) I128::PairwiseAdd8(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    __m128i ax = _mm_add_epi8(a, _mm_srli_epi16(a, 8));
    __m128i bx = _mm_add_epi8(b, _mm_srli_epi16(b, 8));
    return I128::NarrowFrom16To8(ax, bx);
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    return vpaddq_s8(a, b);
# else
    int8x8_t al = vget_low_s8(a);
    int8x8_t ah = vget_high_s8(a);
    int8x8_t rl = vpadd_s8(al, ah);
    int8x8_t bl = vget_low_s8(b);
    int8x8_t bh = vget_high_s8(b);
    int8x8_t rh = vpadd_s8(bl, bh);
    return vcombine_s8(rl, rh);
# endif
#endif
}

// r.i16[0] = value[0] + value[1], ..., r.i16[3] = value[6] + value[7]
NLIB_M(i128) I128::PairwiseAdd16(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_hadd_epi16(a, b);
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    return vreinterpretq_s8_s16(vpaddq_s16(vreinterpretq_s16_s8(a), vreinterpretq_s16_s8(b)));
# else
    int16x4_t al = vget_low_s16(vreinterpretq_s16_s8(a));
    int16x4_t ah = vget_high_s16(vreinterpretq_s16_s8(a));
    int16x4_t rl = vpadd_s16(al, ah);
    int16x4_t bl = vget_low_s16(vreinterpretq_s16_s8(b));
    int16x4_t bh = vget_high_s16(vreinterpretq_s16_s8(b));
    int16x4_t rh = vpadd_s16(bl, bh);
    return NLIB_CMB(s16, rl, rh);
# endif
#endif
}

// r.i32[0] = value[0] + value[1], r.i32[1] = value[2] + value[3]
NLIB_M(i128) I128::PairwiseAdd32(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_hadd_epi32(a, b);
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    return vreinterpretq_s8_s32(vpaddq_s32(vreinterpretq_s32_s8(a), vreinterpretq_s32_s8(b)));
# else
    int32x2_t al = vget_low_s32(vreinterpretq_s32_s8(a));
    int32x2_t ah = vget_high_s32(vreinterpretq_s32_s8(a));
    int32x2_t rl = vpadd_s32(al, ah);
    int32x2_t bl = vget_low_s32(vreinterpretq_s32_s8(b));
    int32x2_t bh = vget_high_s32(vreinterpretq_s32_s8(b));
    int32x2_t rh = vpadd_s32(bl, bh);
    return NLIB_CMB(s32, rl, rh);
# endif
#endif
}

// r.i16[i] = a.i16[i] * b.i16[i]
NLIB_M(i128) I128::Mult16(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_mullo_epi16(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vmulq, s16, a, b);
#endif
}

// r = c + a * b
NLIB_M(i128) I128::MultAdd16(i128arg a, i128arg b, i128arg c) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_add_epi16(c, _mm_mullo_epi16(a, b));
#elif defined(NLIB_NEON)
    return NLIB_OP3(vmlaq, s16, c, a, b);
#endif
}

// r = c - a * b
NLIB_M(i128) I128::MultSub16(i128arg a, i128arg b, i128arg c) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_sub_epi16(c, _mm_mullo_epi16(a, b));
#elif defined(NLIB_NEON)
    return NLIB_OP3(vmlsq, s16, c, a, b);
#endif
}

// r.i32[i] = a.i32[i] * b.i32[i]
NLIB_M(i128) I128::Mult32(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_mullo_epi32(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vmulq, s32, a, b);
#endif
}

// r = c + a * b
NLIB_M(i128) I128::MultAdd32(i128arg a, i128arg b, i128arg c) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_add_epi32(c, _mm_mullo_epi32(a, b));
#elif defined(NLIB_NEON)
    return NLIB_OP3(vmlaq, s32, c, a, b);
#endif
}

// r = c - a * b
NLIB_M(i128) I128::MultSub32(i128arg a, i128arg b, i128arg c) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_sub_epi32(c, _mm_mullo_epi32(a, b));
#elif defined(NLIB_NEON)
    return NLIB_OP3(vmlsq, s32, c, a, b);
#endif
}

// r.s8[i] = max(a.s8[i], b.s8[i])
NLIB_M(i128) I128::MaxInt8(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_max_epi8(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vmaxq, s8, a, b);
#endif
}

// r.s16[i] = max(a.s16[i], b.s16[i])
NLIB_M(i128) I128::MaxInt16(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_max_epi16(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vmaxq, s16, a, b);
#endif
}

// r.s32[i] = max(a.s32[i], b.s32[i])
NLIB_M(i128) I128::MaxInt32(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_max_epi32(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vmaxq, s32, a, b);
#endif
}

// r.u8[i] = max(a.u8[i], b.u8[i])
NLIB_M(i128) I128::MaxUint8(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_max_epu8(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vmaxq, u8, a, b);
#endif
}

// r.u16[i] = max(a.u16[i], b.u16[i])
NLIB_M(i128) I128::MaxUint16(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_max_epu16(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vmaxq, u16, a, b);
#endif
}

// r.u32[i] = max(a.u32[i], b.u32[i])
NLIB_M(i128) I128::MaxUint32(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_max_epu32(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vmaxq, u32, a, b);
#endif
}

// r.s8[i] = min(a.s8[i], b.s8[i])
NLIB_M(i128) I128::MinInt8(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_min_epi8(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vminq, s8, a, b);
#endif
}

// r.s16[i] = min(a.s16[i], b.s16[i])
NLIB_M(i128) I128::MinInt16(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_min_epi16(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vminq, s16, a, b);
#endif
}

// r.s32[i] = min(a.s32[i], b.s32[i])
NLIB_M(i128) I128::MinInt32(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_min_epi32(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vminq, s32, a, b);
#endif
}

// r.u8[i] = min(a.u8[i], b.u8[i])
NLIB_M(i128) I128::MinUint8(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_min_epu8(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vminq, u8, a, b);
#endif
}

// r.u16[i] = min(a.u16[i], b.u16[i])
NLIB_M(i128) I128::MinUint16(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_min_epu16(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vminq, u16, a, b);
#endif
}

// r.u32[i] = min(a.u32[i], b.u32[i])
NLIB_M(i128) I128::MinUint32(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_min_epu32(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vminq, u32, a, b);
#endif
}

// r.s8[i] = abs(value.s8[i])
NLIB_M(i128) I128::AbsInt8(i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_abs_epi8(value);
#elif defined(NLIB_NEON)
    return NLIB_OP1(vabsq, s8, value);
#endif
}

// r.s16[i] = abs(value.s16[i])
NLIB_M(i128) I128::AbsInt16(i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_abs_epi16(value);
#elif defined(NLIB_NEON)
    return NLIB_OP1(vabsq, s16, value);
#endif
}

// r.s32[i] = abs(value.s32[i])
NLIB_M(i128) I128::AbsInt32(i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_abs_epi32(value);
#elif defined(NLIB_NEON)
    return NLIB_OP1(vabsq, s32, value);
#endif
}

// r.s8[i] = abs(a.s8[i] - b.s8[i])
NLIB_M(i128) I128::AbsDiffInt8(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_abs_epi8(_mm_sub_epi8(a, b));
#elif defined(NLIB_NEON)
    return NLIB_OP2(vabdq, s8, a, b);
#endif
}

// r.s16[i] = abs(a.s16[i] - b.s16[i])
NLIB_M(i128) I128::AbsDiffInt16(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_abs_epi16(_mm_sub_epi16(a, b));
#elif defined(NLIB_NEON)
    return NLIB_OP2(vabdq, s16, a, b);
#endif
}

// r.s32[i] = abs(a.s32[i] - b.s32[i])
NLIB_M(i128) I128::AbsDiffInt32(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_abs_epi32(_mm_sub_epi32(a, b));
#elif defined(NLIB_NEON)
    return NLIB_OP2(vabdq, s32, a, b);
#endif
}

// r.s8[i] = -value.s8[i]
NLIB_M(i128) I128::NegateInt8(i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_sub_epi8(_mm_setzero_si128(), value);
#elif defined(NLIB_NEON)
    return NLIB_OP1(vnegq, s8, value);
#endif
}

// r.s16[i] = -value.s16[i]
NLIB_M(i128) I128::NegateInt16(i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_sub_epi16(_mm_setzero_si128(), value);
#elif defined(NLIB_NEON)
    return NLIB_OP1(vnegq, s16, value);
#endif
}

// r.s32[i] = -value.s32[i]
NLIB_M(i128) I128::NegateInt32(i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_sub_epi32(_mm_setzero_si128(), value);
#elif defined(NLIB_NEON)
    return NLIB_OP1(vnegq, s32, value);
#endif
}

// r = a & b
NLIB_M(i128) I128::And(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_and_si128(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vandq, s8, a, b);
#endif
}

// r = a | b
NLIB_M(i128) I128::Or(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_or_si128(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vorrq, s8, a, b);
#endif
}

// r = a ^ b
NLIB_M(i128) I128::Xor(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_xor_si128(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(veorq, s8, a, b);
#endif
}

// r = ~a
NLIB_M(i128) I128::Not(i128arg a) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_andnot_si128(a, _mm_cmpeq_epi8(a, a));
#elif defined(NLIB_NEON)
    return NLIB_OP1(vmvnq, s8, a);
#endif
}

// r = ~a & b
NLIB_M(i128) I128::AndNot(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_andnot_si128(a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vbicq, s8, b, a);
#endif
}

// r = ~a | b
NLIB_M(i128) I128::OrNot(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    __m128i not_a = _mm_andnot_si128(a, _mm_cmpeq_epi8(a, a));
    return _mm_or_si128(not_a, b);
#elif defined(NLIB_NEON)
    return NLIB_OP2(vornq, s8, b, a);
#endif
}

NLIB_M(i128) I128::Test8(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_NEON)
    return vtstq_s8(a, b);
#else
    return I128::Not(I128::CmpEqZero8(I128::And(a, b)));
#endif
}

NLIB_M(i128) I128::Test16(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_NEON)
    return NLIB_OP2(vtstq, s16, a, b);
#else
    return I128::Not(I128::CmpEqZero16(I128::And(a, b)));
#endif
}

NLIB_M(i128) I128::Test32(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_NEON)
    return NLIB_OP2(vtstq, s32, a, b);
#else
    return I128::Not(I128::CmpEqZero32(I128::And(a, b)));
#endif
}

// r.i8[i] = a.i8[i] == b.i8[i] ? 0xFF : 0
NLIB_M(i128) I128::CmpEq8(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cmpeq_epi8(a, b);
#elif defined(NLIB_NEON)
    return NLIB_CMP(vceqq, s8, a, b, u8);
#endif
}

// r.i16[i] = a.i16[i] == b.i16[i] ? 0xFFFF : 0
NLIB_M(i128) I128::CmpEq16(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cmpeq_epi16(a, b);
#elif defined(NLIB_NEON)
    return NLIB_CMP(vceqq, s16, a, b, u16);
#endif
}

// r.i32[i] = a.i32[i] == b.i32[i] ? 0xFFFFFFFF : 0
NLIB_M(i128) I128::CmpEq32(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cmpeq_epi32(a, b);
#elif defined(NLIB_NEON)
    return NLIB_CMP(vceqq, s32, a, b, u32);
#endif
}

// r.i64[i] = a.i64[i] == b.i64[i] ? 0xFFFFFFFFFFFFFFFF : 0
NLIB_M(i128) I128::CmpEq64(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cmpeq_epi64(a, b);
#elif defined(NLIB_NEON)
#ifdef __aarch64__
    return NLIB_CMP(vceqq, s64, a, b, u64);
#else
    uint32x4_t x0 = vceqq_u32(vreinterpretq_u32_s8(a), vreinterpretq_u32_s8(b));
    uint32x2x2_t x1 = vtrn_u32(vget_low_u32(x0), vget_high_u32(x0));
    uint32x2_t x2 = vand_u32(x1.val[0], x1.val[1]);
    int64x2_t result = vmovl_s32(vreinterpret_s32_u32(x2));
    return vreinterpretq_s8_s64(result);
#endif
#endif
}

// r.s8[i] = a.s8[i] < b.s8[i] ? 0xFF : 0
NLIB_M(i128) I128::CmpLtInt8(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cmplt_epi8(a, b);
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcltq, s8, a, b, u8);
#endif
}

// r.s16[i] = a.s16[i] < b.s16[i] ? 0xFFFF : 0
NLIB_M(i128) I128::CmpLtInt16(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cmplt_epi16(a, b);
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcltq, s16, a, b, u16);
#endif
}

// r.s32[i] = a.s32[i] < b.s32[i] ? 0xFFFFFFFF : 0
NLIB_M(i128) I128::CmpLtInt32(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cmplt_epi32(a, b);
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcltq, s32, a, b, u32);
#endif
}

// r.s64[i] = a.s64[i] < b.s64[i] ? 0xFFFFFFFFFFFFFFFF : 0
NLIB_M(i128) I128::CmpLtInt64(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE42)
    return _mm_cmpgt_epi64(b, a);
#elif defined(NLIB_NEON)
#ifdef __aarch64__
    return NLIB_CMP(vcltq, s64, a, b, u64);
#else
    int32x2x2_t trn_a = vtrn_s32(vreinterpret_s32_s8(vget_low_s8(a)),
                                 vreinterpret_s32_s8(vget_high_s8(a)));
    int32x2x2_t trn_b = vtrn_s32(vreinterpret_s32_s8(vget_low_s8(b)),
                                 vreinterpret_s32_s8(vget_high_s8(b)));
    uint32x2_t upper_lt = vclt_s32(trn_a.val[1], trn_b.val[1]);
    uint32x2_t upper_eq = vceq_s32(trn_a.val[1], trn_b.val[1]);
    uint32x2_t lower_lt = vclt_u32(trn_a.val[0], trn_b.val[0]);
    uint32x2_t x2 = vorr_u32(upper_lt, vand_u32(upper_eq, lower_lt));
    int64x2_t result = vmovl_s32(vreinterpret_s32_u32(x2));
    return vreinterpretq_s8_s64(result);
#endif
#else
    i128 cmp = I128::CmpLtInt32(a, b);
    i128 eq = I128::CmpEq32(a, b);
    i128 cmp_lt = I128::CmpLtUint32(a, b);
    i128 upper_lt = I128::Permute32<1, 1, 3, 3>(cmp, cmp);
    i128 lower_lt = I128::Permute32<0, 0, 2, 2>(cmp_lt, cmp_lt);
    i128 upper_eq = I128::Permute32<1, 1, 3, 3>(eq, eq);
    return I128::Or(upper_lt, I128::And(upper_eq, lower_lt));
#endif
}

// r.s8[i] = a.s8[i] > b.s8[i] ? 0xFF : 0
NLIB_M(i128) I128::CmpGtInt8(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cmpgt_epi8(a, b);
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcgtq, s8, a, b, u8);
#endif
}

// r.s16[i] = a.s16[i] > b.s16[i] ? 0xFFFF : 0
NLIB_M(i128) I128::CmpGtInt16(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cmpgt_epi16(a, b);
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcgtq, s16, a, b, u16);
#endif
}

// r.s32[i] = a.s32[i] > b.s32[i] ? 0xFFFFFFFF : 0
NLIB_M(i128) I128::CmpGtInt32(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cmpgt_epi32(a, b);
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcgtq, s32, a, b, u32);
#endif
}

// r.s64[i] = a.s64[i] > b.s64[i] ? 0xFFFFFFFFFFFFFFFF : 0
NLIB_M(i128) I128::CmpGtInt64(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE42)
    return _mm_cmpgt_epi64(a, b);
#elif defined(NLIB_NEON) && defined(__aarch64__)
    return NLIB_CMP(vcgtq, s64, a, b, u64);
#else
    return I128::CmpLtInt64(b, a);
#endif
}

// r.u8[i] = a.u8[i] < b.u8[i] ? 0xFF : 0
NLIB_M(i128) I128::CmpLtUint8(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    i128 ofs = I128::SetValue(0x80, each_uint8);
    return _mm_cmplt_epi8(_mm_add_epi8(a, ofs), _mm_add_epi8(b, ofs));
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcltq, u8, a, b, u8);
#endif
}

// r.u8[i] = a.u8[i] > b.u8[i] ? 0xFF : 0
NLIB_M(i128) I128::CmpGtUint8(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    i128 ofs = I128::SetValue(0x80, each_uint8);
    return _mm_cmpgt_epi8(_mm_add_epi8(a, ofs), _mm_add_epi8(b, ofs));
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcgtq, u8, a, b, u8);
#endif
}

// r.u16[i] = a.u16[i] < b.u16[i] ? 0xFFFF : 0
NLIB_M(i128) I128::CmpLtUint16(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    i128 ofs = I128::SetValue(0x8000U, each_uint16);
    return _mm_cmplt_epi16(_mm_add_epi16(a, ofs), _mm_add_epi16(b, ofs));
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcltq, u16, a, b, u16);
#endif
}

// r.u16[i] = a.u16[i] > b.u16[i] ? 0xFFFF : 0
NLIB_M(i128) I128::CmpGtUint16(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    i128 ofs = I128::SetValue(0x8000U, each_uint16);
    return _mm_cmpgt_epi16(_mm_add_epi16(a, ofs), _mm_add_epi16(b, ofs));
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcgtq, u16, a, b, u16);
#endif
}

// r.u32[i] = a.u32[i] < b.u32[i] ? 0xFFFFFFFF : 0
NLIB_M(i128) I128::CmpLtUint32(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    i128 ofs = I128::SetValue(0x80000000U, each_uint32);
    return _mm_cmplt_epi32(_mm_add_epi32(a, ofs), _mm_add_epi32(b, ofs));
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcltq, u32, a, b, u32);
#endif
}

// r.u32[i] = a.u32[i] > b.u32[i] ? 0xFFFFFFFF : 0
NLIB_M(i128) I128::CmpGtUint32(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    i128 ofs = I128::SetValue(0x80000000U, each_uint32);
    return _mm_cmpgt_epi32(_mm_add_epi32(a, ofs), _mm_add_epi32(b, ofs));
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcgtq, u32, a, b, u32);
#endif
}

// r.u64[i] = a.u64[i] < b.u64[i] ? 0xFFFFFFFFFFFFFFFF : 0
NLIB_M(i128) I128::CmpLtUint64(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE42)
    i128 ofs = I128::SetValue(0x8000000000000000ULL, each_uint64);
    return _mm_cmpgt_epi64(_mm_add_epi64(b, ofs), _mm_add_epi64(a, ofs));
#elif defined(NLIB_NEON)
#ifdef __aarch64__
    return NLIB_CMP(vcltq, u64, a, b, u64);
#else
    uint32x2x2_t trn_a = vtrn_u32(vreinterpret_u32_s8(vget_low_s8(a)),
                                  vreinterpret_u32_s8(vget_high_s8(a)));
    uint32x2x2_t trn_b = vtrn_u32(vreinterpret_u32_s8(vget_low_s8(b)),
                                  vreinterpret_u32_s8(vget_high_s8(b)));
    uint32x2_t upper_lt = vclt_u32(trn_a.val[1], trn_b.val[1]);
    uint32x2_t upper_eq = vceq_u32(trn_a.val[1], trn_b.val[1]);
    uint32x2_t lower_lt = vclt_u32(trn_a.val[0], trn_b.val[0]);
    uint32x2_t x2 = vorr_u32(upper_lt, vand_u32(upper_eq, lower_lt));
    int64x2_t result = vmovl_s32(vreinterpret_s32_u32(x2));
    return vreinterpretq_s8_s64(result);
#endif
#else
    i128 cmp = I128::CmpLtUint32(a, b);
    i128 eq = I128::CmpEq32(a, b);
    i128 upper_lt = I128::Permute32<1, 1, 3, 3>(cmp, cmp);
    i128 lower_lt = I128::Permute32<0, 0, 2, 2>(cmp, cmp);
    i128 upper_eq = I128::Permute32<1, 1, 3, 3>(eq, eq);
    return I128::Or(upper_lt, I128::And(upper_eq, lower_lt));
#endif
}

// r.u64[i] = a.u64[i] > b.u64[i] ? 0xFFFFFFFFFFFFFFFF : 0
NLIB_M(i128) I128::CmpGtUint64(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE42)
    i128 ofs = I128::SetValue(0x8000000000000000ULL, each_uint64);
    return _mm_cmpgt_epi64(_mm_add_epi64(a, ofs), _mm_add_epi64(b, ofs));
#elif defined(NLIB_NEON) && defined(__aarch64__)
    return NLIB_CMP(vcgtq, u64, a, b, u64);
#else
    return I128::CmpLtUint64(b, a);
#endif
}

// r.s8[i] = a.s8[i] <= b.s8[i] ? 0xFF : 0
NLIB_M(i128) I128::CmpLeInt8(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_or_si128(_mm_cmplt_epi8(a, b), _mm_cmpeq_epi8(a, b));
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcleq, s8, a, b, u8);
#endif
}

// r.s16[i] = a.s16[i] <= b.s16[i] ? 0xFFFF : 0
NLIB_M(i128) I128::CmpLeInt16(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_or_si128(_mm_cmplt_epi16(a, b), _mm_cmpeq_epi16(a, b));
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcleq, s16, a, b, u16);
#endif
}

// r.s32[i] = a.s32[i] <= b.s32[i] ? 0xFFFFFFFF : 0
NLIB_M(i128) I128::CmpLeInt32(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_or_si128(_mm_cmplt_epi32(a, b), _mm_cmpeq_epi32(a, b));
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcleq, s32, a, b, u32);
#endif
}

// r.s32[i] = a.s32[i] <= b.s32[i] ? 0xFFFFFFFFFFFFFFFF : 0
NLIB_M(i128) I128::CmpLeInt64(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE42)
    return _mm_or_si128(_mm_cmpgt_epi64(b, a), _mm_cmpeq_epi64(a, b));
#elif defined(NLIB_NEON) && defined(__aarch64__)
    return NLIB_CMP(vcleq, s64, a, b, u64);
#else
    return I128::Not(I128::CmpGtInt64(a, b));
#endif
}

// r.s8[i] = a.s8[i] >= b.s8[i] ? 0xFF : 0
NLIB_M(i128) I128::CmpGeInt8(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_or_si128(_mm_cmpgt_epi8(a, b), _mm_cmpeq_epi8(a, b));
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcgeq, s8, a, b, u8);
#endif
}

// r.s16[i] = a.s16[i] >= b.s16[i] ? 0xFFFF : 0
NLIB_M(i128) I128::CmpGeInt16(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_or_si128(_mm_cmpgt_epi16(a, b), _mm_cmpeq_epi16(a, b));
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcgeq, s16, a, b, u16);
#endif
}

// r.s32[i] = a.s32[i] >= b.s32[i] ? 0xFFFFFFFF : 0
NLIB_M(i128) I128::CmpGeInt32(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_or_si128(_mm_cmpgt_epi32(a, b), _mm_cmpeq_epi32(a, b));
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcgeq, s32, a, b, u32);
#endif
}

// r.s64[i] = a.s64[i] >= b.s64[i] ? 0xFFFFFFFFFFFFFFFF : 0
NLIB_M(i128) I128::CmpGeInt64(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE42)
    return _mm_or_si128(_mm_cmpgt_epi64(a, b), _mm_cmpeq_epi64(a, b));
#elif defined(NLIB_NEON) && defined(__aarch64__)
    return NLIB_CMP(vcgeq, s64, a, b, u64);
#else
    return I128::Not(I128::CmpLtInt64(a, b));
#endif
}

// r.u8[i] = a.u8[i] <= b.u8[i] ? 0xFF : 0
NLIB_M(i128) I128::CmpLeUint8(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cmpeq_epi8(_mm_min_epu8(a, b), a);
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcleq, u8, a, b, u8);
#endif
}

// r.u16[i] = a.u16[i] <= b.u16[i] ? 0xFFFF : 0
NLIB_M(i128) I128::CmpLeUint16(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cmpeq_epi16(_mm_min_epu16(a, b), a);
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcleq, u16, a, b, u16);
#endif
}

// r.u32[i] = a.u32[i] <= b.u32[i] ? 0xFFFFFFFF : 0
NLIB_M(i128) I128::CmpLeUint32(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cmpeq_epi32(_mm_min_epu32(a, b), a);
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcleq, u32, a, b, u32);
#endif
}

// r.u64[i] = a.u64[i] <= b.u64[i] ? 0xFFFFFFFFFFFFFFFF : 0
NLIB_M(i128) I128::CmpLeUint64(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE42)
    i128 ofs = I128::SetValue(0x8000000000000000ULL, each_uint64);
    i128 mask = _mm_cmpgt_epi64(_mm_add_epi64(b, ofs), _mm_add_epi64(a, ofs));
    return _mm_or_si128(mask, _mm_cmpeq_epi64(a, b));
#elif defined(NLIB_NEON) && defined(__aarch64__)
    return NLIB_CMP(vcleq, u64, a, b, u64);
#else
    return I128::Not(I128::CmpGtUint64(a, b));
#endif
}

// r.u8[i] = a.u8[i] >= b.u8[i] ? 0xFF : 0
NLIB_M(i128) I128::CmpGeUint8(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cmpeq_epi8(_mm_max_epu8(a, b), a);
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcgeq, u8, a, b, u8);
#endif
}

// r.u16[i] = a.u16[i] >= b.u16[i] ? 0xFFFF : 0
NLIB_M(i128) I128::CmpGeUint16(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cmpeq_epi16(_mm_max_epu16(a, b), a);
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcgeq, u16, a, b, u16);
#endif
}

// r.u32[i] = a.u32[i] >= b.u32[i] ? 0xFFFFFFFF : 0
NLIB_M(i128) I128::CmpGeUint32(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cmpeq_epi32(_mm_max_epu32(a, b), a);
#elif defined(NLIB_NEON)
    return NLIB_CMP(vcgeq, u32, a, b, u32);
#endif
}

// r.u64[i] = a.u64[i] >= b.u64[i] ? 0xFFFFFFFFFFFFFFFF : 0
NLIB_M(i128) I128::CmpGeUint64(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE42)
    i128 ofs = I128::SetValue(0x8000000000000000ULL, each_uint64);
    i128 mask = _mm_cmpgt_epi64(_mm_add_epi64(a, ofs), _mm_add_epi64(b, ofs));
    return _mm_or_si128(mask, _mm_cmpeq_epi64(a, b));
#elif defined(NLIB_NEON) && defined(__aarch64__)
    return NLIB_CMP(vcgeq, u64, a, b, u64);
#else
    return I128::Not(I128::CmpLtUint64(a, b));
#endif
}

NLIB_M(i128) I128::CmpEqZero8(i128arg value) NLIB_NOEXCEPT {
#if defined(__aarch64__)
    return vceqzq_s8(value);
#else
    return I128::CmpEq8(value, I128::SetZero());
#endif
}

NLIB_M(i128) I128::CmpEqZero16(i128arg value) NLIB_NOEXCEPT {
#if defined(__aarch64__)
    return vreinterpretq_s8_s16(vceqzq_s16(vreinterpretq_s16_s8(value)));
#else
    return I128::CmpEq16(value, I128::SetZero());
#endif
}

NLIB_M(i128) I128::CmpEqZero32(i128arg value) NLIB_NOEXCEPT {
#if defined(__aarch64__)
    return vreinterpretq_s8_s32(vceqzq_s32(vreinterpretq_s32_s8(value)));
#else
    return I128::CmpEq32(value, I128::SetZero());
#endif
}

NLIB_M(i128) I128::CmpEqZero64(i128arg value) NLIB_NOEXCEPT {
#if defined(__aarch64__)
    return vreinterpretq_s8_s64(vceqzq_s64(vreinterpretq_s64_s8(value)));
#else
    return I128::CmpEq64(value, I128::SetZero());
#endif
}

// r.u8[i] = value.u8[i] << count
NLIB_M(i128) I128::ShiftLeftLogical8(i128arg value, int count) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    __m128i hi = _mm_shuffle_epi32(value, _MM_SHUFFLE(1, 0, 3, 2));
    __m128i xh = _mm_slli_epi16(_mm_cvtepu8_epi16(hi), count);
    __m128i xl = _mm_slli_epi16(_mm_cvtepu8_epi16(value), count);
    return I128::NarrowFrom16To8(xl, xh);
#elif defined(NLIB_NEON)
    return NLIB_SFT(vshlq, u8, value, count, s8);
#endif
}

// r.u8[i] = value.u8[i] >> count
NLIB_M(i128) I128::ShiftRightLogical8(i128arg value, int count) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    __m128i hi = _mm_shuffle_epi32(value, _MM_SHUFFLE(1, 0, 3, 2));
    __m128i xh = _mm_srli_epi16(_mm_cvtepu8_epi16(hi), count);
    __m128i xl = _mm_srli_epi16(_mm_cvtepu8_epi16(value), count);
    return _mm_packus_epi16(xl, xh);
#elif defined(NLIB_NEON)
    return NLIB_SFT(vshlq, u8, value, -count, s8);
#endif
}

// r.s8[i] = value.s8[i] >> count
NLIB_M(i128) I128::ShiftRightArithmetic8(i128arg value, int count) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    __m128i hi = _mm_shuffle_epi32(value, _MM_SHUFFLE(1, 0, 3, 2));
    __m128i xh = _mm_srai_epi16(_mm_cvtepi8_epi16(hi), count);
    __m128i xl = _mm_srai_epi16(_mm_cvtepi8_epi16(value), count);
    return _mm_packus_epi16(xl, xh);
#elif defined(NLIB_NEON)
    return NLIB_SFT(vshlq, s8, value, -count, s8);
#endif
}

// r.u16[i] = value.u16[i] << count
NLIB_M(i128) I128::ShiftLeftLogical16(i128arg value, int count) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_slli_epi16(value, count);
#elif defined(NLIB_NEON)
    return NLIB_SFT(vshlq, u16, value, count, s16);
#endif
}

// r.u16[i] = value.u16[i] >> count
NLIB_M(i128) I128::ShiftRightLogical16(i128arg value, int count) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_srli_epi16(value, count);
#elif defined(NLIB_NEON)
    return NLIB_SFT(vshlq, u16, value, -count, s16);
#endif
}

// r.s16[i] = value.s16[i] >> count
NLIB_M(i128) I128::ShiftRightArithmetic16(i128arg value, int count) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_srai_epi16(value, count);
#elif defined(NLIB_NEON)
    return NLIB_SFT(vshlq, s16, value, -count, s16);
#endif
}

// r.u32[i] = value.u32[i] << count
NLIB_M(i128) I128::ShiftLeftLogical32(i128arg value, int count) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_slli_epi32(value, count);
#elif defined(NLIB_NEON)
    return NLIB_SFT(vshlq, u32, value, count, s32);
#endif
}

// r.u32[i] = value.u32[i] >> count
NLIB_M(i128) I128::ShiftRightLogical32(i128arg value, int count) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_srli_epi32(value, count);
#elif defined(NLIB_NEON)
    return NLIB_SFT(vshlq, u32, value, -count, s32);
#endif
}

// r.s32[i] = value.s32[i] >> count
NLIB_M(i128) I128::ShiftRightArithmetic32(i128arg value, int count) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_srai_epi32(value, count);
#elif defined(NLIB_NEON)
    return NLIB_SFT(vshlq, s32, value, -count, s32);
#endif
}

// r.u32[i] = value.u64[i] << count
NLIB_M(i128) I128::ShiftLeftLogical64(i128arg value, int count) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_slli_epi64(value, count);
#elif defined(NLIB_NEON)
    return NLIB_SFT(vshlq, u64, value, count, s64);
#endif
}

// r.u32[i] = value.u64[i] >> count
NLIB_M(i128) I128::ShiftRightLogical64(i128arg value, int count) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_srli_epi64(value, count);
#elif defined(NLIB_NEON)
    return NLIB_SFT(vshlq, u64, value, -count, s64);
#endif
}

template<size_t N>
NLIB_M(i128) I128::ShiftLeftLogical8(i128arg value) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N >= 0 && N <= 8);
#ifdef NLIB_NEON
    return vshlq_n_s8(value, N);
#else
    return I128::ShiftLeftLogical8(value, N);
#endif
}

template <size_t N>
NLIB_M(i128) I128::ShiftRightLogical8(i128arg value) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N >= 0 && N <= 8);
#ifdef NLIB_NEON
    uint8x16_t tmp = vreinterpretq_u8_s8(value);
    return vreinterpretq_s8_u8(vshrq_n_u8(tmp, N));
#else
    return I128::ShiftRightLogical8(value, N);
#endif
}

template <size_t N>
NLIB_M(i128) I128::ShiftRightArithmetic8(i128arg value) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N >= 0 && N <= 8);
#ifdef NLIB_NEON
    return vshrq_n_s8(value, N);
#else
    return I128::ShiftRightArithmetic8(value, N);
#endif
}

template <size_t N>
NLIB_M(i128) I128::ShiftLeftLogical16(i128arg value) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N >= 0 && N <= 16);
#ifdef NLIB_NEON
    uint16x8_t tmp = vreinterpretq_u16_s8(value);
    return vreinterpretq_s8_u16(vshlq_n_u16(tmp, N));
#else
    return I128::ShiftLeftLogical16(value, N);
#endif
}

template <size_t N>
NLIB_M(i128) I128::ShiftRightLogical16(i128arg value) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N >= 0 && N <= 16);
#ifdef NLIB_NEON
    uint16x8_t tmp = vreinterpretq_u16_s8(value);
    return vreinterpretq_s8_u16(vshrq_n_u16(tmp, N));
#else
    return I128::ShiftRightLogical16(value, N);
#endif
}

template <size_t N>
NLIB_M(i128) I128::ShiftRightArithmetic16(i128arg value) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N >= 0 && N <= 16);
#ifdef NLIB_NEON
    int16x8_t tmp = vreinterpretq_s16_s8(value);
    return vreinterpretq_s8_s16(vshrq_n_s16(tmp, N));
#else
    return I128::ShiftRightArithmetic16(value, N);
#endif
}

template <size_t N>
NLIB_M(i128) I128::ShiftLeftLogical32(i128arg value) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N >= 0 && N <= 32);
#ifdef NLIB_NEON
    uint32x4_t tmp = vreinterpretq_u32_s8(value);
    return vreinterpretq_s8_u32(vshlq_n_u32(tmp, N));
#else
    return I128::ShiftLeftLogical32(value, N);
#endif
}

template <size_t N>
NLIB_M(i128) I128::ShiftRightLogical32(i128arg value) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N >= 0 && N <= 32);
#ifdef NLIB_NEON
    uint32x4_t tmp = vreinterpretq_u32_s8(value);
    return vreinterpretq_s8_u32(vshrq_n_u32(tmp, N));
#else
    return I128::ShiftRightLogical32(value, N);
#endif
}

template <size_t N>
NLIB_M(i128) I128::ShiftRightArithmetic32(i128arg value) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N >= 0 && N <= 32);
#ifdef NLIB_NEON
    int32x4_t tmp = vreinterpretq_s32_s8(value);
    return vreinterpretq_s8_s32(vshrq_n_s32(tmp, N));
#else
    return I128::ShiftRightArithmetic32(value, N);
#endif
}

template <size_t N>
NLIB_M(i128) I128::ShiftLeftLogical64(i128arg value) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N >= 0 && N <= 64);
#ifdef NLIB_NEON
    uint64x2_t tmp = vreinterpretq_u64_s8(value);
    return vreinterpretq_s8_u64(vshlq_n_u64(tmp, N));
#else
    return I128::ShiftLeftLogical64(value, N);
#endif
}

template <size_t N>
NLIB_M(i128) I128::ShiftRightLogical64(i128arg value) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N >= 0 && N <= 64);
#ifdef NLIB_NEON
    uint64x2_t tmp = vreinterpretq_u64_s8(value);
    return vreinterpretq_s8_u64(vshrq_n_u64(tmp, N));
#else
    return I128::ShiftRightLogical64(value, N);
#endif
}

#ifdef NLIB_NEON
template<>
NLIB_M(i128) I128::ShiftLeftLogical8<8>(i128arg value) NLIB_NOEXCEPT {
    NLIB_UNUSED(value);
    return I128::SetZero();
}
template<>
NLIB_M(i128) I128::ShiftRightLogical8<0>(i128arg value) NLIB_NOEXCEPT {
    return value;
}
template<>
NLIB_M(i128) I128::ShiftLeftLogical16<16>(i128arg value) NLIB_NOEXCEPT {
    NLIB_UNUSED(value);
    return I128::SetZero();
}
template<>
NLIB_M(i128) I128::ShiftRightLogical16<0>(i128arg value) NLIB_NOEXCEPT {
    return value;
}
template<>
NLIB_M(i128) I128::ShiftRightArithmetic16<0>(i128arg value) NLIB_NOEXCEPT {
    return value;
}
template<>
NLIB_M(i128) I128::ShiftLeftLogical32<32>(i128arg value) NLIB_NOEXCEPT {
    NLIB_UNUSED(value);
    return I128::SetZero();
}
template<>
NLIB_M(i128) I128::ShiftRightLogical32<0>(i128arg value) NLIB_NOEXCEPT {
    return value;
}
template<>
NLIB_M(i128) I128::ShiftRightArithmetic32<0>(i128arg value) NLIB_NOEXCEPT {
    return value;
}
template<>
NLIB_M(i128) I128::ShiftLeftLogical64<64>(i128arg value) NLIB_NOEXCEPT {
    NLIB_UNUSED(value);
    return I128::SetZero();
}
template<>
NLIB_M(i128) I128::ShiftRightLogical64<0>(i128arg value) NLIB_NOEXCEPT {
    return value;
}
#endif

template <size_t N>
// r.i8[i + N] = r.i8[i], fill zero
NLIB_M(i128) I128::ByteShiftLeft(i128arg value) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N < 16);
#if defined(NLIB_SSE41)
    return _mm_slli_si128(value, N);
#elif defined(NLIB_NEON)
    return vextq_s8(vdupq_n_s8(0), value, 16 - N);
#endif
}

template <size_t N>
// r.i8[i - N] = r.i8[i], fill zero
NLIB_M(i128) I128::ByteShiftRight(i128arg value) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N < 16);
#if defined(NLIB_SSE41)
    return _mm_srli_si128(value, N);
#elif defined(NLIB_NEON)
    return vextq_s8(value, vdupq_n_s8(0), N);
#endif
}

template <size_t N>
// left <- fedcba9876543210 -> right
NLIB_M(i128) I128::ByteRotateRight(i128arg value) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N < 16);
#if defined(NLIB_SSE41)
    return _mm_alignr_epi8(value, value, N);
#elif defined(NLIB_NEON)
    return vextq_s8(value, value, N);
#endif
}

template <size_t N>
// _mm_alignr_epi8(a, b, N), which returns ((a << 128) | b) >> (N * 8)
NLIB_M(i128) I128::AlignR(i128arg a, i128arg b) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(N < 16);
#if defined(NLIB_SSE41)
    return _mm_alignr_epi8(a, b, N);
#elif defined(NLIB_NEON)
    return vextq_s8(b, a, N);
#endif
}

// r.u8[i] = value.u8[i * 2]
NLIB_M(i128) I128::NarrowFrom16To8(i128arg lo, i128arg hi) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    i128 mask = I128::SetValue(0x00FFU, each_uint16);
    __m128i lo_mask = _mm_and_si128(lo, mask);
    __m128i hi_mask = _mm_and_si128(hi, mask);
    return _mm_packus_epi16(lo_mask, hi_mask);
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    uint8x8_t l = vmovn_u16(vreinterpretq_u16_s8(lo));
    return vreinterpretq_s8_u8(vmovn_high_u16(l, vreinterpretq_u16_s8(hi)));
# else
    uint8x8_t l = vmovn_u16(vreinterpretq_u16_s8(lo));
    uint8x8_t h = vmovn_u16(vreinterpretq_u16_s8(hi));
    return NLIB_CMB(u8, l, h);
# endif
#endif
}

// r.u16[i] = value.u16[i * 2]
NLIB_M(i128) I128::NarrowFrom32To16(i128arg lo, i128arg hi) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    i128 mask = I128::SetValue(0xFFFFU, each_uint32);
    __m128i lo_mask = _mm_and_si128(lo, mask);
    __m128i hi_mask = _mm_and_si128(hi, mask);
    return _mm_packus_epi32(lo_mask, hi_mask);
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    uint16x4_t l = vmovn_u32(vreinterpretq_u32_s8(lo));
    return vreinterpretq_s8_u16(vmovn_high_u32(l, vreinterpretq_u32_s8(hi)));
# else
    uint16x4_t l = vmovn_u32(vreinterpretq_u32_s8(lo));
    uint16x4_t h = vmovn_u32(vreinterpretq_u32_s8(hi));
    return NLIB_CMB(u16, l, h);
# endif
#endif
}

// r.u32[i] = value.u32[i * 2]
NLIB_M(i128) I128::NarrowFrom64To32(i128arg lo, i128arg hi) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    __m128i lo_ = _mm_shuffle_epi32(lo, _MM_SHUFFLE(3, 1, 2, 0));
    __m128i hi_ = _mm_shuffle_epi32(hi, _MM_SHUFFLE(3, 1, 2, 0));
    return _mm_unpacklo_epi64(lo_, hi_);
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    uint32x2_t l = vmovn_u64(vreinterpretq_u64_s8(lo));
    return vreinterpretq_s8_u32(vmovn_high_u64(l, vreinterpretq_u64_s8(hi)));
# else
    uint32x2_t l = vmovn_u64(vreinterpretq_u64_s8(lo));
    uint32x2_t h = vmovn_u64(vreinterpretq_u64_s8(hi));
    return NLIB_CMB(u32, l, h);
# endif
#endif
}

// r.u8[i] = 255 if value.u16[i] > 255
NLIB_M(i128) I128::ConvertFromUint16ToUint8Saturated(i128arg lo, i128arg hi) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    i128 b7FFF = I128::SetValue(0x7FFFU, each_uint16);
    __m128i lotmp = _mm_and_si128(lo, b7FFF);
    __m128i hitmp = _mm_and_si128(hi, b7FFF);
    return _mm_packus_epi16(lotmp, hitmp);
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    uint8x8_t l = vqmovn_u16(vreinterpretq_u16_s8(lo));
    return vreinterpretq_s8_u8(vqmovn_high_u16(l, vreinterpretq_u16_s8(hi)));
# else
    uint8x8_t l = vqmovn_u16(vreinterpretq_u16_s8(lo));
    uint8x8_t h = vqmovn_u16(vreinterpretq_u16_s8(hi));
    return NLIB_CMB(u8, l, h);
# endif
#endif
}

// r.s8[i] = 127 if value.s16[i] > 127, -128 if value.s16[i] < -128
NLIB_M(i128) I128::ConvertFromInt16ToInt8Saturated(i128arg lo, i128arg hi) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_packs_epi16(lo, hi);
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    int8x8_t l = vqmovn_s16(vreinterpretq_s16_s8(lo));
    return vqmovn_high_s16(l, vreinterpretq_s16_s8(hi));
# else
    int8x8_t l = vqmovn_s16(vreinterpretq_s16_s8(lo));
    int8x8_t h = vqmovn_s16(vreinterpretq_s16_s8(hi));
    return NLIB_CMB(s8, l, h);
# endif
#endif
}

// r.u16[i] = 65535 if value.u32[i] > 65535
NLIB_M(i128) I128::ConvertFromUint32ToUint16Saturated(i128arg lo, i128arg hi) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    i128 b7FFFFFFF = I128::SetValue(0x7FFFFFFFU, each_uint32);
    __m128i lotmp = _mm_and_si128(lo, b7FFFFFFF);
    __m128i hitmp = _mm_and_si128(hi, b7FFFFFFF);
    return _mm_packus_epi32(lotmp, hitmp);
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    uint16x4_t l = vqmovn_u32(vreinterpretq_u32_s8(lo));
    return vreinterpretq_s8_u16(vqmovn_high_u32(l, vreinterpretq_u32_s8(hi)));
# else
    uint16x4_t l = vqmovn_u32(vreinterpretq_u32_s8(lo));
    uint16x4_t h = vqmovn_u32(vreinterpretq_u32_s8(hi));
    return NLIB_CMB(u16, l, h);
# endif
#endif
}

// r.s16[i] = 32767 if value.s32[i] > 32767, -32768 if value.s32[i] < -32768
NLIB_M(i128) I128::ConvertFromInt32ToInt16Saturated(i128arg lo, i128arg hi) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_packs_epi32(lo, hi);
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    int16x4_t l = vqmovn_s32(vreinterpretq_s32_s8(lo));
    return vreinterpretq_s8_s16(vqmovn_high_s32(l, vreinterpretq_s32_s8(hi)));
# else
    int16x4_t l = vqmovn_s32(vreinterpretq_s32_s8(lo));
    int16x4_t h = vqmovn_s32(vreinterpretq_s32_s8(hi));
    return NLIB_CMB(s16, l, h);
# endif
#endif
}

// r.s8[2 * i] = value.s8[i], r.u8[2 * i + 1] = value.s8[i] < 0 ? 0xFF : 0
NLIB_M(i128) I128::ConvertFromInt8ToInt16Lo(i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cvtepi8_epi16(value);
#elif defined(NLIB_NEON)
    return vreinterpretq_s8_s16(vmovl_s8(vget_low_s8(value)));
#endif
}

// r.s8[2 * i] = value.s8[i + 8], r.u8[2 * i + 1] = value.s8[i + 8] < 0 ? 0xFF : 0
NLIB_M(i128) I128::ConvertFromInt8ToInt16Hi(i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cvtepi8_epi16(_mm_shuffle_epi32(value, _MM_SHUFFLE(1, 0, 3, 2)));
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    int16x8_t result = vmovl_high_s8(value);
# else
    int16x8_t result = vmovl_s8(vget_high_s8(value));
# endif
    return vreinterpretq_s8_s16(result);
#endif
}

// r.s16[2 * i] = value.s16[i], r.u16[2 * i + 1] = value.s16[i] < 0 ? 0xFFFF : 0
NLIB_M(i128) I128::ConvertFromInt16ToInt32Lo(i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cvtepi16_epi32(value);
#elif defined(NLIB_NEON)
    int16x8_t x = vreinterpretq_s16_s8(value);
    int32x4_t result = vmovl_s16(vget_low_s16(x));
    return vreinterpretq_s8_s32(result);
#endif
}

// r.s16[2 * i] = value.s16[i + 4], r.u16[2 * i + 1] = value.s16[i + 4] < 0 ? 0xFFFF : 0
NLIB_M(i128) I128::ConvertFromInt16ToInt32Hi(i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cvtepi16_epi32(_mm_shuffle_epi32(value, _MM_SHUFFLE(1, 0, 3, 2)));
#elif defined(NLIB_NEON)
    int16x8_t x = vreinterpretq_s16_s8(value);
# ifdef __aarch64__
    int32x4_t result = vmovl_high_s16(x);
# else
    int32x4_t result = vmovl_s16(vget_high_s16(x));
# endif
    return vreinterpretq_s8_s32(result);
#endif
}

// r.s32[2 * i] = value.s32[i], r.u32[2 * i + 1] = value.s32[i] < 0 ? 0xFFFFFFFF : 0
NLIB_M(i128) I128::ConvertFromInt32ToInt64Lo(i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cvtepi32_epi64(value);
#elif defined(NLIB_NEON)
    int32x4_t x = vreinterpretq_s32_s8(value);
    int64x2_t result = vmovl_s32(vget_low_s32(x));
    return vreinterpretq_s8_s64(result);
#endif
}

// r.s32[2 * i] = value.s32[i + 2], r.u32[2 * i + 1] = value.s32[i + 2] < 0 ? 0xFFFFFFFF : 0
NLIB_M(i128) I128::ConvertFromInt32ToInt64Hi(i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cvtepi32_epi64(_mm_shuffle_epi32(value, _MM_SHUFFLE(1, 0, 3, 2)));
#elif defined(NLIB_NEON)
    int32x4_t x = vreinterpretq_s32_s8(value);
# ifdef __aarch64__
    int64x2_t result = vmovl_high_s32(x);
# else
    int64x2_t result = vmovl_s32(vget_high_s32(x));
# endif
    return vreinterpretq_s8_s64(result);
#endif
}

// r.u16[i] = value.u8[i]
NLIB_M(i128) I128::ConvertFromUint8ToUint16Lo(i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cvtepu8_epi16(value);
#elif defined(NLIB_NEON)
    uint8x16_t x = vreinterpretq_u8_s8(value);
    uint16x8_t result = vmovl_u8(vget_low_u8(x));
    return vreinterpretq_s8_u16(result);
#endif
}

// r.u16[i] = value.u8[i + 8]
NLIB_M(i128) I128::ConvertFromUint8ToUint16Hi(i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cvtepu8_epi16(_mm_shuffle_epi32(value, _MM_SHUFFLE(1, 0, 3, 2)));
#elif defined(NLIB_NEON)
    uint8x16_t x = vreinterpretq_u8_s8(value);
# ifdef __aarch64__
    uint16x8_t result = vmovl_high_u8(x);
# else
    uint16x8_t result = vmovl_u8(vget_high_u8(x));
# endif
    return vreinterpretq_s8_u16(result);
#endif
}

// r.u32[i] = value.u16[i]
NLIB_M(i128) I128::ConvertFromUint16ToUint32Lo(i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cvtepu16_epi32(value);
#elif defined(NLIB_NEON)
    uint16x8_t x = vreinterpretq_u16_s8(value);
    uint32x4_t result = vmovl_u16(vget_low_u16(x));
    return vreinterpretq_s8_u32(result);
#endif
}

// r.u32[i] = value.u16[i + 4]
NLIB_M(i128) I128::ConvertFromUint16ToUint32Hi(i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cvtepu16_epi32(_mm_shuffle_epi32(value, _MM_SHUFFLE(1, 0, 3, 2)));
#elif defined(NLIB_NEON)
    uint16x8_t x = vreinterpretq_u16_s8(value);
# ifdef __aarch64__
    uint32x4_t result = vmovl_high_u16(x);
# else
    uint32x4_t result = vmovl_u16(vget_high_u16(x));
# endif
    return vreinterpretq_s8_u32(result);
#endif
}

// r.u64[i] = value.u32[i]
NLIB_M(i128) I128::ConvertFromUint32ToUint64Lo(i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cvtepu32_epi64(value);
#elif defined(NLIB_NEON)
    uint32x4_t x = vreinterpretq_u32_s8(value);
    uint64x2_t result = vmovl_u32(vget_low_u32(x));
    return vreinterpretq_s8_u64(result);
#endif
}

// r.u64[i] = value.u32[i + 2]
NLIB_M(i128) I128::ConvertFromUint32ToUint64Hi(i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_cvtepu32_epi64(_mm_shuffle_epi32(value, _MM_SHUFFLE(1, 0, 3, 2)));
#elif defined(NLIB_NEON)
    uint32x4_t x = vreinterpretq_u32_s8(value);
# ifdef __aarch64__
    uint64x2_t result = vmovl_high_u32(x);
# else
    uint64x2_t result = vmovl_u32(vget_high_u32(x));
# endif
    return vreinterpretq_s8_u64(result);
#endif
}

// 01234567 abcdefgh -> 0a1b2c3d4e5f6g7h
NLIB_M(i128) I128::Zip8Lo(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_unpacklo_epi8(a, b);
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    return vzip1q_s8(a, b);
# else
    return vzipq_s8(a, b).val[0];
# endif
#endif
}

// 01234567 abcdefgh -> 0a1b2c3d4e5f6g7h
NLIB_M(i128) I128::Zip8Hi(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_unpackhi_epi8(a, b);
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    return vzip2q_s8(a, b);
# else
    return vzipq_s8(a, b).val[1];
# endif
#endif
}

NLIB_M(i128) I128::Unzip8Lo(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    i128 mask = I128::SetValue(0x00FFU, each_uint16);
    __m128i lo_mask = _mm_and_si128(a, mask);
    __m128i hi_mask = _mm_and_si128(b, mask);
    return _mm_packus_epi16(lo_mask, hi_mask);
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    return vuzp1q_s8(a, b);
# else
    return vuzpq_s8(a, b).val[0];
# endif
#endif
}

NLIB_M(i128) I128::Unzip8Hi(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    i128 mask = I128::SetValue(0xFF00U, each_uint16);
    __m128i lo_mask = _mm_srli_si128(_mm_and_si128(a, mask), 1);
    __m128i hi_mask = _mm_srli_si128(_mm_and_si128(b, mask), 1);
    return _mm_packus_epi16(lo_mask, hi_mask);
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    return vuzp2q_s8(a, b);
# else
    return vuzpq_s8(a, b).val[1];
# endif
#endif
}

// 0123 abcd -> 0a1b2c3d
NLIB_M(i128) I128::Zip16Lo(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_unpacklo_epi16(a, b);
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    return NLIB_OP2(vzip1q, u16, a, b);
# else
    return vreinterpretq_s8_u16(vzipq_u16(
        vreinterpretq_u16_s8(a), vreinterpretq_u16_s8(b)).val[0]);
# endif
#endif
}

// 0123 abcd -> 0a1b2c3d
NLIB_M(i128) I128::Zip16Hi(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_unpackhi_epi16(a, b);
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    return NLIB_OP2(vzip2q, u16, a, b);
# else
    return vreinterpretq_s8_u16(vzipq_u16(
        vreinterpretq_u16_s8(a), vreinterpretq_u16_s8(b)).val[1]);
# endif
#endif
}

NLIB_M(i128) I128::Unzip16Lo(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    i128 mask = I128::SetValue(0xFFFFU, each_uint32);
    __m128i lo_mask = _mm_and_si128(a, mask);
    __m128i hi_mask = _mm_and_si128(b, mask);
    return _mm_packus_epi32(lo_mask, hi_mask);
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    return NLIB_OP2(vuzp1q, u16, a, b);
# else
    return vreinterpretq_s8_u16(vuzpq_u16(
        vreinterpretq_u16_s8(a), vreinterpretq_u16_s8(b)).val[0]);
# endif
#endif
}

NLIB_M(i128) I128::Unzip16Hi(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    i128 mask = I128::SetValue(0xFFFF0000U, each_uint32);
    __m128i lo_mask = _mm_srli_si128(_mm_and_si128(a, mask), 2);
    __m128i hi_mask = _mm_srli_si128(_mm_and_si128(b, mask), 2);
    return _mm_packus_epi32(lo_mask, hi_mask);
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    return NLIB_OP2(vuzp2q, u16, a, b);
# else
    return vreinterpretq_s8_u16(vuzpq_u16(
        vreinterpretq_u16_s8(a), vreinterpretq_u16_s8(b)).val[1]);
# endif
#endif
}

// 01 ab -> 0a1b
NLIB_M(i128) I128::Zip32Lo(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_unpacklo_epi32(a, b);
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    return NLIB_OP2(vzip1q, u32, a, b);
# else
    return vreinterpretq_s8_u32(vzipq_u32(
        vreinterpretq_u32_s8(a), vreinterpretq_u32_s8(b)).val[0]);
# endif
#endif
}

// 01 ab -> 0a1b
NLIB_M(i128) I128::Zip32Hi(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_unpackhi_epi32(a, b);
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    return NLIB_OP2(vzip2q, u32, a, b);
# else
    return vreinterpretq_s8_u32(vzipq_u32(
        vreinterpretq_u32_s8(a), vreinterpretq_u32_s8(b)).val[1]);
# endif
#endif
}

NLIB_M(i128) I128::Unzip32Lo(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    __m128i x0 = _mm_shuffle_epi32(a, _MM_SHUFFLE(3, 1, 2, 0));
    __m128i x1 = _mm_shuffle_epi32(b, _MM_SHUFFLE(2, 0, 3, 1));
    return _mm_blend_epi16(x0, x1, 0xF0);
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    return NLIB_OP2(vuzp1q, u32, a, b);
# else
    return vreinterpretq_s8_u32(vuzpq_u32(
        vreinterpretq_u32_s8(a), vreinterpretq_u32_s8(b)).val[0]);
# endif
#endif
}

NLIB_M(i128) I128::Unzip32Hi(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    __m128i x0 = _mm_shuffle_epi32(a, _MM_SHUFFLE(2, 0, 3, 1));
    __m128i x1 = _mm_shuffle_epi32(b, _MM_SHUFFLE(3, 1, 2, 0));
    return _mm_blend_epi16(x0, x1, 0xF0);
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    return NLIB_OP2(vuzp2q, u32, a, b);
# else
    return vreinterpretq_s8_u32(vuzpq_u32(
        vreinterpretq_u32_s8(a), vreinterpretq_u32_s8(b)).val[1]);
# endif
#endif
}

template<int V0, int V1, int V2, int V3, int V4, int V5, int V6, int V7,
         int V8, int V9, int V10, int V11, int V12, int V13, int V14, int V15>
NLIB_M(i128) I128::Permute8(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if __has_builtin(__builtin_shufflevector) && defined(NLIB_NEON)
    return __builtin_shufflevector(
        a, b,
        V0, V1, V2, V3, V4, V5, V6, V7,
        V8, V9, V10, V11, V12, V13, V14, V15);
#elif __has_builtin(__builtin_shufflevector) && defined(NLIB_SSE41)
    return __builtin_shufflevector((__v16qi)a, (__v16qi)b,
                                   V0, V1, V2, V3, V4, V5, V6, V7,
                                   V8, V9, V10, V11, V12, V13, V14, V15);
#else
    NLIB_ALIGNAS(16) int8_t mask_a[16] = {
        (V0 < 0 || V0 > 15) ? -128 : V0,
        (V1 < 0 || V1 > 15) ? -128 : V1,
        (V2 < 0 || V2 > 15) ? -128 : V2,
        (V3 < 0 || V3 > 15) ? -128 : V3,
        (V4 < 0 || V4 > 15) ? -128 : V4,
        (V5 < 0 || V5 > 15) ? -128 : V5,
        (V6 < 0 || V6 > 15) ? -128 : V6,
        (V7 < 0 || V7 > 15) ? -128 : V7,
        (V8 < 0 || V8 > 15) ? -128 : V8,
        (V9 < 0 || V9 > 15) ? -128 : V9,
        (V10 < 0 || V10 > 15) ? -128 : V10,
        (V11 < 0 || V11 > 15) ? -128 : V11,
        (V12 < 0 || V12 > 15) ? -128 : V12,
        (V13 < 0 || V13 > 15) ? -128 : V13,
        (V14 < 0 || V14 > 15) ? -128 : V14,
        (V15 < 0 || V15 > 15) ? -128 : V15
    };
    NLIB_ALIGNAS(16) int8_t mask_b[16] = {
        V0 < 16 ? -128 : (V0 - 16),
        V1 < 16 ? -128 : (V1 - 16),
        V2 < 16 ? -128 : (V2 - 16),
        V3 < 16 ? -128 : (V3 - 16),
        V4 < 16 ? -128 : (V4 - 16),
        V5 < 16 ? -128 : (V5 - 16),
        V6 < 16 ? -128 : (V6 - 16),
        V7 < 16 ? -128 : (V7 - 16),
        V8 < 16 ? -128 : (V8 - 16),
        V9 < 16 ? -128 : (V9 - 16),
        V10 < 16 ? -128 : (V10 - 16),
        V11 < 16 ? -128 : (V11 - 16),
        V12 < 16 ? -128 : (V12 - 16),
        V13 < 16 ? -128 : (V13 - 16),
        V14 < 16 ? -128 : (V14 - 16),
        V15 < 16 ? -128 : (V15 - 16)
    };
    i128 tmp_a = I128::Shuffle8(a, I128::LoadA16(mask_a));
    i128 tmp_b = I128::Shuffle8(b, I128::LoadA16(mask_b));
    return I128::Or(tmp_a, tmp_b);
#endif
}

template<int V0, int V1, int V2, int V3, int V4, int V5, int V6, int V7>
NLIB_M(i128) I128::Permute16(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if __has_builtin(__builtin_shufflevector) && defined(NLIB_NEON)
    return vreinterpretq_s8_u16(__builtin_shufflevector(
        vreinterpretq_u16_s8(a), vreinterpretq_u16_s8(b),
        V0, V1, V2, V3, V4, V5, V6, V7));
#elif __has_builtin(__builtin_shufflevector) && defined(NLIB_SSE41)
    return __builtin_shufflevector((__v8hi)a, (__v8hi)b,
                                   V0, V1, V2, V3, V4, V5, V6, V7);
#else
    NLIB_ALIGNAS(16) int8_t mask_a[16] = {
        (V0 < 0 || V0 > 7) ? -128 : V0 * 2,
        (V0 < 0 || V0 > 7) ? -128 : V0 * 2 + 1,
        (V1 < 0 || V1 > 7) ? -128 : V1 * 2,
        (V1 < 0 || V1 > 7) ? -128 : V1 * 2 + 1,
        (V2 < 0 || V2 > 7) ? -128 : V2 * 2,
        (V2 < 0 || V2 > 7) ? -128 : V2 * 2 + 1,
        (V3 < 0 || V3 > 7) ? -128 : V3 * 2,
        (V3 < 0 || V3 > 7) ? -128 : V3 * 2 + 1,
        (V4 < 0 || V4 > 7) ? -128 : V4 * 2,
        (V4 < 0 || V4 > 7) ? -128 : V4 * 2 + 1,
        (V5 < 0 || V5 > 7) ? -128 : V5 * 2,
        (V5 < 0 || V5 > 7) ? -128 : V5 * 2 + 1,
        (V6 < 0 || V6 > 7) ? -128 : V6 * 2,
        (V6 < 0 || V6 > 7) ? -128 : V6 * 2 + 1,
        (V7 < 0 || V7 > 7) ? -128 : V7 * 2,
        (V7 < 0 || V7 > 7) ? -128 : V7 * 2 + 1
    };
    NLIB_ALIGNAS(16) int8_t mask_b[16] = {
        V0 < 8 ? -128 : (V0 - 8) * 2,
        V0 < 8 ? -128 : (V0 - 8) * 2 + 1,
        V1 < 8 ? -128 : (V1 - 8) * 2,
        V1 < 8 ? -128 : (V1 - 8) * 2 + 1,
        V2 < 8 ? -128 : (V2 - 8) * 2,
        V2 < 8 ? -128 : (V2 - 8) * 2 + 1,
        V3 < 8 ? -128 : (V3 - 8) * 2,
        V3 < 8 ? -128 : (V3 - 8) * 2 + 1,
        V4 < 8 ? -128 : (V4 - 8) * 2,
        V4 < 8 ? -128 : (V4 - 8) * 2 + 1,
        V5 < 8 ? -128 : (V5 - 8) * 2,
        V5 < 8 ? -128 : (V5 - 8) * 2 + 1,
        V6 < 8 ? -128 : (V6 - 8) * 2,
        V6 < 8 ? -128 : (V6 - 8) * 2 + 1,
        V7 < 8 ? -128 : (V7 - 8) * 2,
        V7 < 8 ? -128 : (V7 - 8) * 2 + 1
    };
    i128 tmp_a = I128::Shuffle8(a, I128::LoadA16(mask_a));
    i128 tmp_b = I128::Shuffle8(b, I128::LoadA16(mask_b));
    return I128::Or(tmp_a, tmp_b);
#endif
}

template<int V0, int V1, int V2, int V3>
NLIB_M(i128) I128::Permute32(i128arg a, i128arg b) NLIB_NOEXCEPT {
#if __has_builtin(__builtin_shufflevector) && defined(NLIB_NEON)
    return vreinterpretq_s8_u32(__builtin_shufflevector(
        vreinterpretq_u32_s8(a), vreinterpretq_u32_s8(b),
        V0, V1, V2, V3));
#elif __has_builtin(__builtin_shufflevector) && defined(NLIB_SSE41)
    return __builtin_shufflevector((__v4si)a, (__v4si)b,
                                   V0, V1, V2, V3);
#else
    NLIB_ALIGNAS(16) int8_t mask_a[16] = {
        (V0 < 0 || V0 > 3) ? -128 : V0 * 4,
        (V0 < 0 || V0 > 3) ? -128 : V0 * 4 + 1,
        (V0 < 0 || V0 > 3) ? -128 : V0 * 4 + 2,
        (V0 < 0 || V0 > 3) ? -128 : V0 * 4 + 3,
        (V1 < 0 || V1 > 3) ? -128 : V1 * 4,
        (V1 < 0 || V1 > 3) ? -128 : V1 * 4 + 1,
        (V1 < 0 || V1 > 3) ? -128 : V1 * 4 + 2,
        (V1 < 0 || V1 > 3) ? -128 : V1 * 4 + 3,
        (V2 < 0 || V2 > 3) ? -128 : V2 * 4,
        (V2 < 0 || V2 > 3) ? -128 : V2 * 4 + 1,
        (V2 < 0 || V2 > 3) ? -128 : V2 * 4 + 2,
        (V2 < 0 || V2 > 3) ? -128 : V2 * 4 + 3,
        (V3 < 0 || V3 > 3) ? -128 : V3 * 4,
        (V3 < 0 || V3 > 3) ? -128 : V3 * 4 + 1,
        (V3 < 0 || V3 > 3) ? -128 : V3 * 4 + 2,
        (V3 < 0 || V3 > 3) ? -128 : V3 * 4 + 3
    };
    NLIB_ALIGNAS(16) int8_t mask_b[16] = {
        V0 < 4 ? -128 : (V0 - 4) * 4,
        V0 < 4 ? -128 : (V0 - 4) * 4 + 1,
        V0 < 4 ? -128 : (V0 - 4) * 4 + 2,
        V0 < 4 ? -128 : (V0 - 4) * 4 + 3,
        V1 < 4 ? -128 : (V1 - 4) * 4,
        V1 < 4 ? -128 : (V1 - 4) * 4 + 1,
        V1 < 4 ? -128 : (V1 - 4) * 4 + 2,
        V1 < 4 ? -128 : (V1 - 4) * 4 + 3,
        V2 < 4 ? -128 : (V2 - 4) * 4,
        V2 < 4 ? -128 : (V2 - 4) * 4 + 1,
        V2 < 4 ? -128 : (V2 - 4) * 4 + 2,
        V2 < 4 ? -128 : (V2 - 4) * 4 + 3,
        V3 < 4 ? -128 : (V3 - 4) * 4,
        V3 < 4 ? -128 : (V3 - 4) * 4 + 1,
        V3 < 4 ? -128 : (V3 - 4) * 4 + 2,
        V3 < 4 ? -128 : (V3 - 4) * 4 + 3
    };
    i128 tmp_a = I128::Shuffle8(a, I128::LoadA16(mask_a));
    i128 tmp_b = I128::Shuffle8(b, I128::LoadA16(mask_b));
    return I128::Or(tmp_a, tmp_b);
#endif
}


// 0123456789abcdef -> 1032547698badcfe
NLIB_M(i128) I128::Reverse16(i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    NLIB_ALIGNAS(16) const int8_t mask_[16] = {
        1, 0, 3, 2, 5, 4, 7, 6, 9, 8, 11, 10, 13, 12, 15, 14
    };
    return _mm_shuffle_epi8(value, *reinterpret_cast<const __m128i*>(&mask_[0]));
#elif defined(NLIB_NEON)
    return NLIB_OP1(vrev16q, u8, value);
#endif
}

// 0123456789abcdef -> 32107654ba98fedc
NLIB_M(i128) I128::Reverse32(i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    NLIB_ALIGNAS(16) const int8_t mask_[16] = {
        3, 2, 1, 0, 7, 6, 5, 4, 11, 10, 9, 8, 15, 14, 13, 12
    };
    return _mm_shuffle_epi8(value, *reinterpret_cast<const __m128i*>(&mask_[0]));
#elif defined(NLIB_NEON)
    return NLIB_OP1(vrev32q, u8, value);
#endif
}

// 0123456789abcdef -> 76543210fedcba98
NLIB_M(i128) I128::Reverse64(i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    NLIB_ALIGNAS(16) const int8_t mask_[16] = {
        7, 6, 5, 4, 3, 2, 1, 0, 15, 14, 13, 12, 11, 10, 9, 8
    };
    return _mm_shuffle_epi8(value, *reinterpret_cast<const __m128i*>(&mask_[0]));
#elif defined(NLIB_NEON)
    return NLIB_OP1(vrev64q, u8, value);
#endif
}

// r = Or(ismasked(value.u8[i]) ? (1 << i) : 0)
NLIB_M(int) I128::MoveMask8(i128arg value) NLIB_NOEXCEPT {  // NOLINT
#if defined(NLIB_SSE41)
    return _mm_movemask_epi8(value);
#elif defined(NLIB_NEON)
    uint8x16_t powers = vreinterpretq_u8_u64(vdupq_n_u64(0x8040201008040201ULL));
    uint8x16_t a = vandq_u8(value, powers);
# ifdef __aarch64__
    return vaddv_u8(vget_low_u8(a)) | (vaddv_u8(vget_high_u8(a)) << 8);
# else
    uint8x8_t al = vget_low_u8(a);
    uint8x8_t ah = vget_high_u8(a);
    uint8x8_t tmp = vpadd_u8(al, ah);
    tmp = vpadd_u8(tmp, tmp);
    tmp = vpadd_u8(tmp, tmp);
    return vget_lane_u16(vreinterpret_u16_u8(tmp), 0);
# endif
#endif
}

// r = Or(ismasked(value.u16[i]) ? (1 << i) : 0)
NLIB_M(int) I128::MoveMask16(i128arg value) NLIB_NOEXCEPT {  // NOLINT
#if defined(NLIB_SSE41)
    __m128i tmp = _mm_packs_epi16(value, value);
    return _mm_movemask_epi8(tmp) & 255;
#elif defined(NLIB_NEON)
    uint16x4_t powers_lo = vcreate_u16(0x0008000400020001ULL);
    uint16x4_t powers_hi = vshl_n_u16(powers_lo, 4);
    uint16x8_t powers = vcombine_u16(powers_lo, powers_hi);
    uint16x8_t a = vandq_u16(vreinterpretq_u16_s8(value), powers);
# ifdef __aarch64__
    return vaddvq_u16(a);
# else
    uint8x8_t tmp = vmovn_u16(a);
    tmp = vpadd_u8(tmp, tmp);
    tmp = vpadd_u8(tmp, tmp);
    tmp = vpadd_u8(tmp, tmp);
    return vget_lane_u8(tmp, 0);
# endif
#endif
}

// r = Or(ismasked(value.u32[i]) ? (1 << i) : 0)
NLIB_M(int) I128::MoveMask32(i128arg value) NLIB_NOEXCEPT {  // NOLINT
#if defined(NLIB_SSE41)
    __m128i tmp = _mm_packs_epi16(value, value);
    tmp = _mm_packs_epi16(tmp, tmp);
    return _mm_movemask_epi8(tmp) & 15;
#elif defined(NLIB_NEON)
    uint32x2_t powers_lo = vcreate_u32(0x0000000200000001ULL);
    uint32x2_t powers_hi = vshl_n_u32(powers_lo, 2);
    uint32x4_t powers = vcombine_u32(powers_lo, powers_hi);
    uint32x4_t a = vandq_u32(vreinterpretq_u32_s8(value), powers);
# ifdef __aarch64__
    return vaddvq_u32(a);
# else
    uint16x4_t tmp = vmovn_u32(a);
    tmp = vpadd_u16(tmp, tmp);
    tmp = vpadd_u16(tmp, tmp);
    return vget_lane_u8(vreinterpret_u8_u16(tmp), 0);
# endif
#endif
}

NLIB_M(i128) I128::SetMask8(int mask) NLIB_NOEXCEPT {
#if defined(NLIB_NEON)
    int8x8_t m = vcreate_s8(0x8040201008040201ULL);
    int8x8_t s0 = vdup_n_s8(mask & 0xFF);
    int8x8_t s1 = vdup_n_s8(mask >> 8);
    return vtstq_s8(vcombine_s8(m, m), vcombine_s8(s0, s1));
#elif defined(NLIB_SSE41)
    i128 m = I128::SetValue(0x8040201008040201ULL, each_uint64);
    i128 s0 = I128::SetValue(mask & 0xFF, each_int8);
    i128 s1 = I128::SetValue(static_cast<int8_t>(mask >> 8), each_int8);
    i128 s = _mm_blend_epi16(s0, s1, 0xF0);
    return I128::Test8(m, s);
#endif
}

NLIB_M(i128) I128::SetMask16(int mask) NLIB_NOEXCEPT {
#if defined(NLIB_NEON)
    uint16x4_t powers_lo = vcreate_u16(0x0008000400020001ULL);
    uint16x4_t powers_hi = vshl_n_u16(powers_lo, 4);
    uint16x8_t powers = vcombine_u16(powers_lo, powers_hi);
    uint16x8_t s = vdupq_n_u16(mask);
    return vreinterpretq_s8_u16(vtstq_u16(powers, s));
#elif defined(NLIB_SSE41)
    i128 m0 = I128::SetValue(0x0008000400020001ULL, each_uint64);
    i128 m1 = I128::SetValue(0x0080004000200010ULL, each_uint64);
    i128 m = _mm_blend_epi16(m0, m1, 0xF0);
    i128 s = I128::SetValue(static_cast<int16_t>(mask), each_int16);
    return I128::Test16(m, s);
#endif
}

NLIB_M(i128) I128::SetMask32(int mask) NLIB_NOEXCEPT {
#if defined(NLIB_NEON)
    uint32x2_t powers_lo = vcreate_u32(0x0000000200000001ULL);
    uint32x2_t powers_hi = vshl_n_u32(powers_lo, 2);
    uint32x4_t powers = vcombine_u32(powers_lo, powers_hi);
    uint32x4_t s = vdupq_n_u32(mask);
    return vreinterpretq_s8_u32(vtstq_u32(powers, s));
#elif defined(NLIB_SSE41)
    i128 m0 = I128::SetValue(0x0000000200000001ULL, each_uint64);
    i128 m1 = I128::SetValue(0x0000000800000004ULL, each_uint64);
    i128 m = _mm_blend_epi16(m0, m1, 0xF0);
    i128 s = I128::SetValue(mask, each_int32);
    return I128::Test32(m, s);
#endif
}

// true if value == 0
NLIB_M(bool) I128::IsZero(i128arg value) NLIB_NOEXCEPT {  // NOLINT
#if defined(NLIB_SSE41)
    return _mm_testz_si128(value, value) != 0;
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    uint32x4_t mask = vceqzq_u32(vreinterpretq_u32_s8(value));
    return vaddvq_s32(vreinterpretq_s32_u32(mask)) == -4;
# else
    // NOTE: there is a better way
    // https://stackoverflow.com/questions/15389539/fastest-way-to-test-a-128-bit-neon-register-for-a-value-of-0-using-intrinsics  // NOLINT
    int8x8_t tmp = vorr_s8(vget_low_s8(value), vget_high_s8(value));
    return vget_lane_u64(vreinterpret_u64_s8(tmp), 0) == 0;
# endif
#endif
}

// true if all bits on
NLIB_M(bool) I128::IsFull(i128arg value) NLIB_NOEXCEPT {  // NOLINT
#if defined(NLIB_SSE41)
    return _mm_testc_si128(value, _mm_cmpeq_epi8(value, value)) != 0;
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    uint32x4_t mask = vceqzq_u32(vreinterpretq_u32_s8(vmvnq_s8(value)));
    return vaddvq_s32(vreinterpretq_s32_u32(mask)) == -4;
# else
    int8x8_t tmp = vand_s8(vget_low_s8(value), vget_high_s8(value));
    return vget_lane_s64(vreinterpret_s64_s8(tmp), 0) == -1;
# endif
#endif
}

// r.i8[i] = ismasked(mask.i8[i]) ? a.i8[i] : b.i8[i]
NLIB_M(i128) I128::Select(i128arg mask, i128arg a, i128arg b) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_blendv_epi8(b, a, mask);
#elif defined(NLIB_NEON)
    return NLIB_OP3(vbslq, u32, mask, a, b);
#endif
}

// r[i] = 0 if shuffle.u8[i] == 0x80, r[i] = value[shuffle.u8[i]] if 0 <= shuffle.u8[i] < 16
NLIB_M(i128) I128::Shuffle8(i128arg value, i128arg shuffle) NLIB_NOEXCEPT {
#if defined(NLIB_SSE41)
    return _mm_shuffle_epi8(value, shuffle);
#elif defined(NLIB_NEON)
# ifdef __aarch64__
    return vqtbl1q_s8(value, vreinterpretq_u8_s8(shuffle));
# else
    int8x8x2_t x;
    x.val[0] = vget_low_s8(value);
    x.val[1] = vget_high_s8(value);
    int8x8_t lo = vtbl2_s8(x, vget_low_s8(shuffle));
    int8x8_t hi = vtbl2_s8(x, vget_high_s8(shuffle));
    return vcombine_s8(lo, hi);
# endif
#endif
}

// popcnt of mask(returns nlib_popcnt16(I128::MoveMask8(value))
NLIB_ALWAYS_INLINE int __vectorcall I128::PopCntMask8(i128arg value) NLIB_NOEXCEPT {
#if defined(NLIB_NEON)
# ifdef __aarch64__
    int8x16_t tmp = vnegq_s8(value);
    return vaddvq_s8(tmp);
# else
    int8x16_t tmp = vnegq_s8(value);
    int8x8_t lo = vget_low_s8(tmp);
    int8x8_t hi = vget_high_s8(tmp);
    lo = vadd_s8(lo, hi);
    lo = vpadd_s8(lo, lo);
    lo = vpadd_s8(lo, lo);
    lo = vpadd_s8(lo, lo);
    return vget_lane_s8(lo, 0);
# endif
#else
    return nlib_popcnt16(static_cast<uint16_t>(I128::MoveMask8(value)));
#endif
}

NLIB_ALWAYS_INLINE int __vectorcall I128::ClzMask8(i128arg value) NLIB_NOEXCEPT {
    return nlib_clz(static_cast<uint32_t>(I128::MoveMask8(value))) - 16;
}

NLIB_ALWAYS_INLINE int __vectorcall I128::CtzMask8(i128arg value) NLIB_NOEXCEPT {
    return nlib_ctz(static_cast<uint32_t>(I128::MoveMask8(value) | 0x10000));
}

#ifdef NLIB_NEON
# undef vreinterpretq_s8_s8
# undef NLIB_OP1
# undef NLIB_OP2
# undef NLIB_OP3
# undef NLIB_CMP
# undef NLIB_SFT
# undef NLIB_CMB
#endif

#endif  // NLIB_DOXYGEN

#undef NLIB_M
#undef NLIB_M2

#if defined(NLIB_SSE41)
#define NLIB_I128_TRANSPOSE32(row0, row1, row2, row3)                \
    {                                                                \
        row0 = _mm_shuffle_epi32(row0, _MM_SHUFFLE(3, 1, 2, 0));     \
        row1 = _mm_shuffle_epi32(row1, _MM_SHUFFLE(3, 1, 2, 0));     \
        row2 = _mm_shuffle_epi32(row2, _MM_SHUFFLE(3, 1, 2, 0));     \
        row3 = _mm_shuffle_epi32(row3, _MM_SHUFFLE(3, 1, 2, 0));     \
        __m128i t0_transpose32_ = _mm_unpacklo_epi32(row0, row1);    \
        __m128i t1_transpose32_ = _mm_unpackhi_epi32(row0, row1);    \
        __m128i t2_transpose32_ = _mm_unpacklo_epi32(row2, row3);    \
        __m128i t3_transpose32_ = _mm_unpackhi_epi32(row2, row3);    \
        row0 = _mm_unpacklo_epi64(t0_transpose32_, t2_transpose32_); \
        row1 = _mm_unpacklo_epi64(t1_transpose32_, t3_transpose32_); \
        row2 = _mm_unpackhi_epi64(t0_transpose32_, t2_transpose32_); \
        row3 = _mm_unpackhi_epi64(t1_transpose32_, t3_transpose32_); \
}
#elif defined(NLIB_NEON)
# ifdef __aarch64__
#define NLIB_I128_TRANSPOSE32(row0, row1, row2, row3)                           \
    {                                                                           \
        uint32x4x2_t trn_f0_ = vtrnq_u32(vreinterpretq_u32_s8(row0),            \
                                         vreinterpretq_u32_s8(row1));           \
        uint32x4x2_t trn_f1_ = vtrnq_u32(vreinterpretq_u32_s8(row2),            \
                                         vreinterpretq_u32_s8(row3));           \
        uint64x2_t row0_, row1_, row2_, row3_;                                  \
        row0_ = vtrn1q_u64(vreinterpretq_u64_u32(trn_f0_.val[0]),               \
                           vreinterpretq_u64_u32(trn_f1_.val[0]));              \
        row0 = vreinterpretq_s8_u64(row0_);                                     \
        row1_ = vtrn1q_u64(vreinterpretq_u64_u32(trn_f0_.val[1]),               \
                           vreinterpretq_u64_u32(trn_f1_.val[1]));              \
        row1 = vreinterpretq_s8_u64(row1_);                                     \
        row2_ = vtrn2q_u64(vreinterpretq_u64_u32(trn_f0_.val[0]),               \
                           vreinterpretq_u64_u32(trn_f1_.val[0]));              \
        row2 = vreinterpretq_s8_u64(row2_);                                     \
        row3_ = vtrn2q_u64(vreinterpretq_u64_u32(trn_f0_.val[1]),               \
                           vreinterpretq_u64_u32(trn_f1_.val[1]));              \
        row3 = vreinterpretq_s8_u64(row3_);                                     \
    }
# else
#define NLIB_I128_TRANSPOSE32(row0, row1, row2, row3)                           \
    {                                                                           \
        uint32x4x2_t trn_f0_ = vtrnq_u32(vreinterpretq_u32_s8(row0),            \
                                         vreinterpretq_u32_s8(row1));           \
        uint32x4x2_t trn_f1_ = vtrnq_u32(vreinterpretq_u32_s8(row2),            \
                                         vreinterpretq_u32_s8(row3));           \
        uint32x4_t row0_, row1_, row2_, row3_;                                  \
        uint32x2_t lo, hi;                                                      \
        lo = vget_low_u32(trn_f0_.val[0]); hi = vget_low_u32(trn_f1_.val[0]);   \
        row0_ = vcombine_u32(lo, hi);                                           \
        row0 = vreinterpretq_s8_u32(row0_);                                     \
        lo = vget_low_u32(trn_f0_.val[1]); hi = vget_low_u32(trn_f1_.val[1]);   \
        row1_ = vcombine_u32(lo, hi);                                           \
        row1 = vreinterpretq_s8_u32(row1_);                                     \
        lo = vget_high_u32(trn_f0_.val[0]); hi = vget_high_u32(trn_f1_.val[0]); \
        row2_ = vcombine_u32(lo, hi);                                           \
        row2 = vreinterpretq_s8_u32(row2_);                                     \
        lo = vget_high_u32(trn_f0_.val[1]); hi = vget_high_u32(trn_f1_.val[1]); \
        row3_ = vcombine_u32(lo, hi);                                           \
        row3 = vreinterpretq_s8_u32(row3_);                                     \
    }
# endif
#endif

#endif  // NLIB_SIMD

}  // namespace simd
NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_SIMD_SIMDINT_H_
