﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_SIMD_SIMDMATRIX_H_
#define INCLUDE_NN_NLIB_SIMD_SIMDMATRIX_H_

#include "nn/nlib/simd/SimdFloat.h"
#include "nn/nlib/simd/SimdVector4.h"
#include "nn/nlib/simd/SimdQuaternion.h"
#include "nn/nlib/simd/SimdVector3.h"
// #include "nn/nlib/simd/SimdGeometry.h"

NLIB_NAMESPACE_BEGIN
namespace simd {

class NLIB_VIS_HIDDEN Matrix {
 public:
    static SimdMatrix __vectorcall LoadFloat4x4(const Float4x4* p) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall LoadFloat3x4(const Float3x4* p) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall LoadFloat4x3(const Float4x3* p) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall LoadFloat3x3(const Float3x3* p) NLIB_NOEXCEPT;
    static void __vectorcall StoreFloat4x4(Float4x4* p, SimdMatrixArg m) NLIB_NOEXCEPT;
    static void __vectorcall StoreFloat3x4(Float3x4* p, SimdMatrixArg m) NLIB_NOEXCEPT;
    static void __vectorcall StoreFloat4x3(Float4x3* p, SimdMatrixArg m) NLIB_NOEXCEPT;
    static void __vectorcall StoreFloat3x3(Float3x3* p, SimdMatrixArg m) NLIB_NOEXCEPT;

    static SimdVector __vectorcall Determinant(SimdMatrixArg m) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall Identity() NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall Inverse(SimdVector* det, SimdMatrixArg m) NLIB_NOEXCEPT;

    static bool __vectorcall IsIdentity(SimdMatrixArg m) NLIB_NOEXCEPT;
    static bool __vectorcall IsInfinite(SimdMatrixArg m) NLIB_NOEXCEPT;
    static bool __vectorcall IsNaN(SimdMatrixArg m) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall Mult(SimdMatrixArg a, SimdMatrixArg b) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall Transpose(SimdMatrixArg m) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall MultTranspose(SimdMatrixArg a, SimdMatrixArg b) NLIB_NOEXCEPT;

    static SimdMatrix __vectorcall FromScaling(float scale_x, float scale_y,
                                               float scale_z) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall FromScaling(SimdVectorArg scale) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall FromTranslation(float ofs_x, float ofs_y,
                                                   float ofs_z) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall FromTranslation(SimdVectorArg ofs) NLIB_NOEXCEPT;

    static SimdMatrix __vectorcall FromRotationX(float sin_value, float cos_value) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall FromRotationY(float sin_value, float cos_value) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall FromRotationZ(float sin_value, float cos_value) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall FromRotationAxisAndSinCos(SimdVectorArg axis_normalized,
                                                             float sin_value,
                                                             float cos_value) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall FromRotationQuaternion(SimdQuaternionArg quat) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall
        FromRotationZXY(SimdVectorArg sin_xyz, SimdVectorArg cos_xyz) NLIB_NOEXCEPT;

    static SimdMatrix __vectorcall LookToLh(SimdVectorArg eye_pos, SimdVectorArg eye_dir_normalized,
                                            SimdVectorArg up_dir_normalized) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall LookAtLh(SimdVectorArg eye_pos, SimdVectorArg at_pos,
                                            SimdVectorArg up_dir_normalized) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall LookToRh(SimdVectorArg eye_pos, SimdVectorArg eye_dir_normalized,
                                            SimdVectorArg up_dir_normalized) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall LookAtRh(SimdVectorArg eye_pos, SimdVectorArg at_pos,
                                            SimdVectorArg up_dir_normalized) NLIB_NOEXCEPT;

    static SimdMatrix __vectorcall
        PerspectiveLh(float width, float height, float near_z, float far_z) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall
        PerspectiveRh(float width, float height, float near_z, float far_z) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall PerspectiveFovLh(float half_fovy_sin, float half_fovy_cos,
                                                    float aspect, float near_z,
                                                    float far_z) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall PerspectiveFovRh(float half_fovy_sin, float half_fovy_cos,
                                                    float aspect, float near_z,
                                                    float far_z) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall PerspectiveOffCenterLh(float left, float right, float bottom,
                                                          float top, float near_z,
                                                          float far_z) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall PerspectiveOffCenterRh(float left, float right, float bottom,
                                                          float top, float near_z,
                                                          float far_z) NLIB_NOEXCEPT;

    static SimdMatrix __vectorcall
        OrthographicLh(float width, float height, float near_z, float far_z) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall
        OrthographicRh(float width, float height, float near_z, float far_z) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall OrthographicOffCenterLh(float left, float right, float bottom,
                                                           float top, float near_z,
                                                           float far_z) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall OrthographicOffCenterRh(float left, float right, float bottom,
                                                           float top, float near_z,
                                                           float far_z) NLIB_NOEXCEPT;

    static SimdMatrix __vectorcall
        Shadow(SimdPlaneArg shadow_plane, SimdVector light_pos) NLIB_NOEXCEPT;
    static SimdMatrix __vectorcall Reflect(SimdPlaneArg reflection_plane) NLIB_NOEXCEPT;
    static void __vectorcall Decompose(SimdVector* scale, SimdMatrix* rot, SimdVector* trans,
                                       SimdMatrixArg m) NLIB_NOEXCEPT;
    // AffineTransform

 private:
    Matrix();  // forbidden
};

#ifndef NLIB_DOXYGEN

#define NLIB_M(tp) inline tp __vectorcall

NLIB_M(SimdMatrix) Matrix::LoadFloat4x4(const Float4x4* p) NLIB_NOEXCEPT {
    SimdMatrix m;
    m.r[0] = F128::LoadA16(&p->m[0][0]);
    m.r[1] = F128::LoadA16(&p->m[1][0]);
    m.r[2] = F128::LoadA16(&p->m[2][0]);
    m.r[3] = F128::LoadA16(&p->m[3][0]);
    return m;
}

// Load and convert to column major order matrix
NLIB_M(SimdMatrix) Matrix::LoadFloat3x4(const Float3x4* p) NLIB_NOEXCEPT {
    SimdMatrix m;
    m.r[0] = F128::LoadA16(&p->m[0][0]);
    m.r[1] = F128::LoadA16(&p->m[1][0]);
    m.r[2] = F128::LoadA16(&p->m[2][0]);
    m.r[3] = F128::Set0001();
    NLIB_F128_TRANSPOSE(m.r[0], m.r[1], m.r[2], m.r[3]);
    return m;
}

NLIB_M(SimdMatrix) Matrix::LoadFloat4x3(const Float4x3* p) NLIB_NOEXCEPT {
    f128 t0 = F128::LoadA16(&p->m[0][0]);
    f128 t1 = F128::LoadA16(&p->m[1][1]);
    f128 t2 = F128::LoadA16(&p->m[2][2]);
    SimdMatrix m;
    m.r[0] = F128::SetZeroToLane<3>(t0);
    f128 tmp1 = F128::Permute<3, 4, 5, -1>(t0, t1);
    m.r[1] = F128::SetZeroToLane<3>(tmp1);
    f128 tmp2 = F128::Permute<2, 3, 4, -1>(t1, t2);
    m.r[2] = F128::SetZeroToLane<3>(tmp2);
    m.r[3] = F128::Permute<1, 2, 3, 7>(t2, F128::SetOne());
    return m;
}

NLIB_M(SimdMatrix) Matrix::LoadFloat3x3(const Float3x3* p) NLIB_NOEXCEPT {
    f128 t0 = F128::LoadA4(&p->m[0][0]);
    f128 t1 = F128::LoadA4(&p->m[1][0]);
    f128 t2 = F128::LoadA4(&p->m[1][2]);
    f128 zero = F128::SetZero();
    SimdMatrix m;
    m.r[0] = F128::SetZeroToLane<3>(t0);
    m.r[1] = F128::SetZeroToLane<3>(t1);
    m.r[2] = F128::Permute<1, 2, 3, 7>(t2, zero);
    m.r[3] = F128::Set0001();
    return m;
}

inline void __vectorcall Matrix::StoreFloat4x4(Float4x4* p, SimdMatrixArg m) NLIB_NOEXCEPT {
    F128::StoreA16(&p->m[0][0], m.r[0]);
    F128::StoreA16(&p->m[1][0], m.r[1]);
    F128::StoreA16(&p->m[2][0], m.r[2]);
    F128::StoreA16(&p->m[3][0], m.r[3]);
}

inline void __vectorcall Matrix::StoreFloat4x3(Float4x3* p, SimdMatrixArg m) NLIB_NOEXCEPT {
    f128 t0 = F128::Permute<0, 1, 2, 4>(m.r[0], m.r[1]);
    f128 t1 = F128::Permute<1, 2, 4, 5>(m.r[1], m.r[2]);
    f128 t2 = F128::Permute<2, 4, 5, 6>(m.r[2], m.r[3]);
    F128::StoreA16(&p->m[0][0], t0);
    F128::StoreA16(&p->m[1][1], t1);
    F128::StoreA16(&p->m[2][2], t2);
}

inline void __vectorcall Matrix::StoreFloat3x3(Float3x3* p, SimdMatrixArg m) NLIB_NOEXCEPT {
    f128 t0 = F128::Permute<0, 1, 2, 4>(m.r[0], m.r[1]);
    f128 t1 = F128::Permute<1, 2, 4, 5>(m.r[1], m.r[2]);
    F128::StoreA4(&p->m[0][0], t0);
    F128::StoreA4(&p->m[1][1], t1);
    p->m[2][2] = F128::GetFloatFromLane<2>(m.r[2]);
}

// XMMatrixDeterminant
// r = determinant(m)
NLIB_M(f128) Matrix::Determinant(SimdMatrixArg m) NLIB_NOEXCEPT {
    // a0 b0 c0 d0
    // a1 b1 c1 d1
    // a2 b2 c2 d2
    // a3 b3 c3 d3
    //
    // a0|bcd123| - b0|acd123| + c0|abd123| + d0|abc123| -> SubAdd & Dot

    // |bcd123| = b1|cd23| - c1|bd23| + d1|bc23| -> lane0
    // |acd123| = a1|cd23| - c1|ad23| + d1|ac23| -> lane1
    // |abd123| = a1|bd23| - b1|ad23| + d1|ab23| -> lane2
    // |abc123| = a1|bc23| - b1|ac23| + c1|ab23| -> lane3

    f128 c0det, c1det, c2det;
    {
        f128 ccbb_2 = F128::Swizzle<2, 2, 1, 1>(m.r[2]);
        f128 ccbb_3 = F128::Swizzle<2, 2, 1, 1>(m.r[3]);

        f128 dddc_2 = F128::Swizzle<3, 3, 3, 2>(m.r[2]);
        f128 dddc_3 = F128::Swizzle<3, 3, 3, 2>(m.r[3]);

        f128 baaa_2 = F128::Swizzle<1, 0, 0, 0>(m.r[2]);
        f128 baaa_3 = F128::Swizzle<1, 0, 0, 0>(m.r[3]);

        f128 tmp0 = F128::Mult(ccbb_2, dddc_3);
        f128 tmp1 = F128::Mult(baaa_2, dddc_3);
        f128 tmp2 = F128::Mult(baaa_2, ccbb_3);

        c0det = F128::MultSub(dddc_2, ccbb_3, tmp0);
        c1det = F128::MultSub(dddc_2, baaa_3, tmp1);
        c2det = F128::MultSub(ccbb_2, baaa_3, tmp2);
    }

    f128 ccbb_1 = F128::Swizzle<2, 2, 1, 1>(m.r[1]);
    f128 baaa_1 = F128::Swizzle<1, 0, 0, 0>(m.r[1]);
    f128 dddc_1 = F128::Swizzle<3, 3, 3, 2>(m.r[1]);
    f128 r0x = F128::NegateEx<true, false, true, false>(m.r[0]);

    f128 det3_neg = F128::Mult(c1det, ccbb_1);
    det3_neg = F128::MultSub(c0det, baaa_1, det3_neg);
    det3_neg = F128::MultSub(c2det, dddc_1, det3_neg);
    return Vector4::Dot(r0x, det3_neg);
}

// XMMatrixIdentity
// r[0] = { 1, 0, 0, 0 }
// r[1] = { 0, 1, 0, 0 }
// r[2] = { 0, 0, 1, 0 }
// r[3] = { 0, 0, 0, 1 }
NLIB_M(SimdMatrix) Matrix::Identity() NLIB_NOEXCEPT {
    SimdMatrix m;
#if defined(NLIB_NEON) && !defined(NLIB_F128_SIMD_NOUSE)
    float32x2_t x10 = vcreate_f32(0x000000003F800000ULL);
    float32x2_t x01 = vcreate_f32(0x3F80000000000000ULL);
    float32x2_t x00 = vcreate_f32(0ULL);
    m.r[0] = vcombine_f32(x10, x00);
    m.r[1] = vcombine_f32(x01, x00);
    m.r[2] = vcombine_f32(x00, x10);
    m.r[3] = vcombine_f32(x00, x01);
#else
    m.r[0] = F128::LoadA16(F128::v1000_);
    m.r[1] = F128::LoadA16(F128::v0100_);
    m.r[2] = F128::LoadA16(F128::v0010_);
    m.r[3] = F128::LoadA16(F128::v0001_);
#endif
    return m;
}

// XMMatrixTranspose
// use NLIB_F128_TRANSPOSE(m.r[0], m.r[1], m.r[2], m.r[3]) if you can change 'm' itself
NLIB_M(SimdMatrix) Matrix::Transpose(SimdMatrixArg m) NLIB_NOEXCEPT {
    SimdMatrix ret;
    f128 r0 = m.r[0];
    f128 r1 = m.r[1];
    f128 r2 = m.r[2];
    f128 r3 = m.r[3];
    NLIB_F128_TRANSPOSE(r0, r1, r2, r3);
    ret.r[0] = r0;
    ret.r[1] = r1;
    ret.r[2] = r2;
    ret.r[3] = r3;
    return ret;
}

inline void __vectorcall Matrix::StoreFloat3x4(Float3x4* p, SimdMatrixArg m) NLIB_NOEXCEPT {
    SimdMatrix M = Matrix::Transpose(m);
    F128::StoreA16(&p->m[0][0], M.r[0]);
    F128::StoreA16(&p->m[1][0], M.r[1]);
    F128::StoreA16(&p->m[2][0], M.r[2]);
}

#if 1
// XMMatrixInverse
NLIB_M(SimdMatrix) Matrix::Inverse(SimdVector* det, SimdMatrixArg m) NLIB_NOEXCEPT {
    // faster on NEON
    // SimdMatrix M = Transpose(m);
    // M:
    // a0 b0 c0 d0
    // a1 b1 c1 d1
    // a2 b2 c2 d2
    // a3 b3 c3 d3

    // Inv:
    // ( |bcd123|, -|acd123|,  |abd123|, -|abc123|) / detm
    // (-|bcd023|,  |acd023|, -|abd023|,  |abc023}) / detm
    // ( |bcd012|, -|acd012|,  |abd012|, -|abc012|) / detm
    // (-|bcd013|,  |acd013|, -|abd013|,  |abc013|) / detm

    // row0:
    // |bcd123| = b1|cd23| - c1|bd23| + d1|bc23| -> lane0
    // |acd123| = a1|cd23| - c1|ad23| + d1|ac23| -> lane1
    // |abd123| = a1|bd23| - b1|ad23| + d1|ab23| -> lane2
    // |abc123| = a1|bc23| - b1|ac23| + c1|ab23| -> lane3

    // row1:
    // |bcd023| = b0|cd23| - c0|bd23| + d0|bc23| -> lane0
    // |acd023| = a0|cd23| - c0|ad23| + d0|ac23| -> lane1
    // |abd023| = a0|bd23| - b0|ad23| + d0|ab23| -> lane2
    // |abc023| = a0|bc23| - b0|ac23| + c0|ab23| -> lane3
    f128 detValueReciprocal;
    SimdMatrix ret;
    f128 mydet;
    {
        f128 c0det, c1det, c2det;
        {
            f128 ccbb_2 = F128::Permute<2, 2, 6, 6>(m.r[2], m.r[1]);
            f128 ccbb_3 = F128::Permute<3, 3, 7, 7>(m.r[2], m.r[1]);

            f128 dddc_2 = F128::Permute<2, 2, 2, 6>(m.r[3], m.r[2]);
            f128 dddc_3 = F128::Permute<3, 3, 3, 7>(m.r[3], m.r[2]);

            f128 baaa_2 = F128::Permute<2, 6, 6, 6>(m.r[1], m.r[0]);
            f128 baaa_3 = F128::Permute<3, 7, 7, 7>(m.r[1], m.r[0]);

            f128 tmp0 = F128::Mult(ccbb_2, dddc_3);
            f128 tmp1 = F128::Mult(baaa_2, dddc_3);
            f128 tmp2 = F128::Mult(baaa_2, ccbb_3);

            c0det = F128::MultSub(dddc_2, ccbb_3, tmp0);
            c1det = F128::MultSub(dddc_2, baaa_3, tmp1);
            c2det = F128::MultSub(ccbb_2, baaa_3, tmp2);
        }
        {
            f128 baaa_1 = F128::Permute<1, 5, 5, 5>(m.r[1], m.r[0]);
            f128 ccbb_1 = F128::Permute<1, 1, 5, 5>(m.r[2], m.r[1]);
            f128 dddc_1 = F128::Permute<1, 1, 1, 5>(m.r[3], m.r[2]);

            f128 r0x_m = F128::Permute<0, -1, 4, -1>(m.r[0], m.r[2]);
            f128 r0x_p = F128::Permute<-1, 0, -1, 4>(m.r[1], m.r[3]);
            f128 r0x = F128::Permute<0, 5, 2, 7>(F128::Negate(r0x_m), r0x_p);

            f128 det3 = F128::Mult(c1det, ccbb_1);
            det3 = F128::MultSub(c0det, baaa_1, det3);
            det3 = F128::MultSub(c2det, dddc_1, det3);

            mydet = Vector4::Dot(r0x, det3);

            det3 = F128::NegateEx<true, false, true, false>(det3);
            detValueReciprocal = F128::Recp(mydet);

            ret.r[0] = F128::Mult(detValueReciprocal, det3);
        }
        {
            f128 baaa_0 = F128::Permute<0, 4, 4, 4>(m.r[1], m.r[0]);
            f128 ccbb_0 = F128::Permute<0, 0, 4, 4>(m.r[2], m.r[1]);
            f128 dddc_0 = F128::Permute<0, 0, 0, 4>(m.r[3], m.r[2]);

            f128 det3 = F128::Mult(c0det, baaa_0);
            det3 = F128::MultAdd(c2det, dddc_0, det3);
            det3 = F128::MultSub(c1det, ccbb_0, det3);
            det3 = F128::NegateEx<true, false, true, false>(det3);
            ret.r[1] = F128::Mult(detValueReciprocal, det3);
        }
    }

    // row3:
    // |bcd012| = b2|cd01| - c2|bd01| + d2|bc01| -> lane0
    // |acd012| = a2|cd01| - c2|ad01| + d2|ac01| -> lane1
    // |abd012| = a2|bd01| - b2|ad01| + d2|ab01| -> lane2
    // |abc012| = a2|bc01| - b2|ac01| + c2|ab01| -> lane3

    // row2:
    // |bcd013| = b3|cd01| - c3|bd01| + d3|bc01| -> lane0
    // |acd013| = a3|cd01| - c3|ad01| + d3|ac01| -> lane1
    // |abd013| = a3|bd01| - b3|ad01| + d3|ab01| -> lane2
    // |abc013| = a3|bc01| - b3|ac01| + c3|ab01| -> lane3
    {
        f128 c0det, c1det, c2det;
        {
            f128 ccbb_0 = F128::Permute<0, 0, 4, 4>(m.r[2], m.r[1]);
            f128 ccbb_1 = F128::Permute<1, 1, 5, 5>(m.r[2], m.r[1]);

            f128 dddc_0 = F128::Permute<0, 0, 0, 4>(m.r[3], m.r[2]);
            f128 dddc_1 = F128::Permute<1, 1, 1, 5>(m.r[3], m.r[2]);

            f128 baaa_0 = F128::Permute<0, 4, 4, 4>(m.r[1], m.r[0]);
            f128 baaa_1 = F128::Permute<1, 5, 5, 5>(m.r[1], m.r[0]);

            f128 tmp0 = F128::Mult(ccbb_0, dddc_1);
            f128 tmp1 = F128::Mult(baaa_0, dddc_1);
            f128 tmp2 = F128::Mult(baaa_0, ccbb_1);

            c0det = F128::MultSub(dddc_0, ccbb_1, tmp0);
            c1det = F128::MultSub(dddc_0, baaa_1, tmp1);
            c2det = F128::MultSub(ccbb_0, baaa_1, tmp2);
        }
        {
            f128 baaa_3 = F128::Permute<3, 7, 7, 7>(m.r[1], m.r[0]);
            f128 ccbb_3 = F128::Permute<3, 3, 7, 7>(m.r[2], m.r[1]);
            f128 dddc_3 = F128::Permute<3, 3, 3, 7>(m.r[3], m.r[2]);

            f128 det3 = F128::Mult(c1det, ccbb_3);
            det3 = F128::MultSub(c0det, baaa_3, det3);
            det3 = F128::MultSub(c2det, dddc_3, det3);
            det3 = F128::NegateEx<true, false, true, false>(det3);

            ret.r[2] = F128::Mult(detValueReciprocal, det3);
        }
        {
            f128 baaa_2 = F128::Permute<2, 6, 6, 6>(m.r[1], m.r[0]);
            f128 ccbb_2 = F128::Permute<2, 2, 6, 6>(m.r[2], m.r[1]);
            f128 dddc_2 = F128::Permute<2, 2, 2, 6>(m.r[3], m.r[2]);

            f128 det3 = F128::Mult(c0det, baaa_2);
            det3 = F128::MultAdd(c2det, dddc_2, det3);
            det3 = F128::MultSub(c1det, ccbb_2, det3);
            det3 = F128::NegateEx<true, false, true, false>(det3);

            ret.r[3] = F128::Mult(detValueReciprocal, det3);
        }
    }
    if (det) {
        *det = mydet;
    }
    return ret;
}
#else
// XMMatrixInverse
NLIB_M(SimdMatrix) Matrix::Inverse(SimdVector* det, SimdMatrixArg m) NLIB_NOEXCEPT {
    SimdMatrix M = Transpose(m);
    // M:
    // a0 b0 c0 d0
    // a1 b1 c1 d1
    // a2 b2 c2 d2
    // a3 b3 c3 d3

    // Inv:
    // ( |bcd123|, -|acd123|,  |abd123|, -|abc123|) / detm
    // (-|bcd023|,  |acd023|, -|abd023|,  |abc023}) / detm
    // ( |bcd012|, -|acd012|,  |abd012|, -|abc012|) / detm
    // (-|bcd013|,  |acd013|, -|abd013|,  |abc013|) / detm

    // row0:
    // |bcd123| = b1|cd23| - c1|bd23| + d1|bc23| -> lane0
    // |acd123| = a1|cd23| - c1|ad23| + d1|ac23| -> lane1
    // |abd123| = a1|bd23| - b1|ad23| + d1|ab23| -> lane2
    // |abc123| = a1|bc23| - b1|ac23| + c1|ab23| -> lane3

    // row1:
    // |bcd023| = b0|cd23| - c0|bd23| + d0|bc23| -> lane0
    // |acd023| = a0|cd23| - c0|ad23| + d0|ac23| -> lane1
    // |abd023| = a0|bd23| - b0|ad23| + d0|ab23| -> lane2
    // |abc023| = a0|bc23| - b0|ac23| + c0|ab23| -> lane3
    f128 detValueReciprocal;
    SimdMatrix ret;
    {
        f128 c0det, c1det, c2det;
        {
            f128 ccbb_2 = F128::Swizzle<2, 2, 1, 1>(M.r[2]);
            f128 ccbb_3 = F128::Swizzle<2, 2, 1, 1>(M.r[3]);

            f128 dddc_2 = F128::Swizzle<3, 3, 3, 2>(M.r[2]);
            f128 dddc_3 = F128::Swizzle<3, 3, 3, 2>(M.r[3]);

            f128 baaa_2 = F128::Swizzle<1, 0, 0, 0>(M.r[2]);
            f128 baaa_3 = F128::Swizzle<1, 0, 0, 0>(M.r[3]);

            f128 tmp0 = F128::Mult(ccbb_2, dddc_3);
            f128 tmp1 = F128::Mult(baaa_2, dddc_3);
            f128 tmp2 = F128::Mult(baaa_2, ccbb_3);

            c0det = F128::MultSub(dddc_2, ccbb_3, tmp0);
            c1det = F128::MultSub(dddc_2, baaa_3, tmp1);
            c2det = F128::MultSub(ccbb_2, baaa_3, tmp2);
        }
        {
            f128 baaa_1 = F128::Swizzle<1, 0, 0, 0>(M.r[1]);
            f128 ccbb_1 = F128::Swizzle<2, 2, 1, 1>(M.r[1]);
            f128 dddc_1 = F128::Swizzle<3, 3, 3, 2>(M.r[1]);
            f128 r0x = F128::NegateEx<true, false, true, false>(M.r[0]);

            f128 det3 = F128::Mult(c1det, ccbb_1);
            det3 = F128::MultSub(c0det, baaa_1, det3);
            det3 = F128::MultSub(c2det, dddc_1, det3);

            detValueReciprocal = Vector4::Dot(r0x, det3);
            if (det) {
                *det = detValueReciprocal;
            }

            det3 = F128::NegateEx<true, false, true, false>(det3);
            detValueReciprocal = F128::Recp(detValueReciprocal);

            ret.r[0] = F128::Mult(detValueReciprocal, det3);
        }
        {
            f128 baaa_0 = F128::Swizzle<1, 0, 0, 0>(M.r[0]);
            f128 ccbb_0 = F128::Swizzle<2, 2, 1, 1>(M.r[0]);
            f128 dddc_0 = F128::Swizzle<3, 3, 3, 2>(M.r[0]);

            f128 det3 = F128::Mult(c0det, baaa_0);
            det3 = F128::MultAdd(c2det, dddc_0, det3);
            det3 = F128::MultSub(c1det, ccbb_0, det3);
            det3 = F128::NegateEx<true, false, true, false>(det3);
            ret.r[1] = F128::Mult(detValueReciprocal, det3);
        }
    }

    // row3:
    // |bcd012| = b2|cd01| - c2|bd01| + d2|bc01| -> lane0
    // |acd012| = a2|cd01| - c2|ad01| + d2|ac01| -> lane1
    // |abd012| = a2|bd01| - b2|ad01| + d2|ab01| -> lane2
    // |abc012| = a2|bc01| - b2|ac01| + c2|ab01| -> lane3

    // row2:
    // |bcd013| = b3|cd01| - c3|bd01| + d3|bc01| -> lane0
    // |acd013| = a3|cd01| - c3|ad01| + d3|ac01| -> lane1
    // |abd013| = a3|bd01| - b3|ad01| + d3|ab01| -> lane2
    // |abc013| = a3|bc01| - b3|ac01| + c3|ab01| -> lane3
    {
        f128 c0det, c1det, c2det;
        {
            f128 ccbb_0 = F128::Swizzle<2, 2, 1, 1>(M.r[0]);
            f128 ccbb_1 = F128::Swizzle<2, 2, 1, 1>(M.r[1]);

            f128 dddc_0 = F128::Swizzle<3, 3, 3, 2>(M.r[0]);
            f128 dddc_1 = F128::Swizzle<3, 3, 3, 2>(M.r[1]);

            f128 baaa_0 = F128::Swizzle<1, 0, 0, 0>(M.r[0]);
            f128 baaa_1 = F128::Swizzle<1, 0, 0, 0>(M.r[1]);

            f128 tmp0 = F128::Mult(ccbb_0, dddc_1);
            f128 tmp1 = F128::Mult(baaa_0, dddc_1);
            f128 tmp2 = F128::Mult(baaa_0, ccbb_1);

            c0det = F128::MultSub(dddc_0, ccbb_1, tmp0);
            c1det = F128::MultSub(dddc_0, baaa_1, tmp1);
            c2det = F128::MultSub(ccbb_0, baaa_1, tmp2);
        }
        {
            f128 baaa_3 = F128::Swizzle<1, 0, 0, 0>(M.r[3]);
            f128 ccbb_3 = F128::Swizzle<2, 2, 1, 1>(M.r[3]);
            f128 dddc_3 = F128::Swizzle<3, 3, 3, 2>(M.r[3]);

            f128 det3 = F128::Mult(c1det, ccbb_3);
            det3 = F128::MultSub(c0det, baaa_3, det3);
            det3 = F128::MultSub(c2det, dddc_3, det3);
            det3 = F128::NegateEx<true, false, true, false>(det3);

            ret.r[2] = F128::Mult(detValueReciprocal, det3);
        }
        {
            f128 baaa_2 = F128::Swizzle<1, 0, 0, 0>(M.r[2]);
            f128 ccbb_2 = F128::Swizzle<2, 2, 1, 1>(M.r[2]);
            f128 dddc_2 = F128::Swizzle<3, 3, 3, 2>(M.r[2]);

            f128 det3 = F128::Mult(c0det, baaa_2);
            det3 = F128::MultAdd(c2det, dddc_2, det3);
            det3 = F128::MultSub(c1det, ccbb_2, det3);
            det3 = F128::NegateEx<true, false, true, false>(det3);

            ret.r[3] = F128::Mult(detValueReciprocal, det3);
        }
    }
    return ret;
}
#endif

// XMMatrixIsIdentity
// true if m is an indentity matrix
NLIB_M(bool) Matrix::IsIdentity(SimdMatrixArg m) NLIB_NOEXCEPT {  // NOLINT
    SimdMatrix x = Matrix::Identity();
    f128 cmp0 = F128::CmpEq(x.r[0], m.r[0]);
    f128 cmp1 = F128::CmpEq(x.r[1], m.r[1]);
    f128 cmp2 = F128::CmpEq(x.r[2], m.r[2]);
    f128 cmp3 = F128::CmpEq(x.r[3], m.r[3]);
    cmp0 = F128::And(cmp0, cmp1);
    cmp2 = F128::And(cmp2, cmp3);
    cmp0 = F128::And(cmp0, cmp2);
    return F128::IsAllMaskTrue(cmp0);
}

// XMMatrixIsInfinite
// true if there is (i, j) which satisfies isinf(m[i][j])
NLIB_M(bool) Matrix::IsInfinite(SimdMatrixArg m) NLIB_NOEXCEPT {  // NOLINT
#ifdef NLIB_F128_SIMD_NOUSE
    f128 cmp0 = F128::IsInfinite(m.r[0]);
    f128 cmp1 = F128::IsInfinite(m.r[1]);
    f128 cmp2 = F128::IsInfinite(m.r[2]);
    f128 cmp3 = F128::IsInfinite(m.r[3]);
    cmp0 = F128::Or(cmp0, cmp1);
    cmp2 = F128::Or(cmp2, cmp3);
    cmp0 = F128::Or(cmp0, cmp2);
    return !F128::IsAllMaskFalse(cmp0);
#else
    f128 inf_value = F128::SetInfinity();
    f128 cmp0 = F128::CmpEq(inf_value, F128::Abs(m.r[0]));
    f128 cmp1 = F128::CmpEq(inf_value, F128::Abs(m.r[1]));
    f128 cmp2 = F128::CmpEq(inf_value, F128::Abs(m.r[2]));
    f128 cmp3 = F128::CmpEq(inf_value, F128::Abs(m.r[3]));
    cmp0 = F128::Or(cmp0, cmp1);
    cmp2 = F128::Or(cmp2, cmp3);
    cmp0 = F128::Or(cmp0, cmp2);
    return !F128::IsAllMaskFalse(cmp0);
#endif
}

// XMMatrixIsNaN
// true if there is (i, j) which satisfies isnan(m[i][j])
NLIB_M(bool) Matrix::IsNaN(SimdMatrixArg m) NLIB_NOEXCEPT {  // NOLINT
    f128 cmp0 = F128::IsNaN(m.r[0]);
    f128 cmp1 = F128::IsNaN(m.r[1]);
    f128 cmp2 = F128::IsNaN(m.r[2]);
    f128 cmp3 = F128::IsNaN(m.r[3]);
    cmp0 = F128::Or(cmp0, cmp1);
    cmp2 = F128::Or(cmp2, cmp3);
    cmp0 = F128::Or(cmp0, cmp2);
    return !F128::IsAllMaskFalse(cmp0);
}

// XMMatrixMultiply
// r = a * b
NLIB_M(SimdMatrix) Matrix::Mult(SimdMatrixArg a, SimdMatrixArg b) NLIB_NOEXCEPT {
    SimdMatrix m;
    m.r[0] = Vector4::Transform(a.r[0], b);
    m.r[1] = Vector4::Transform(a.r[1], b);
    m.r[2] = Vector4::Transform(a.r[2], b);
    m.r[3] = Vector4::Transform(a.r[3], b);
    return m;
}

// XMMatrixMultiplyTranspose
// r = transpose(a * b)
NLIB_M(SimdMatrix) Matrix::MultTranspose(SimdMatrixArg a, SimdMatrixArg b) NLIB_NOEXCEPT {
    f128 r0 = Vector4::Transform(a.r[0], b);
    f128 r1 = Vector4::Transform(a.r[1], b);
    f128 r2 = Vector4::Transform(a.r[2], b);
    f128 r3 = Vector4::Transform(a.r[3], b);
    SimdMatrix ret;
    NLIB_F128_TRANSPOSE(r0, r1, r2, r3);
    ret.r[0] = r0;
    ret.r[1] = r1;
    ret.r[2] = r2;
    ret.r[3] = r3;
    return ret;
}

// XMMatrixScaling
// r[0] = { x, 0, 0, 0 }
// r[1] = { 0, y, 0, 0 }
// r[2] = { 0, 0, z, 0 }
// r[3] = { 0, 0, 0, 1 }
NLIB_M(SimdMatrix) Matrix::FromScaling(float scale_x, float scale_y, float scale_z) NLIB_NOEXCEPT {
    SimdMatrix m;
    f128 zero = F128::SetZero();
    m.r[0] = F128::SetFloatToLane<0>(zero, scale_x);
    m.r[1] = F128::SetFloatToLane<1>(zero, scale_y);
    m.r[2] = F128::SetFloatToLane<2>(zero, scale_z);
    m.r[3] = F128::Set0001();
    return m;
}

// XMMatrixScalingFromVector
// r[0] = { x, 0, 0, 0 }
// r[1] = { 0, y, 0, 0 }
// r[2] = { 0, 0, z, 0 }
// r[3] = { 0, 0, 0, 1 }
NLIB_M(SimdMatrix) Matrix::FromScaling(SimdVectorArg scale) NLIB_NOEXCEPT {
    SimdMatrix m;
    f128 zero = F128::SetZero();
    m.r[0] = F128::Splat<false, true, true, true>(scale, zero);
    m.r[1] = F128::Splat<true, false, true, true>(scale, zero);
    m.r[2] = F128::Splat<true, true, false, true>(scale, zero);
    m.r[3] = F128::Set0001();
    return m;
}

// XMMatrixTranslation
// r[0] = { 1, 0, 0, 0 }
// r[1] = { 0, 1, 0, 0 }
// r[2] = { 0, 0, 1, 0 }
// r[3] = { x, y, z, 1 }
NLIB_M(SimdMatrix) Matrix::FromTranslation(float ofs_x, float ofs_y, float ofs_z) NLIB_NOEXCEPT {
    SimdMatrix m;
    m.r[0] = F128::Set1000();
    m.r[1] = F128::Set0100();
    m.r[2] = F128::Set0010();
    m.r[3] = F128::SetValue(ofs_x, ofs_y, ofs_z, 1.f);
    return m;
}

// XMMatrixTranslationFromVector
// r[0] = { 1, 0, 0, 0 }
// r[1] = { 0, 1, 0, 0 }
// r[2] = { 0, 0, 1, 0 }
// r[3] = { x, y, z, 1 }
NLIB_M(SimdMatrix) Matrix::FromTranslation(SimdVectorArg ofs) NLIB_NOEXCEPT {
    SimdMatrix m;
    m.r[0] = F128::Set1000();
    m.r[1] = F128::Set0100();
    m.r[2] = F128::Set0010();
    m.r[3] = F128::Permute<0, 1, 2, 4>(ofs, m.r[0]);
    return m;
}

// XMMatrixRotationX
// r[0] = { 1 0 0 0 }
// r[1] = { 0 c s 0 }
// r[2] = { 0 -s c 0 }
// r[3] = { 0 0 0 1 }
NLIB_M(SimdMatrix) Matrix::FromRotationX(float sin_value, float cos_value) NLIB_NOEXCEPT {
    SimdMatrix m;
    SimdVector zero = F128::SetZero();
    f128 r1 = F128::SetFloatToLane<1>(zero, cos_value);
    r1 = F128::SetFloatToLane<2>(r1, sin_value);
    f128 r2 = F128::SetFloatToLane<1>(zero, -sin_value);
    r2 = F128::SetFloatToLane<2>(r2, cos_value);

    m.r[0] = F128::Set1000();
    m.r[1] = r1;
    m.r[2] = r2;
    m.r[3] = F128::Set0001();
    return m;
}

// XMMatrixRotationY
// r[0] = { c 0 -s 0 }
// r[1] = { 0 1 0 0 }
// r[2] = { s 0 c 0 }
// r[3] = { 0 0 0 1 }
NLIB_M(SimdMatrix) Matrix::FromRotationY(float sin_value, float cos_value) NLIB_NOEXCEPT {
    SimdMatrix m;
    SimdVector zero = F128::SetZero();
    f128 r0 = F128::SetFloatToLane<0>(zero, cos_value);
    r0 = F128::SetFloatToLane<2>(r0, -sin_value);
    f128 r2 = F128::SetFloatToLane<0>(zero, sin_value);
    r2 = F128::SetFloatToLane<2>(r2, cos_value);

    m.r[0] = r0;
    m.r[1] = F128::Set0100();
    m.r[2] = r2;
    m.r[3] = F128::Set0001();
    return m;
}

// XMMatrixRotationZ
// r[0] = { c s 0 0 }
// r[1] = { -s c 0 0 }
// r[2] = { 0 0 1 0 }
// r[3] = { 0 0 0 1 }
NLIB_M(SimdMatrix) Matrix::FromRotationZ(float sin_value, float cos_value) NLIB_NOEXCEPT {
    SimdMatrix m;
    SimdVector zero = F128::SetZero();
    f128 r0 = F128::SetFloatToLane<0>(zero, cos_value);
    r0 = F128::SetFloatToLane<1>(r0, sin_value);
    f128 r1 = F128::SetFloatToLane<0>(zero, -sin_value);
    r1 = F128::SetFloatToLane<1>(r1, cos_value);

    m.r[0] = r0;
    m.r[1] = r1;
    m.r[2] = F128::Set0010();
    m.r[3] = F128::Set0001();
    return m;
}

// XMMatrixRotationAxis
// The result may be different from DirectXMath's XMMatrixRotationAxis.
// It is because the calculation order is different.
NLIB_M(SimdMatrix) Matrix::FromRotationAxisAndSinCos(SimdVectorArg axis_normalized, float sin_value,
                                                     float cos_value) NLIB_NOEXCEPT {
    // m00, m11, m22, *
    f128 diagonal, c1;
    {
        f128 nn = F128::Mult(axis_normalized, axis_normalized);
        f128 c = F128::SetValue(cos_value, each_float);
        c1 = F128::SetValue(1.f - cos_value, each_float);
        diagonal = F128::MultAdd(c1, nn, c);
        diagonal = F128::SetZeroToLane<3>(diagonal);
    }

    f128 zxy = F128::Swizzle<2, 0, 1, 2>(axis_normalized);
    f128 s = F128::SetValue(sin_value, each_float);
    f128 xy_yz_xz = F128::Mult(axis_normalized, F128::Swizzle<1, 2, 0, 3>(axis_normalized));
    xy_yz_xz = F128::Mult(c1, xy_yz_xz);
    f128 plus = F128::MultAdd(s, zxy, xy_yz_xz);   // xy(1-c)+sz, yz(1-c)+sx, xz(1-c)+sy
    f128 minus = F128::MultSub(s, zxy, xy_yz_xz);  // xy(1-c)-sz, yz(1-c)-sx, xz(1-c)-sy

    f128 t0 = F128::Permute<1, 2, 4, 5>(plus, minus);
    f128 t1 = F128::Permute<0, 6, 2, 3>(plus, minus);

    SimdMatrix m;
    m.r[0] = F128::Permute<4, 0, 1, 7>(t1, diagonal);
    m.r[1] = F128::Permute<2, 5, 0, 7>(t0, diagonal);
    m.r[2] = F128::Permute<1, 3, 6, 7>(t0, diagonal);
    m.r[3] = F128::Set0001();
    return m;
}

// XMMatrixRotationQuaternion
// the result may differ from directxmath's one because the computation sequence is different.
NLIB_M(SimdMatrix) Matrix::FromRotationQuaternion(SimdQuaternionArg quat) NLIB_NOEXCEPT {
    // m00, m11, m22, *
    f128 q2 = F128::Add(quat, quat);
    f128 qq2 = F128::Mult(quat, q2);
    f128 t0, t1;

    t0 = F128::Swizzle<1, 0, 0, -1>(qq2);  // 2y^2, 2x^2, 2x^2, *
    t1 = F128::Swizzle<2, 2, 1, -1>(qq2);  // 2z^2, 2z^2, 2y^2, *
    // 1-2y^2-2z^2, 1-2x^2-2z^2, 1-2x^2-2y^2, 0
    f128 diagonal = F128::Sub(F128::Sub(F128::SetOne(), t0), t1);
    diagonal = F128::SetFloatToLane<3>(diagonal, 0.f);

    t0 = F128::Swizzle<1, 0, 0, -1>(quat);  //  y,  x,  x, *
    t1 = F128::Swizzle<2, 2, 1, -1>(q2);    // 2z, 2z, 2y, *
    f128 yz_xz_xy = F128::Mult(t0, t1);    // 2yz, 2xz, 2xy, *

    t0 = F128::SetValue<3>(quat, each_select32);  // w, w, w, *
    f128 wx_wy_wz = F128::Mult(q2, t0);           // 2wx, 2wy, 2wz

    f128 plus = F128::Add(yz_xz_xy, wx_wy_wz);   // 2yz+2wx, 2xz+2wy, 2xy+2wz, *
    f128 minus = F128::Sub(yz_xz_xy, wx_wy_wz);  // 2yz-2wx, 2xz-2wy, 2xy-2wz, *

    t0 = F128::Permute<1, 2, 4, 5>(plus, minus);
    t1 = F128::Permute<0, 6, 2, 3>(plus, minus);

    SimdMatrix m;
    m.r[0] = F128::Permute<4, 1, 3, 7>(t0, diagonal);
    m.r[1] = F128::Permute<1, 5, 0, 7>(t1, diagonal);
    m.r[2] = F128::Permute<0, 2, 6, 7>(t0, diagonal);
    m.r[3] = F128::Set0001();
    return m;
}

// XMMatrixRotationRollPitchYaw
NLIB_M(SimdMatrix) Matrix::FromRotationZXY(SimdVectorArg sin_xyz,
                                           SimdVectorArg cos_xyz) NLIB_NOEXCEPT {
    // CzCy+SzSxSy  SzCx  -CzSy+SzSxCy  0
    // -SzCy+CzSxSy CzCx   SzSy+CzSxCy  0
    // CxSy         -Sx    CxCy         0
    f128 m00_12_02_10;
    {
        f128 sz_cz_sz_cz = F128::Permute<2, 6, 2, 6>(sin_xyz, cos_xyz);
        f128 sy_cy_cy_sy = F128::Permute<1, 5, 5, 1>(sin_xyz, cos_xyz);
        f128 tmp = F128::Mult(sz_cz_sz_cz, sy_cy_cy_sy);
        m00_12_02_10 = F128::Mult<0>(sin_xyz, tmp, each_select32);
        tmp = F128::Swizzle<1, 0, 3, 2>(tmp);
        tmp = F128::NegateEx<false, false, true, true>(tmp);
        m00_12_02_10 = F128::Add(tmp, m00_12_02_10);
    }
    f128 m20_01_22_11;
    {
        f128 sy_sz_cy_cz = F128::Permute<1, 2, 5, 6>(sin_xyz, cos_xyz);
        m20_01_22_11 = F128::Mult<0>(cos_xyz, sy_sz_cy_cz, each_select32);
    }

    f128 r2 = F128::SetFloatToLane<3>(m20_01_22_11, 0.f);
    f128 r1 = F128::Permute<3, 7, 1, 1>(m00_12_02_10, m20_01_22_11);

    SimdMatrix m;
    m.r[0] = F128::Permute<0, 5, 2, 7>(m00_12_02_10, r2);
    m.r[1] = F128::SetZeroToLane<3>(r1);
    m.r[2] = F128::SetFloatToLane<1>(r2, -F128::GetFloatFromLane<0>(sin_xyz));
    m.r[3] = F128::Set0001();
    return m;
}

// XMMatrixLookToLH
NLIB_M(SimdMatrix) Matrix::LookToLh(SimdVectorArg eye_pos, SimdVectorArg eye_dir_normalized,
                                    SimdVectorArg up_dir_normalized) NLIB_NOEXCEPT {
    SimdVector r0 = Vector3::Cross(up_dir_normalized, eye_dir_normalized);
    SimdVector r1 = Vector3::Cross(eye_dir_normalized, r0);
    SimdVector neg = F128::Negate(eye_pos);
#ifdef NLIB_NEON
    neg = F128::SetZeroToLane<3>(neg);
    f128 d012 = Vector4::Dot3(neg, r0, r1, eye_dir_normalized);
    SimdMatrix m;
    m.r[0] = r0;
    m.r[1] = r1;
    m.r[2] = eye_dir_normalized;
    m.r[3] = F128::Set0001();
    m = Transpose(m);
    m.r[3] = F128::SetFloatToLane<3>(d012, 1.f);
    return m;
#else
    f128 d0 = Vector3::Dot(r0, neg);
    f128 d1 = Vector3::Dot(r1, neg);
    f128 d2 = Vector3::Dot(eye_dir_normalized, neg);
    SimdMatrix m;
    m.r[0] = F128::Splat<false, false, false, true>(r0, d0);
    m.r[1] = F128::Splat<false, false, false, true>(r1, d1);
    m.r[2] = F128::Splat<false, false, false, true>(eye_dir_normalized, d2);
    m.r[3] = F128::Set0001();
    return Transpose(m);
#endif
}

// XMMatrixLookAtLH
NLIB_M(SimdMatrix) Matrix::LookAtLh(SimdVectorArg eye_pos, SimdVectorArg at_pos,
                                    SimdVectorArg up_dir_normalized) NLIB_NOEXCEPT {
    SimdVector eye_dir = F128::Sub(at_pos, eye_pos);
    eye_dir = Vector3::Normalize(eye_dir);
    return LookToLh(eye_pos, eye_dir, up_dir_normalized);
}

// XMMatrixLookToRH
NLIB_M(SimdMatrix) Matrix::LookToRh(SimdVectorArg eye_pos, SimdVectorArg eye_dir_normalized,
                                    SimdVectorArg up_dir_normalized) NLIB_NOEXCEPT {
    return LookToLh(eye_pos, F128::Negate(eye_dir_normalized), up_dir_normalized);
}

// XMMatrixLookAtRH
NLIB_M(SimdMatrix) Matrix::LookAtRh(SimdVectorArg eye_pos, SimdVectorArg at_pos,
                                    SimdVectorArg up_dir_normalized) NLIB_NOEXCEPT {
    SimdVector eye_dir = F128::Sub(eye_pos, at_pos);
    eye_dir = Vector3::Normalize(eye_dir);
    return LookToLh(eye_pos, eye_dir, up_dir_normalized);
}

// XMMatrixPerspectiveLH
NLIB_M(SimdMatrix) Matrix::PerspectiveLh(float width, float height, float near_z,
                                         float far_z) NLIB_NOEXCEPT {
    float near2 = near_z + near_z;
    float range = far_z / (far_z - near_z);
    f128 zero = F128::SetZero();
    f128 v = F128::SetValue(near2 / width, near2 / height, range, -range * near_z);
    SimdMatrix m;
    m.r[0] = F128::Splat<false, true, true, true>(v, zero);
    m.r[1] = F128::Splat<true, false, true, true>(v, zero);
    f128 tmp = F128::Splat<true, true, false, true>(v, zero);
    m.r[2] = F128::SetFloatToLane<3>(tmp, 1.f);
    m.r[3] = F128::Permute<0, 1, 7, 0>(zero, v);
    return m;
}

// XMMatrixPerspectiveRH
NLIB_M(SimdMatrix) Matrix::PerspectiveRh(float width, float height, float near_z,
                                         float far_z) NLIB_NOEXCEPT {
    float near2 = near_z + near_z;
    float range = far_z / (near_z - far_z);
    f128 zero = F128::SetZero();
    f128 v = F128::SetValue(near2 / width, near2 / height, range, range * near_z);
    SimdMatrix m;
    m.r[0] = F128::Splat<false, true, true, true>(v, zero);
    m.r[1] = F128::Splat<true, false, true, true>(v, zero);
    f128 tmp = F128::Splat<true, true, false, true>(v, zero);
    m.r[2] = F128::SetFloatToLane<3>(tmp, -1.f);
    m.r[3] = F128::Permute<0, 1, 7, 0>(zero, v);
    return m;
}

// XMMatrixPerspectiveFovLH
NLIB_M(SimdMatrix) Matrix::PerspectiveFovLh(float half_fovy_sin, float half_fovy_cos, float aspect,
                                            float near_z, float far_z) NLIB_NOEXCEPT {
    float height = half_fovy_cos / half_fovy_sin;
    float width = height / aspect;
    float range = far_z / (far_z - near_z);

    f128 zero = F128::SetZero();
    f128 v = F128::SetValue(width, height, range, -range * near_z);
    SimdMatrix m;
    m.r[0] = F128::Splat<false, true, true, true>(v, zero);
    m.r[1] = F128::Splat<true, false, true, true>(v, zero);
    f128 tmp = F128::Splat<true, true, false, true>(v, zero);
    m.r[2] = F128::SetFloatToLane<3>(tmp, 1.f);
    m.r[3] = F128::Permute<0, 1, 7, 0>(zero, v);
    return m;
}

// XMMatrixPerspectiveFovRH
NLIB_M(SimdMatrix) Matrix::PerspectiveFovRh(float half_fovy_sin, float half_fovy_cos, float aspect,
                                            float near_z, float far_z) NLIB_NOEXCEPT {
    float height = half_fovy_cos / half_fovy_sin;
    float width = height / aspect;
    float range = far_z / (near_z - far_z);

    f128 zero = F128::SetZero();
    f128 v = F128::SetValue(width, height, range, range * near_z);
    SimdMatrix m;
    m.r[0] = F128::Splat<false, true, true, true>(v, zero);
    m.r[1] = F128::Splat<true, false, true, true>(v, zero);
    f128 tmp = F128::Splat<true, true, false, true>(v, zero);
    m.r[2] = F128::SetFloatToLane<3>(tmp, -1.f);
    m.r[3] = F128::Permute<0, 1, 7, 0>(zero, v);
    return m;
}

// XMMatrixPerspectiveOffCenterLH
NLIB_M(SimdMatrix) Matrix::PerspectiveOffCenterLh(float left, float right, float bottom, float top,
                                                  float near_z, float far_z) NLIB_NOEXCEPT {
    float near2 = near_z + near_z;
    f128 div;
    {
        f128 a = F128::SetValue(1.f, 1.f, far_z, 1.f);
        f128 b = F128::SetValue(right - left, top - bottom, far_z - near_z, 1.f);
        div = F128::Div(a, b);
        // recpWidth, recpHeight, range, 1.f
    }
    f128 zero = F128::SetZero();
    f128 v0 = F128::SetValue(near2, near2, -near_z, 1.f);
    f128 r2 = F128::SetValue(-(left + right), -(top + bottom), 1.f, 1.f);
    v0 = F128::Mult(v0, div);

    SimdMatrix m;
    m.r[0] = F128::Splat<false, true, true, true>(v0, zero);
    m.r[1] = F128::Splat<true, false, true, true>(v0, zero);
    m.r[2] = F128::Mult(r2, div);
    m.r[3] = F128::Splat<true, true, false, true>(v0, zero);
    return m;
}

// XMMatrixPerspectiveOffCenterRH
NLIB_M(SimdMatrix) Matrix::PerspectiveOffCenterRh(float left, float right, float bottom, float top,
                                                  float near_z, float far_z) NLIB_NOEXCEPT {
    float near2 = near_z + near_z;
    f128 div;
    {
        f128 a = F128::SetValue(1.f, 1.f, far_z, 1.f);
        f128 b = F128::SetValue(right - left, top - bottom, near_z - far_z, -1.f);
        div = F128::Div(a, b);
        // recpWidth, recpHeight, range, 1.f
    }
    f128 zero = F128::SetZero();
    f128 v0 = F128::SetValue(near2, near2, near_z, 1.f);
    f128 r2 = F128::SetValue((left + right), (top + bottom), 1.f, 1.f);
    v0 = F128::Mult(v0, div);

    SimdMatrix m;
    m.r[0] = F128::Splat<false, true, true, true>(v0, zero);
    m.r[1] = F128::Splat<true, false, true, true>(v0, zero);
    m.r[2] = F128::Mult(r2, div);
    m.r[3] = F128::Splat<true, true, false, true>(v0, zero);
    return m;
}

// XMMatrixOrthographicLH
NLIB_M(SimdMatrix) Matrix::OrthographicLh(float width, float height, float near_z,
                                          float far_z) NLIB_NOEXCEPT {
    f128 div;
    {
        f128 a = F128::SetValue(2.f, 2.f, 1.f, -near_z);
        f128 b = F128::SetValue(width, height, far_z - near_z, far_z - near_z);
        div = F128::Div(a, b);
    }
    f128 zero = F128::SetZero();

    SimdMatrix m;
    m.r[0] = F128::Splat<false, true, true, true>(div, zero);
    m.r[1] = F128::Splat<true, false, true, true>(div, zero);
    m.r[2] = F128::Splat<true, true, false, true>(div, zero);
    f128 tmp = F128::Permute<0, 1, 7, 1>(zero, div);
    m.r[3] = F128::SetFloatToLane<3>(tmp, 1.f);
    return m;
}

// XMMatrixOrthographicRH
NLIB_M(SimdMatrix) Matrix::OrthographicRh(float width, float height, float near_z,
                                          float far_z) NLIB_NOEXCEPT {
    f128 div;
    {
        f128 a = F128::SetValue(2.f, 2.f, 1.f, near_z);
        f128 b = F128::SetValue(width, height, near_z - far_z, near_z - far_z);
        div = F128::Div(a, b);
    }
    f128 zero = F128::SetZero();

    SimdMatrix m;
    m.r[0] = F128::Splat<false, true, true, true>(div, zero);
    m.r[1] = F128::Splat<true, false, true, true>(div, zero);
    m.r[2] = F128::Splat<true, true, false, true>(div, zero);
    f128 tmp = F128::Permute<0, 1, 7, 1>(zero, div);
    m.r[3] = F128::SetFloatToLane<3>(tmp, 1.f);
    return m;
}

// XMMatrixOrthographicOffCenterLH
NLIB_M(SimdMatrix) Matrix::OrthographicOffCenterLh(float left, float right, float bottom, float top,
                                                   float near_z, float far_z) NLIB_NOEXCEPT {
    f128 div;
    {
        f128 a = F128::SetOne();
        f128 b = F128::SetValue(right - left, top - bottom, far_z - near_z, 1.f);
        div = F128::Div(a, b);
    }
    f128 zero = F128::SetZero();
    f128 v0 = F128::SetValue(2.f, 2.f, 1.f, 1.f);
    f128 r3 = F128::SetValue(-(left + right), -(top + bottom), -near_z, 1.f);
    v0 = F128::Mult(v0, div);

    SimdMatrix m;
    m.r[0] = F128::Splat<false, true, true, true>(v0, zero);
    m.r[1] = F128::Splat<true, false, true, true>(v0, zero);
    m.r[2] = F128::Splat<true, true, false, true>(v0, zero);
    m.r[3] = F128::Mult(r3, div);
    return m;
}

// XMMatrixOrthographicOffCenterRH
NLIB_M(SimdMatrix) Matrix::OrthographicOffCenterRh(float left, float right, float bottom, float top,
                                                   float near_z, float far_z) NLIB_NOEXCEPT {
    f128 div;
    {
        f128 a = F128::SetOne();
        f128 b = F128::SetValue(right - left, top - bottom, near_z - far_z, 1.f);
        div = F128::Div(a, b);
    }
    f128 zero = F128::SetZero();
    f128 v0 = F128::SetValue(2.f, 2.f, 1.f, 1.f);
    f128 r3 = F128::SetValue(-(left + right), -(top + bottom), near_z, 1.f);
    v0 = F128::Mult(v0, div);

    SimdMatrix m;
    m.r[0] = F128::Splat<false, true, true, true>(v0, zero);
    m.r[1] = F128::Splat<true, false, true, true>(v0, zero);
    m.r[2] = F128::Splat<true, true, false, true>(v0, zero);
    m.r[3] = F128::Mult(r3, div);
    return m;
}

// XMMatrixShadow
NLIB_M(SimdMatrix) Matrix::Shadow(SimdPlaneArg shadow_plane, SimdVector light_pos) NLIB_NOEXCEPT {
    // Plane::Normalize(shadow_plane);
    SimdPlane plane = F128::Mult(Vector3::RecpLength(shadow_plane), shadow_plane);
    // distance from plane to the light
    f128 r0 = Vector4::DotEx<true, false, false, false>(plane, light_pos);
    plane = F128::Negate(plane);
    f128 r1 = F128::RotateLeft<1>(r0);
    f128 r2 = F128::RotateLeft<2>(r0);
    f128 r3 = F128::RotateLeft<3>(r0);

    SimdMatrix m;
    m.r[0] = F128::MultAdd<0>(plane, light_pos, r0, each_select32);
    m.r[1] = F128::MultAdd<1>(plane, light_pos, r1, each_select32);
    m.r[2] = F128::MultAdd<2>(plane, light_pos, r2, each_select32);
    m.r[3] = F128::MultAdd<3>(plane, light_pos, r3, each_select32);
    return m;
}

// XMMatrixReflect
NLIB_M(SimdMatrix) Matrix::Reflect(SimdPlaneArg reflection_plane) NLIB_NOEXCEPT {
    // SimdPlane plane = Plane::Normalize(reflection_plane);
    SimdPlane plane = F128::Mult(Vector3::RecpLength(reflection_plane), reflection_plane);
    f128 minus_2n = F128::Mult(-2.f, plane);
    minus_2n = F128::SetZeroToLane<3>(minus_2n);

    SimdMatrix m = Matrix::Identity();
    m.r[0] = F128::MultAdd<0>(plane, minus_2n, m.r[0], each_select32);
    m.r[1] = F128::MultAdd<1>(plane, minus_2n, m.r[1], each_select32);
    m.r[2] = F128::MultAdd<2>(plane, minus_2n, m.r[2], each_select32);
    m.r[3] = F128::MultAdd<3>(plane, minus_2n, m.r[3], each_select32);
    return m;
}

// different from XMMatrixDecompose
// note that scale.xyz >= 0
NLIB_M(void) Matrix::Decompose(SimdVector* scale, SimdMatrix* rot, SimdVector* trans,  // NOLINT
                               SimdMatrixArg m) NLIB_NOEXCEPT {
    // translation
    *trans = m.r[3];

    // scaling
    f128 recp_scale;
    {
        f128 dot_x = Vector3::DotEx<true, false, false, true>(m.r[0], m.r[0]);
        f128 dot_y = Vector3::DotEx<false, true, false, true>(m.r[1], m.r[1]);
        f128 dot_z = Vector3::DotEx<false, false, true, true>(m.r[2], m.r[2]);
        f128 dot = F128::Or(dot_x, dot_y);
        dot = F128::Or(dot, dot_z);
        recp_scale = F128::RecpSqrt(dot);
        *scale = F128::Mult(dot, recp_scale);
    }

    // rotation
    rot->r[0] = F128::Mult<0>(recp_scale, m.r[0], each_select32);
    rot->r[1] = F128::Mult<1>(recp_scale, m.r[1], each_select32);
    rot->r[2] = F128::Mult<2>(recp_scale, m.r[2], each_select32);
    rot->r[3] = F128::Set0001();
}

#undef NLIB_M

#endif  // NLIB_DOXYGEN

}  // namespace simd
NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_SIMD_SIMDMATRIX_H_
