﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_SUCCINCT_LOUDS_H_
#define INCLUDE_NN_NLIB_SUCCINCT_LOUDS_H_

#include "nn/nlib/Swap.h"
#include "nn/nlib/succinct/Sbv.h"

#if defined(_MSC_VER) && defined(nx_succinct_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

NLIB_NAMESPACE_BEGIN
namespace succinct {

class NLIB_VIS_PUBLIC Louds NLIB_FINAL {
 public:
    Louds() NLIB_NOEXCEPT {}
    ~Louds() NLIB_NOEXCEPT {}
    NLIB_MOVE_MEMBER_HELPER_1(Louds, sbv_);
    void swap(Louds& rhs) NLIB_NOEXCEPT { sbv_.swap(rhs.sbv_); }
    template <class TREEITER>
    bool Init(TREEITER& iter, uint32_t num_nodes) NLIB_NOEXCEPT;  // NOLINT

    bool InitDirect(const uint32_t* bv, uint32_t bv_size) NLIB_NOEXCEPT;
    int ToNodeId(uint32_t pos) const NLIB_NOEXCEPT;
    int ToPos(uint32_t nodeid) const NLIB_NOEXCEPT;
    int FirstChild(uint32_t pos) const NLIB_NOEXCEPT;
    int NextSibling(uint32_t pos) const NLIB_NOEXCEPT;
    int PrevSibling(uint32_t pos) const NLIB_NOEXCEPT;
    int Parent(uint32_t pos) const NLIB_NOEXCEPT;
    bool IsLeaf(uint32_t pos) const NLIB_NOEXCEPT {
        return this->FirstChild(pos) < 0;
    }
    size_t MemSize() const NLIB_NOEXCEPT { return sbv_.MemSize(); }
    void Reset() NLIB_NOEXCEPT { sbv_.Reset(); }
    bool Export(BinaryWriter* w) const NLIB_NOEXCEPT { return sbv_.Export(w); }
    bool Import(BinaryReader* r) NLIB_NOEXCEPT { return sbv_.Import(r); }

 private:
    Sbv sbv_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Louds);
};

template <class TREEITER>
bool Louds::Init(TREEITER& iter, uint32_t num_nodes) NLIB_NOEXCEPT {  // NOLINT
    // TREEITER has following methods:
    //   int GetNumChildren();
    //   bool BfsMoveNext();  // move to the next node by breadth first search
    sbv_.Init(2 * num_nodes + 1);

    // diffrent from the Louds paper: reversed bits
    sbv_.TurnOn(1);
    uint32_t idx = 2;  // consider super-root 01
    do {
        idx += iter.GetNumChildren();
        sbv_.TurnOn(idx++);
    } while (iter.BfsMoveNext());

    NLIB_ASSERT(idx == 2 * num_nodes + 1);
    if (!sbv_.Build()) {
        sbv_.Reset();
        return false;
    }
    return true;
}

}  // namespace succinct
NLIB_NAMESPACE_END

#if defined(_MSC_VER) && defined(nx_succinct_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_SUCCINCT_LOUDS_H_
