﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_TESTING_TESTING_H_
#define INCLUDE_NN_NLIB_TESTING_TESTING_H_

#include "nn/nlib/Config.h"

#ifndef NLIB_USE_GTEST
#include "nn/nlib/testing/NTest.h"
#include "nn/nlib/testing/Pred.h"
#include "nn/nlib/testing/Printer.h"
#include "nn/nlib/testing/ParamValue.h"
#include "nn/nlib/testing/ParamType.h"

namespace testing = ::nlib_ns::testing;
NLIB_NAMESPACE_BEGIN
namespace testing {
inline void InitGoogleTest(int* argc, char** argv) { InitNintendoTest(argc, argv); }
inline void InitGoogleTest(int* argc, wchar_t** argv) { InitNintendoTest(argc, argv); }
#define NLIB_TESTING_OSTREAM ::nlib_ns::testing::detail_func::DummyStream
}  // namespace testing
NLIB_NAMESPACE_END
#else
// See: https://code.google.com/p/googletest/
#include "gtest/gtest.h"

#define NLIB_TESTING_OSTREAM std::ostream
namespace testing {
inline void InitNintendoTest(int* argc, char** argv) { InitGoogleTest(argc, argv); }
inline void InitNintendoTest(int* argc, wchar_t** argv) { InitGoogleTest(argc, argv); }
}

#endif

#if defined(_MSC_VER) && defined(_PREFAST_)
// for /analyze option
#include "nn/nlib/Cstring.h"
struct MsvcAnalyzeHelper_ {
    template<class T>
    MsvcAnalyzeHelper_& operator<<(T x) { return *this; }
    operator int() const { return 1; }
};
#undef FAIL
#define FAIL() throw MsvcAnalyzeHelper_()
#undef SUCCEED
#define SUCCEED()

#undef ASSERT_TRUE
#define ASSERT_TRUE(e) if (!(e)) throw MsvcAnalyzeHelper_()
#undef ASSERT_FALSE
#define ASSERT_FALSE(e) if (e) throw MsvcAnalyzeHelper_()
#undef EXPECT_TRUE
#define EXPECT_TRUE(e) if (!(e)) throw MsvcAnalyzeHelper_()
#undef EXPECT_FALSE
#define EXPECT_FALSE(e) if (e) throw MsvcAnalyzeHelper_()

#undef ASSERT_EQ
#define ASSERT_EQ(a, b) if (!((a) == (b))) throw MsvcAnalyzeHelper_()
#undef EXPECT_EQ
#define EXPECT_EQ(a, b) if (!((a) == (b))) throw MsvcAnalyzeHelper_()

#undef ASSERT_NE
#define ASSERT_NE(a, b) if ((a) == (b)) throw MsvcAnalyzeHelper_()
#undef EXPECT_NE
#define EXPECT_NE(a, b) if ((a) == (b)) throw MsvcAnalyzeHelper_()

#undef ASSERT_LE
#define ASSERT_LE(a, b) if ((a) > (b)) throw MsvcAnalyzeHelper_()
#undef EXPECT_LE
#define EXPECT_LE(a, b) if ((a) > (b)) throw MsvcAnalyzeHelper_()

#undef ASSERT_LT
#define ASSERT_LT(a, b) if ((a) >= (b)) throw MsvcAnalyzeHelper_()
#undef EXPECT_LT
#define EXPECT_LT(a, b) if ((a) >= (b)) throw MsvcAnalyzeHelper_()

#undef ASSERT_GE
#define ASSERT_GE(a, b) if ((a) < (b)) throw MsvcAnalyzeHelper_()
#undef EXPECT_GE
#define EXPECT_GE(a, b) if ((a) < (b)) throw MsvcAnalyzeHelper_()

#undef ASSERT_GT
#define ASSERT_GT(a, b) if ((a) <= (b)) throw MsvcAnalyzeHelper_()
#undef EXPECT_GT
#define EXPECT_GT(a, b) if ((a) <= (b)) throw MsvcAnalyzeHelper_()

#undef ASSERT_STREQ
#define ASSERT_STREQ(a, b) if (NLIB_NS::StrCmp((a), (b)) != 0) throw MsvcAnalyzeHelper_()
#undef EXPECT_STREQ
#define EXPECT_STREQ(a, b) if (NLIB_NS::StrCmp((a), (b)) != 0) throw MsvcAnalyzeHelper_()

#undef ASSERT_STRNE
#define ASSERT_STRNE(a, b) if (NLIB_NS::StrCmp((a), (b)) == 0) throw MsvcAnalyzeHelper_()
#undef EXPECT_STRNE
#define EXPECT_STRNE(a, b) if (NLIB_NS::StrCmp((a), (b)) == 0) throw MsvcAnalyzeHelper_()

#undef ASSERT_STRCASEEQ
#define ASSERT_STRCASEEQ(a, b) if (NLIB_NS::StrCmp((a), (b)) != 0) throw MsvcAnalyzeHelper_()
#undef EXPECT_STRCASEEQ
#define EXPECT_STRCASEEQ(a, b) if (NLIB_NS::StrCmp((a), (b)) != 0) throw MsvcAnalyzeHelper_()

#undef ASSERT_STRCASENE
#define ASSERT_STRCASENE(a, b) if (NLIB_NS::StrCmp((a), (b)) == 0) throw MsvcAnalyzeHelper_()
#undef EXPECT_STRCASENE
#define EXPECT_STRCASENE(a, b) if (NLIB_NS::StrCmp((a), (b)) == 0) throw MsvcAnalyzeHelper_()

#undef ASSERT_FLOAT_EQ
#define ASSERT_FLOAT_EQ(a, b) if ((a) != (b)) throw MsvcAnalyzeHelper_()
#undef EXPECT_FLOAT_EQ
#define EXPECT_FLOAT_EQ(a, b) if ((a) != (b)) throw MsvcAnalyzeHelper_()

#undef ASSERT_DOUBLE_EQ
#define ASSERT_DOUBLE_EQ(a, b) if ((a) != (b)) throw MsvcAnalyzeHelper_()
#undef EXPECT_DOUBLE_EQ
#define EXPECT_DOUBLE_EQ(a, b) if ((a) != (b)) throw MsvcAnalyzeHelper_()

#undef ASSERT_NEAR
#define ASSERT_NEAR(a, b, c) if (!((a) == (b))) throw MsvcAnalyzeHelper_()
#undef EXPECT_NEAR
#define EXPECT_NEAR(a, b, c) if (!((a) == (b))) throw MsvcAnalyzeHelper_()

#endif


#endif  // INCLUDE_NN_NLIB_TESTING_TESTING_H_
