﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#ifndef SAMPLES_XML_RPC_XMLRPC_CLIENTSERVER_H_  // NOLINT
#define SAMPLES_XML_RPC_XMLRPC_CLIENTSERVER_H_

#include <map>
#include <string>
#include <vector>

#include "nn/nlib/exi/exi.h"

// Read http://www.xmlrpc.com/spec for XML-RPC
// (In Japanese, http://lowlife.jp/yasusii/stories/9.html)
class XmlRpcServer;
class XmlRpcValue;

#ifdef NLIB_EXI_WCHAR_INTERFACE
typedef std::wstring StdString;
#else
typedef std::string StdString;
#endif

using nlib_ns::exi::ExiChar;

// Server Method:
// Implements the server method to inherit this class.
// The constructor registers the method with the server(XmlRpcServer),
// and the destructor unregisters it.
class XmlRpcServerMethod {
 public:
    XmlRpcServerMethod(const StdString& name, XmlRpcServer* server);
    virtual ~XmlRpcServerMethod();
    virtual void Execute(const XmlRpcValue& params, XmlRpcValue* result) = 0;
    const StdString& GetName() const { return name_; }
    XmlRpcServer* GetServer() const { return server_; }

 private:
    StdString name_;
    XmlRpcServer* server_;

 private:
    XmlRpcServerMethod(const XmlRpcServerMethod&);
    XmlRpcServerMethod& operator=(const XmlRpcServerMethod&);
};

// Server:
// Receives XML and converts it into the XmlRpcValue object,
// then calls the server method corresponding to the name of the function.
class XmlRpcServer {
 public:
    XmlRpcServer() {}
    ~XmlRpcServer() {}
    void AddServerMethod(XmlRpcServerMethod* method) { methods_[method->GetName()] = method; }
    void RemoveServerMethod(const StdString& name) {
        Methods::iterator it = methods_.find(name);
        if (it != methods_.end()) methods_.erase(it);
    }
    void RemoveServerMethod(XmlRpcServerMethod* method) {
        this->RemoveServerMethod(method->GetName());
    }
    XmlRpcServerMethod* FindServerMethod(const StdString& name) const {
        Methods::const_iterator it = methods_.find(name);
        return (it != methods_.end()) ? it->second : NULL;
    }
    bool Process(nlib_ns::exi::XmlStreamReader* reader, nlib_ns::exi::XmlStreamWriter* writer);

 private:
    typedef std::map<StdString, XmlRpcServerMethod*> Methods;
    Methods methods_;
};

// This class is for the client to receive the result of the procedure.
class XmlRpcCallResult {
    XmlRpcServer* server_;
    std::vector<unsigned char> to_server_;

 public:
    XmlRpcCallResult() : server_(NULL) {}
    void SetServer(XmlRpcServer* server) { server_ = server; }
    void SetData(const std::vector<unsigned char>& rhs) { to_server_ = rhs; }
    bool GetResult(XmlRpcValue* result);
};

// Client:
// Sends the method name and the parameters to the server,
// and receives the result via XmlRpcCallResult.
class XmlRpcClient {
    XmlRpcServer* server_;

 public:
    XmlRpcClient() : server_(NULL) {}
    void SetServer(XmlRpcServer* server) { server_ = server; }
    bool Execute(const ExiChar* methodName, const XmlRpcValue& params, XmlRpcCallResult* result);
};

#endif  // SAMPLES_XML_RPC_XMLRPC_CLIENTSERVER_H_ // NOLINT
