﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#include <ctype.h>
#include <math.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <string>

// define NLIB_USE_GTEST to use googletest
// #define NLIB_USE_GTEST
#include "nn/nlib/testing/testing.h"
#include "nn/nlib/NativePathMapper.h"

#ifdef _MSC_VER
#pragma warning(disable : 4996)
#endif

//
// stdlib.h
//
TEST(stdlib, abs) {
    EXPECT_EQ(1, abs(1));
    EXPECT_EQ(0, abs(0));
    EXPECT_EQ(1, abs(-1));
}

TEST(stdlib, labs) {
    EXPECT_EQ(1L, labs(1L));
    EXPECT_EQ(0L, labs(0L));
    EXPECT_EQ(1L, labs(-1L));
}

TEST(stdlib, atof) {
    EXPECT_DOUBLE_EQ(1.0, atof("1.0"));
    EXPECT_DOUBLE_EQ(-1.0, atof("-1.0"));
    EXPECT_DOUBLE_EQ(2.0e-11, atof("2.0e-11"));
    EXPECT_DOUBLE_EQ(2.0e+100, atof("2.0e+100"));

    EXPECT_DOUBLE_EQ(0.0, atof("test1"));
}

TEST(stdlib, atoi) {
    EXPECT_EQ(0, atoi("0"));
    EXPECT_EQ(-1, atoi("-1"));
    EXPECT_EQ(0, atoi("test1"));
}

TEST(stdlib, atol) {
    EXPECT_EQ(0L, atol("0"));
    EXPECT_EQ(-1L, atol("-1"));
    EXPECT_EQ(0L, atol("test1"));
}

TEST(stdlib, strtod) {
    const char str1[] = "1.0test";
    char* p;
    EXPECT_EQ(1.0, strtod(str1, &p));
    EXPECT_EQ(&str1[3], p);

    const char str2[] = "1.0e";
    EXPECT_EQ(1.0, strtod(str2, &p));
    EXPECT_EQ(&str2[3], p);

    const char str3[] = "1.0e+";
    EXPECT_EQ(1.0, strtod(str3, &p));
    EXPECT_EQ(&str3[3], p);

    const char str4[] = "+";
    EXPECT_EQ(0.0, strtod(str4, &p));
    EXPECT_EQ(&str4[0], p);
}

TEST(stdlib, strtol) {
    const char str1[] = "123test";
    char* p;
    EXPECT_EQ(123L, strtol(str1, &p, 10));
    EXPECT_EQ(&str1[3], p);
}

TEST(stdlib, strtoul) {
    const char str1[] = "123test";
    char* p;
    EXPECT_EQ(123U, strtoul(str1, &p, 10));
    EXPECT_EQ(&str1[3], p);
}

//
// ctype.h
//
TEST(ctype, isalpha) {
    for (int c = 0; c < 0x80; ++c) {
        if ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z'))
            EXPECT_NE(0, isalpha(c));
        else
            EXPECT_EQ(0, isalpha(c));
    }
}

TEST(ctype, isupper) {
    for (int c = 0; c < 0x80; ++c) {
        if (c >= 'A' && c <= 'Z')
            EXPECT_NE(0, isupper(c));
        else
            EXPECT_EQ(0, isupper(c));
    }
}

TEST(ctype, islower) {
    for (int c = 0; c < 0x80; ++c) {
        if (c >= 'a' && c <= 'z')
            EXPECT_NE(0, islower(c));
        else
            EXPECT_EQ(0, islower(c));
    }
}

TEST(ctype, isdigit) {
    for (int c = 0; c < 0x80; ++c) {
        if (c >= '0' && c <= '9')
            EXPECT_NE(0, isdigit(c));
        else
            EXPECT_EQ(0, isdigit(c));
    }
}

TEST(ctype, isspace) {
    for (int c = 0; c < 0x80; ++c) {
        if (c == ' ' || c == '\f' || c == '\n' || c == '\r' || c == '\t' || c == '\v')
            EXPECT_NE(0, isspace(c));
        else
            EXPECT_EQ(0, isspace(c));
    }
}

TEST(ctype, isalmum) {
    for (int c = 0; c < 0x80; ++c) {
        if ((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z') || (c >= '0' && c <= '9'))
            EXPECT_NE(0, isalnum(c));
        else
            EXPECT_EQ(0, isalnum(c));
    }
}

TEST(ctype, iscntrl) {
    for (int c = 0; c < 0x80; ++c) {
        if (c <= 0x1F || c == 0x7F)
            EXPECT_NE(0, iscntrl(c)) << c;
        else
            EXPECT_EQ(0, iscntrl(c)) << c;
    }
}

TEST(ctype, isprint) {
    for (int c = 0; c < 0x80; ++c) {
        if (c <= 0x1F || c == 0x7F)
            EXPECT_EQ(0, isprint(c)) << c;
        else
            EXPECT_NE(0, isprint(c)) << c;
    }
}

TEST(ctype, ispunct) {
    for (int c = 0; c < 0x80; ++c) {
        if (isprint(c) != 0 && c != 0x20 && isalnum(c) == 0)
            EXPECT_NE(0, ispunct(c)) << c;
        else
            EXPECT_EQ(0, ispunct(c)) << c;
    }
}

TEST(ctype, isxdigit) {
    for (int c = 0; c < 0x80; ++c) {
        if (isdigit(c) || (c >= 'a' && c <= 'f') || (c >= 'A' && c <= 'F'))
            EXPECT_NE(0, isxdigit(c)) << c;
        else
            EXPECT_EQ(0, isxdigit(c)) << c;
    }
}

//
// string.h
//
TEST(string, strcpy) {  // NOLINT
    const char str[] = "abcde";
    char p[256];
    EXPECT_EQ(0, nlib_strcpy(p, str));
    EXPECT_STREQ(str, p);
}

TEST(string, strncpy) {
    const char str[] = "abcde";
    char p[256];
    p[5] = '#';
    p[6] = '\0';
    EXPECT_EQ(0, nlib_strncpy(p, str, 5));
    EXPECT_STREQ("abcde", p);
}

TEST(string, strlen) {
    EXPECT_EQ(5U, strlen("abcde"));
}

TEST(string, strcmp) {
    char abcde[] = "abcde";
    char ABCDE[] = "ABCDE";

    EXPECT_EQ(0, strcmp(abcde, abcde));
    EXPECT_LT(strcmp(ABCDE, abcde), 0);
    EXPECT_GT(strcmp(abcde, ABCDE), 0);
}

//
// math
//
TEST(math, fabs) {
    EXPECT_EQ(1.0, fabs(1.0));
    EXPECT_EQ(1.0, fabs(-1.0));
}

TEST(math, sqrt) {
    EXPECT_DOUBLE_EQ(2.0, sqrt(4.0));
    EXPECT_DOUBLE_EQ(0.0, sqrt(0.0));
}

TEST(math, pow) {
    EXPECT_DOUBLE_EQ(8.0, pow(2.0, 3.0));
}

TEST(math, fmod) {
    EXPECT_DOUBLE_EQ(2.0, fmod(6.0, 4.0));
}

//
// snprintf
//
TEST(snprintf, Int) {
    char buf[256];

    int i = -100000;
    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%d", i));
    EXPECT_STREQ("-100000", buf);
}

TEST(snprintf, unsignedInt) {
    char buf[256];

    unsigned int i = 100000;
    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%u", i));
    EXPECT_STREQ("100000", buf);
}

TEST(snprintf, signedChar) {
    char buf[256];

    signed char c = 'A';
    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%hhd", c));
    EXPECT_STREQ("65", buf);
}

TEST(snprintf, unsignedChar) {
    char buf[256];

    unsigned char c = 'B';
    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%hhu", c));
    EXPECT_STREQ("66", buf);
}

TEST(snprintf, shortInt) {
    char buf[256];

    short s = -10000;  // NOLINT
    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%hd", s));
    EXPECT_STREQ("-10000", buf);
}

TEST(snprintf, unsignedShortInt) {
    char buf[256];

    unsigned short s = 10000;  // NOLINT
    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%hu", s));
    EXPECT_STREQ("10000", buf);
}

TEST(snprintf, longInt) {
    char buf[256];

    long s = -100000;  // NOLINT
    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%ld", s));
    EXPECT_STREQ("-100000", buf);
}

TEST(snprintf, unsignedLongInt) {
    char buf[256];

    unsigned long s = 100000;  // NOLINT
    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%lu", s));
    EXPECT_STREQ("100000", buf);
}

TEST(snprintf, longLongInt) {
    char buf[256];

    long long s = -100000;  // NOLINT
    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%lld", s));
    EXPECT_STREQ("-100000", buf);
}

TEST(snprintf, unsignedLongLongInt) {
    char buf[256];

    unsigned long long s = 100000;  // NOLINT
    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%llu", s));
    EXPECT_STREQ("100000", buf);
}

TEST(snprintf, sizet) {
    char buf[256];

    size_t s = 10000;
    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%" PRIuS, s));
    EXPECT_STREQ("10000", buf);
}

#ifndef _MSC_VER
TEST(snprintf, intmaxt) {
    char buf[256];

    intmax_t s = 10000;
    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%jd", s));
    EXPECT_STREQ("10000", buf);
}

TEST(snprintf, uintmaxt) {
    char buf[256];

    uintmax_t s = 10000;
    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%ju", s));
    EXPECT_STREQ("10000", buf);
}
#endif

TEST(snprintf, ptrdifft) {
    char buf[256];

    ptrdiff_t s = 10000;
    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%" PRIdS, s));
    EXPECT_STREQ("10000", buf);
}

TEST(snprintf, int32t) {
    char buf[256];

    int32_t s = 10000;
    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%" PRId32, s));
    EXPECT_STREQ("10000", buf);
}

TEST(snprintf, uint32t) {
    char buf[256];

    uint32_t s = 10000;
    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%" PRIu32, s));
    EXPECT_STREQ("10000", buf);
}

TEST(snprintf, int64t) {
    char buf[256];

    int64_t s = 10000;
    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%" PRId64, s));
    EXPECT_STREQ("10000", buf);
}

TEST(snprintf, uint64t) {
    char buf[256];

    uint64_t s = 10000;
    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%" PRIu64, s));
    EXPECT_STREQ("10000", buf);
}

TEST(snprintf, float) {
    char buf[256];
    float v = 1.f;
    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%f", v));
    EXPECT_STREQ("1.000000", buf);

    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%e", 1.0e+10f));

    EXPECT_TRUE(strcmp("1.000000e+10", buf) == 0 || strcmp("1.000000e+010", buf) == 0);  // MSVC
}

TEST(snprintf, double) {
    char buf[256];
    double v = 1.0;
    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%f", v));
    EXPECT_STREQ("1.000000", buf);

    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%e", 1.0e+100));
    EXPECT_STREQ("1.000000e+100", buf);
}

TEST(snprintf, flags) {
    char buf[256];
    int d = 100;

    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%-6d", d));
    EXPECT_STREQ("100   ", buf);

    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%+d", d));
    EXPECT_STREQ("+100", buf);

    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%06d", d));
    EXPECT_STREQ("000100", buf);

    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "% 6d", d));
    EXPECT_STREQ("   100", buf);
}

TEST(snprintf, hex) {
    char buf[256];

    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%x", 254));
    EXPECT_STREQ("fe", buf);

    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%X", 254));
    EXPECT_STREQ("FE", buf);
}

TEST(snprintf, char) {
    char buf[256];

    ASSERT_EQ(0, nlib_snprintf(NULL, buf, "%c", 65));
    EXPECT_STREQ("A", buf);
}

#ifndef XMLFILE
#define XMLFILE "libc_test.xml"
#endif

NLIB_PATHMAPPER_FORSAMPLE

bool SampleMain(int argc, char** argv) {
    InitPathMapperForSample();
    char path[512];
    char buf[512];
    size_t count;
    g_PathMapper.ResolvePath(&count, path, "nlibpath:///readwrite/libc_test.xml");
    nlib_snprintf(&count, buf, "xml:%s", path);

    ::testing::GTEST_FLAG(output) = buf;
    ::testing::InitNintendoTest(&argc, argv);

    // RUN_ALL_TESTS returns 0 if all the tests are successful.
    // You can use the return value of main() function.
    return RUN_ALL_TESTS() == 0;
}

NLIB_MAINFUNC
