#!/bin/sh

if ! type uname > /dev/null 2>&1; then
  echo Cannot find uname.
  exit 1
fi

if ! [[ $(uname) =~ CYGWIN.* ]]; then
  echo Please call this script from cygwin.
  exit 1
fi

if ! type git > /dev/null 2>&1; then
  echo Please install git manually.
  exit 1
fi

if ! type wget > /dev/null 2>&1; then
  echo Please install wget manually.
  exit 1
fi

if ! type tar > /dev/null 2>&1; then
  echo Please install tar manually.
  exit 1
fi

if ! type gawk > /dev/null 2>&1; then
  echo Please install gawk manually.
  exit 1
fi

if ! type gpg > /dev/null 2>&1; then
  echo Please install gpg manually.
  exit 1
fi

if ! type gpgv > /dev/null 2>&1; then
  echo Please install gpgv manually.
  exit 1
fi

if [ ! -e ~/.wgetrc ]; then
  set +u
  echo "" > ~/.wgetrc
  if ! [ "${http_proxy}" = "" ]; then
    echo "http_proxy=${http_proxy}" >> ~/.wgetrc
  elif ! [ "${HTTP_PROXY}" = "" ]; then
    echo "http_proxy=${HTTP_PROXY}" >> ~/.wgetrc
  fi
  if ! [ "${https_proxy}" = "" ]; then
    echo "https_proxy=${https_proxy}" >> ~/.wgetrc
  elif ! [ "${HTTPS_PROXY}" = "" ]; then
    echo "https_proxy=${HTTPS_PROXY}" >> ~/.wgetrc
  fi
  set -u
fi

cd $(dirname $0)

if [ ! -e /usr/local/bin/apt-cyg ]; then
  echo installing apt-cyg
  git clone https://github.com/kou1okada/apt-cyg.git
  chmod +x apt-cyg/apt-cyg && cp ./apt-cyg/apt-cyg /usr/local/bin/
fi

# apt-cyg --mirror http://ftp.iij.ad.jp/pub/cygwin/
apt-cyg update
apt-cyg install tar gzip zip unzip p7zip binutils procps-ng
apt-cyg install gcc-g++ make cmake
apt-cyg install zlib zlib-devel libcurl-devel libboost-devel libxml2
apt-cyg install python python3 python-setuptools
apt-cyg install gdb gperf cccc
apt-cyg install doxygen graphviz
apt-cyg install openssh flex jq rsync

if [ ! -e /usr/local/bin/gcovr ]; then
  # rebaseall may be needed
  easy_install-2.7 pip
  pip install gcovr
fi

if [ ! -e ./sloccount-2.26 ]; then
 wget -O - http://www.dwheeler.com/sloccount/sloccount-2.26.tar.gz | tar zxvf -
 cd sloccount-2.26
 make EXE_SUFFIX=.exe
 make install
fi

echo "You may have to do '/usr/bin/rebaseall' and '/usr/bin/peflagsall' on ash"
