#!/bin/sh

if ! type uname > /dev/null 2>&1; then
  echo Cannot find uname.
  exit 1
fi

if ! [[ $(uname) =~ CYGWIN.* ]]; then
  echo Please call this script from cygwin.
  exit 1
fi

set -eu
id -G | grep -qE '\<(544|0)\>' || { echo setup-msvc-libraries.sh: Please run as admin again to copy files; exit 1 ; }

#
# NOTE:
# setup-cygwin-tools.sh required
#

cd $(dirname $0)

#
# myconv
#
set +e
if [[ ${LANG} =~ UTF-8.* ]]; then
  myconv="iconv -c -f SJIS -t UTF-8"
else
  myconv="cat"
fi

#
# Find cmake for Windows
#
path_remove() { export PATH=$(echo -n $PATH | /usr/bin/awk -v RS=: -v ORS=: '$0 != "'$1'"' | /usr/bin/sed 's/:$//'); }
path_remove /usr/bin
path_remove /usr/local/bin

WINCMAKE=$(/usr/bin/which cmake)
if [ $? -ne 0 ]; then
  echo cmake for windows not found
  exit 1;
fi
export PATH=/usr/local/bin:/usr/bin:${PATH}

pg32=$(env | grep ^ProgramFiles\(x86\) | cut -d= -f2)
if [[ $(uname -m) =~ .*64.* ]]; then
  sys64=System32
else
  sys64=sysnative
fi
set -e

#
# Detect Visual Studio
#
if [ "${VS140COMNTOOLS}" = "" ]; then
if [ "${VS120COMNTOOLS}" = "" ]; then
if [ "${VS110COMNTOOLS}" = "" ]; then
  echo "Sorry, Visual Studio 2012/2013/2015 not found"
  exit 1
else
  vc32="Visual Studio 11 2012"
  vc64="Visual Studio 11 2012 Win64"
fi
else
  vc32="Visual Studio 12 2013"
  vc64="Visual Studio 12 2013 Win64"
fi
else
  vc32="Visual Studio 14 2015"
  vc64="Visual Studio 14 2015 Win64"
fi

#
# Natvis
#
userprofile=$(cygpath -u $USERPROFILE)
if [ -e "$userprofile/My Documents/Visual Studio 2015" ]; then
  mkdir -p "$userprofile/My Documents/Visual Studio 2015/Visualizers/"
  cp ./nlib.natvis "$userprofile/My Documents/Visual Studio 2015/Visualizers/nlib.natvis"
fi
if [ -e "$userprofile/My Documents/Visual Studio 2013" ]; then
  mkdir -p "$userprofile/My Documents/Visual Studio 2013/Visualizers/"
  cp ./nlib.natvis "$userprofile/My Documents/Visual Studio 2013/Visualizers/nlib.natvis"
fi
if [ -e "$userprofile/My Documents/Visual Studio 2012" ]; then
  mkdir -p "$userprofile/My Documents/Visual Studio 2012/Visualizers/"
  cp ./nlib.natvis "$userprofile/My Documents/Visual Studio 2012/Visualizers/nlib.natvis"
fi

#
# zlib
#
if [ ! -e ./.cache/zlib128.zip ]; then
  wget -P ./.cache/ http://www.zlib.net/zlib128.zip
fi

unzip ./.cache/zlib128.zip

pushd .
cd zlib-1.2.8
"$WINCMAKE" -G "${vc32}" -H"." -Bvc32
"$WINCMAKE" --build vc32 --config Release --target INSTALL | ${myconv}
"$WINCMAKE" --build vc32 --config Debug --target INSTALL | ${myconv}
"$WINCMAKE" -G "${vc64}" -H"." -Bvc64
"$WINCMAKE" --build vc64 --config Release --target INSTALL | ${myconv}
"$WINCMAKE" --build vc64 --config Debug --target INSTALL | ${myconv}
popd

#
# libcurl
#
if [ ! -e ./.cache/curl-7.48.0.zip ]; then
  wget -P ./.cache/ https://curl.haxx.se/download/curl-7.48.0.zip
fi

unzip ./.cache/curl-7.48.0.zip
pushd .
cd curl-7.48.0

"$WINCMAKE" -G "${vc32}" -H"." -Bvc32 -DBUILD_CURL_TESTS=OFF -DBUILD_CURL_EXE=OFF -DZLIB_ROOT="${pg32}/zlib"
"$WINCMAKE" --build vc32 --config Release --target INSTALL | ${myconv}


"$WINCMAKE" -G "${vc64}" -H"." -Bvc64 -DBUILD_CURL_TESTS=OFF -DBUILD_CURL_EXE=OFF -DZLIB_ROOT="${ProgramW6432}/zlib"
"$WINCMAKE" --build vc64 --config Release --target INSTALL | ${myconv}
popd

#
# expat
#
if [ -e ./_common/software/expat-2.1.0.tar.gz ]; then
  # only for tests
  gzip -dc ../software/expat-2.1.0.tar.gz | tar -xvf -
  pushd .
  cd expat-2.1.0
  "$WINCMAKE" -G "${vc32}" -H"." -Bvc32
  "$WINCMAKE" --build vc32 --config Release --target INSTALL | ${myconv}
  "$WINCMAKE" -DZLIB_ROOT="${ProgramW6432}/zlib" -G "${vc64}" -H"." -Bvc64
  "$WINCMAKE" --build vc64 --config Release --target INSTALL | ${myconv}
  popd
fi



